/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef INC_APP_H
#define INC_APP_H

/********************************* CONFIGURATION ******************************/

/* ********************** INCLUDE ************************** */
#include <cairo.h> //libcairo2-dev
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <gio/gio.h>
#include <gtk/gtk.h>
#include <inttypes.h>
#include <netdb.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#ifdef LIB_HANDY
#  include <handy.h>
#endif /* LIB_HANDY */
#ifdef MODE_DEV
#  include "dev_mode.h"
#endif /* MODE_DEV */

/* *********************** DEFINE ************************** */
#define APP_NB_STRINGS              261  /* Last string number + 1 */
#define APP_SH_NAME                 "updates-warner"
#define APP_VERSION                 "2.0:210422"

#define SIZE_BUFFER_ERROR           256  /* Buffer size to provide error */
#define SIZE_BUFFER_LANG            2048 /* Buffer size to read the language file */
#define SIZE_BUFFER_XOR             8    /* Buffer size for the trigger_xor */

#define STEP_BUFFER_FILE            256  /* Step to allocate */
#define STEP_SIZE_BUFFER_READ       128  /* Step to allocate the buffer used to retrieve local version */
#define STEP_SIZE_WATCHER_TABLE     10   /* Step to allocate the watcher table */

#define DEFAULT_CONFIG_FILE         APP_SH_NAME".conf"
#define DEFAULT_COUNT_HEIGHT        10   /* Default count of watcher in height */
#define DEFAULT_COUNT_WIDTH         10   /* Default count of watcher in width */
#define DEFAULT_FOLDER_CONFIG       ".config"
#define DEFAULT_FOLDER_CACHE        ".cache/"APP_SH_NAME
#define DEFAULT_ICON_WIDTH          26   /* Default width of icons in main window */
#define DEFAULT_ICON_HEIGHT         26   /* Default height of icons in main window */
#define DEFAULT_TIMER               1440 /* In minutes, default timer between 2 downloads for a watcher */

#if GTK_MAJOR_VERSION >= 3
# define DEFAULT_COLOR_RED           0.5  /* Default level of red color */
# define DEFAULT_COLOR_GREEN         0.5  /* Default level of green color */
# define DEFAULT_COLOR_BLUE          0.5  /* Default level of blue color */
# define DEFAULT_COLOR_ALPHA         0.25 /* Default percent of opacity */
#else
# define DEFAULT_COLOR_RED           32768 /* Default level of red color */
# define DEFAULT_COLOR_GREEN         32768 /* Default level of green color */
# define DEFAULT_COLOR_BLUE          32768 /* Default level of blue color */
# define DEFAULT_OPACITY             0.25  /* Default percent of opacity */
#endif

#define MAX_COUNT_FILE_ERROR        10  /* Count limit of errors in s_file.Error */
#define MAX_COUNT_HEIGHT            64  /* Count limit of watchers in height */
#define MAX_COUNT_OPERATION         5   /* Count limit of number of operation when anaylsing resource */
#define MAX_COUNT_TEMPLATE          100 /* Count limit of templates */
#define MAX_COUNT_WIDTH             64  /* Count limit of watchers in width */

#define MAX_LENGTH_NAME             30  /* Length limit for the watcher name */
#define MAX_LENGTH_VERSION          50  /* Length limit for a version; must be < SIZE_BUFFER_ERROR */
#define MAX_LENGTH_PATTERN          50  /* Length limit for a pattern */
#define MAX_LENGTH_CONF             (2048 * 1024) /* Size limit for the configuration file: worst case scenario: 64x64 watchers means 512 bytes per watchers */
#define MAX_STD_OUPUT_LENGTH        (256 * 1024)  /* Limit of memory allocated to buffer command output; 256kb should be more than enough */
#define MAX_LANG_LINE_LENGTH        200 /* Maximum length for a line in the language file; should be more than enough */

/* ************************ MODE *************************** */
#ifdef MODE_RELEASE
#  define APP_NAME                 "Updates warner"
#  define DEFAULT_APP_RESOURCES    "/usr/share/" APP_SH_NAME
#  define APP_LANG_DIR             DEFAULT_APP_RESOURCES "/languages"
#  define ASSERT(condition, str, ...)
#  define SNO(condition, str, ...)
#  define LOCK_MUTEX( val )        pthread_mutex_lock( &val )
#  define UNLOCK_MUTEX( val )      pthread_mutex_unlock( &val )
/* #  if GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION >= 6 */
/* #    define LOCK_GTK() */
/* #    define UNLOCK_GTK() */
/* #  else */
#    define LOCK_GTK()               gdk_threads_enter();
#    define UNLOCK_GTK()             gdk_threads_leave();
/* #  endif */
#elif ! defined MODE_DEV
#  error "MODE_RELEASE or MODE_DEV must be defined"
#endif /* MODE_RELEASE */

/**** GTK VERSIONS ****/
#if GTK_MAJOR_VERSION < 3
#  define APP_GTK_HBOX( homogeneous, spacing )       gtk_hbox_new( homogeneous, spacing )
#  define APP_GTK_VBOX( homogeneous, spacing )       gtk_vbox_new( homogeneous, spacing )
#  define APP_GTK_HSEPAR()                           gtk_hseparator_new()
#  define APP_GTK_VSEPAR()                           gtk_vseparator_new()
#  define APP_GTK_TOGGLE_SET_ACTIVE( widget, state ) gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(widget), state )
#  define APP_GTK_TOGGLE_GET_ACTIVE( widget )        gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(widget) )
#elif GTK_MAJOR_VERSION == 3
#  define APP_GTK_HBOX( homogeneous, spacing )       gtk_box_new( GTK_ORIENTATION_HORIZONTAL, spacing )
#  define APP_GTK_VBOX( homogeneous, spacing )       gtk_box_new( GTK_ORIENTATION_VERTICAL, spacing )
#  define APP_GTK_HSEPAR()                           gtk_separator_new( GTK_ORIENTATION_HORIZONTAL )
#  define APP_GTK_VSEPAR()                           gtk_separator_new( GTK_ORIENTATION_VERTICAL )
#    define APP_GTK_TOGGLE_SET_ACTIVE( widget, state ) gtk_switch_set_active( GTK_SWITCH(widget), state )
#    define APP_GTK_TOGGLE_GET_ACTIVE( widget )        gtk_switch_get_active( GTK_SWITCH(widget) )
#  if GTK_MINOR_VERSION >= 10
//    gtk_image_menu_item_set_image deprecated
#  endif
#  if GTK_MINOR_VERSION >= 14
//    gtk_status_icon_set_visible deprecated
#  endif
#  if GTK_MINOR_VERSION >= 22
// gdk_window_begin_draw_frame
#  else
// gdk_window_begin_paint_rect
#  endif
#endif

/**** GDK VERSIONS ****/
#if GDK_MAJOR_VERSION == 3
#  if GDK_MINOR_VERSION >= 22
#    define SCREEN_WIDTH_HEIGHT( var_width, var_height )		\
  GdkDisplay *disp = gdk_display_get_default();				\
  var_width = 300;							\
  var_height = 300;							\
  if( disp ) {								\
    GdkMonitor *monitor = gdk_display_get_monitor_at_point( disp, app->Gui.MainX, app->Gui.MainY ); \
    if( monitor ) {							\
      GdkRectangle geometry;						\
      gdk_monitor_get_geometry( monitor, &geometry );			\
      var_width = geometry.width;					\
      var_height = geometry.height;					\
    }									\
  }
#  elif GDK_MINOR_VERSION >= 10
#    warning "gdk_screen_get_monitor_scale_factor() should be used"
#    define SCREEN_WIDTH_HEIGHT( width, height )        width = gdk_screen_width(); height = gdk_screen_height()
#  else
#    define SCREEN_WIDTH_HEIGHT( width, height )        width = gdk_screen_width(); height = gdk_screen_height()
#  endif
#else
#  define SCREEN_WIDTH_HEIGHT( width, height )          width = gdk_screen_width(); height = gdk_screen_height()
#endif

/* ************************ ENUM *************************** */
enum {
  IMG_DEFAULT_WATCHER        /* Id of the default watcher image */
  , IMG_APP                  /* id of the application icon */
  , IMG_ABOUT                /* id of the about icon */
  , IMG_ADD                  /* id of the add icon */
  , IMG_DEL                  /* id of the del icon */
  , IMG_FILE                 /* Id of the file image */
  , IMG_INFOS_BOTTOM         /* Id of the infos_view on bottom image */
  , IMG_INFOS_LEFT           /* Id of the infos_view on left image */
  , IMG_INFOS_RIGHT          /* Id of the infos_view on right image */
  , IMG_INFOS_TOP            /* Id of the infos_view on top image */
  , IMG_MODIFY               /* Id of the 'modify watcher' image */
  , IMG_SETTINGS             /* Id of the 'settings' image */
  , IMG_REFRESH              /* Id of the 'refresh systray' image */
  , IMG_VERSION_ERROR        /* Id of the error warning image */
  , IMG_VERSION_MISMATCH     /* Id of the mismatch warning image */
  , IMG_VERSION_IN_PROGRESS  /* Id of the 'in progress' image */

  , IMG_COUNT                /* MUST BE THE LAST ELEMENT - Count of IMG_ */
};

enum Gui_watcher_list_column {
  COL_STATE = 0       /* Column containg the state of watcher */
  , COL_LOGO          /* Column containg logo of watcher */
  , COL_NAME          /* Column containg Name of watcher */

  , COL_COUNT         /* MUST BE THE LAST ELEMENT - Count of COL_ */
};

enum AboutWindow_widget {
  ABW_WIN_MAIN              /* Window About... */

  , ABW_WIDGET_COUNT        /* MUST BE THE LAST ELEMENT - Count of widgets in the about window */
};

enum ErrorWindow_widget {
  ERW_WIN_MAIN              /* Window Error... */
  , ERW_LBL_MESSAGE         /* Label to print error message */

  , ERW_WIDGET_COUNT        /* MUST BE THE LAST ELEMENT - Count of widgets in the error window */
};

enum MainWindow_widget {
  MW_WIN_MAIN               /* Main window of application */
  , MW_MENU_WATCHER         /* Menu of watcher with right click */
  , MW_ROW_NAME             /* Menu row - Name of watcher */
  , MW_ROW_LAUNCH           /* Menu row - Launch without argument */
  , MW_ROW_LOCAL_UPDATE     /* Menu row - Update local version */
  , MW_ROW_OFFI_UPDATE      /* Menu row - Update official version */
  , MW_ROW_ALL_OFFI_UPDATE  /* Menu row - Update all official versions */
  , MW_ROW_GOTO_WEB         /* Menu row - Go to web site... */
  , MW_ROW_SETTINGS         /* Menu row - Settings */
  , MW_ROW_EDIT             /* Menu row - Edit watcher */
  , MW_TABLE_MAIN           /* Main table container of the main window */

  , MW_WIDGET_COUNT         /* MUST BE THE LAST ELEMENT - Count of widgets in the window */
};

enum WatcherSettings_widget {
  WS_WIN_MAIN               /* Window of watcher settings */
  , WS_ADD_BLOCK_OFFICIAL   /* Menu to add a new block in the official search */
  , WS_ADD_BLOCK_LOCAL      /* Menu to add a new block in the local search */
#ifdef LIB_HANDY
  , WS_BAR_WIN              /* Window's top bar */
#endif
  , WS_BOX_VERSION_AUTO     /* Box containing widget relative to the automatic local version */
  , WS_BUT_ADD              /* Button to add a watcher */
  , WS_BUT_EXEC             /* Button to choose the executable for the local version */
  , WS_BUT_LOCAL_FROM_OFFICIAL /* Button to put the official version to the manual local version entry */
  , WS_BUT_LOGO             /* Button to choose the logo for a new watcher */
  , WS_BUT_SAVE             /* Button to save the modifications of a watcher */
  , WS_BUT_SEARCH_OFFICIAL_ADD /* Button to add a search block to official version */
  , WS_ETY_ARGUMENT         /* Entry to set argument of local executable */
  , WS_ETY_NAME             /* Entry to set watcher name */
  , WS_ETY_LOCAL_EXEC       /* Entry to set watcher local executable */
  , WS_ETY_LOCAL_VERSION    /* Entry to set watcher local version */
  , WS_ETY_OFFI_HEADER      /* Entry to set options to HTTP header */
  , WS_ETY_OFFI_RESULT      /* Entry to set the result for the official version */
  , WS_ETY_OFFI_URL         /* Entry to set the url to the page containing the official version */
  , WS_ETY_TRIGGER_ERROR    /* Entry to set the action to execute when official version parsing failed */
  , WS_ETY_TRIGGER_MISMATCH /* Entry to set the action to execute when versions mismatch */
  , WS_EXPROW_LOCAL         /* Expandable row for local version */
  , WS_EXPROW_OFFICIAL      /* Expandable row for official version */
  , WS_IMG_LOGO             /* Image of watcher */
  , WS_LBL_ERROR            /* Label to print the reason why the add of the watcher was rejected " */
  , WS_MENU_TEMPLATE        /* Menu to use template  */
  , WS_MENU_ITEM_TEMPLATE   /* Menu item to use template  */
  , WS_RAD_VERSION_AUTO     /* Radio button for automatic local version */
  , WS_RAD_VERSION_MANUAL   /* Radio button for manual local version */
  , WS_SPIN_TIMER           /* Spin for timer between 2 resource download */
  , WS_TOGGLE_ERROR         /* Toggle widget to activate trigger on error */
  , WS_TOGGLE_HEADER        /* Toogle widget to activate Http header options */
  , WS_TOGGLE_MISMATCH      /* Toogle widget to activate trigger on mismatch */
  , WS_TOGGLE_OPTIONNAL     /* Toggle widget to set 'optionnal' */
  , WS_TOGGLE_STDERR        /* Toggle widget to use standard error output */
  , WS_VBOX_SEARCH_OFFI     /* Vbox in frame to manage official search blocks */
  , WS_VBOX_SEARCH_LOCAL    /* Vbox in frame to manage local search blocks */

  , WS_WIDGET_COUNT         /* MUST BE THE LAST ELEMENT - Count of widgets in the window */
};

enum PreferenceSettingswidget {
  PS_WIN_MAIN               /* Window of preferences settings */
  , PS_BUT_ADD_TPL          /* Button to show the 'Add template' window */
  , PS_BUT_EDIT_TPL         /* Button to edit a template */
  , PS_BUT_MOVE_UP_TPL      /* Button to move up template */
  , PS_BUT_MOVE_DOWN_TPL    /* Button to move down template */
  , PS_BUT_REMOVE_TPL       /* Button to remove a template */

  , PS_BUT_ADD              /* Button to show the 'add watcher' window */
  , PS_BUT_EDIT             /* Button to edit a watcher */
  , PS_BUT_COLOR            /* Button to choose color of main window background */
  , PS_BUT_MOVE_UP          /* Button to move up watcher */
  , PS_BUT_MOVE_DOWN        /* Button to move down watcher */
  , PS_BUT_REMOVE           /* Button to remove a watcher */


  , PS_BUT_RELOAD           /* Button to reload the configuration file */
  , PS_BUT_SAVE             /* Save button of preference window */
  , PS_CHK_LOCK_POS         /* Check box for lock position */
  , PS_COMBOROW_EXPANSION   /* Combo_row for infosview expansion list */
  , PS_ETY_CACHE            /* Entry to set  the cache folder */
  , PS_HBOX_SYSTRAY         /* Hbox of the systray parameter */
  , PS_HBOX_ALWAYS_ON       /* Hbox of the always_on parameter */
  , PS_RAD_ALWAYS_ON_BOTTOM /* Radio button always on 'bottom' */
  , PS_RAD_ALWAYS_ON_TOP    /* Radio button always on 'top' */
#ifndef LIB_HANDY
  , PS_RAD_INFOS_TO_BOTTOM  /* Radio button 'to bottom' */
  , PS_RAD_INFOS_TO_LEFT    /* Radio button 'to left' */
  , PS_RAD_INFOS_TO_RIGHT   /* Radio button 'to right' */
  , PS_RAD_INFOS_TO_TOP     /* Radio button 'to top' */
#endif
  , PS_ROW_LOCK_POS         /* Row of lock position option */
  , PS_SPIN_MAX_HEIGHT      /* Spin for the maximal height */
  , PS_SPIN_MAX_WIDTH       /* Spin for the maximal width */
  , PS_TOGGLE_SYSTRAY       /* Toggle widget for the systray option */
  , PS_TOGGLE_VEXPAND       /* Toggle Widget for the vertical expand option */
  , PS_SCALE_ICON           /* Scale for icon size */

  , PS_WIDGET_COUNT         /* MUST BE THE LAST ELEMENT - Count of widgets in the window */
};

enum TemplatesSettings_widget {
  TS_WIN_MAIN               /* Window of template settings */
  , TS_ADD_BLOCK            /* Menu to add a new block in template window  */
#ifdef LIB_HANDY
  , TS_BAR_WIN              /* Window's top bar  */
#endif
  , TS_BUT_ADD              /* Button to add a template */
  , TS_BUT_CACHE            /* Button to get cache file */
  , TS_BUT_SAVE             /* Button to save the modifications of a template */
  , TS_ETY_NAME             /* Entry to set template name */
  , TS_ETY_OFFI_HEADER      /* Entry to set the HTTP header options */
  , TS_ETY_RESULT           /* Entry for result */
  , TS_LBL_ERROR            /* Label to print the reason why the add of the template was rejected " */
  , TS_VBOX_OPERATION       /* Vbox in frame to manage search blocks */

  , TS_WIDGET_COUNT         /* MUST BE THE LAST ELEMENT - Count of widgets in the window */
};

/* ******************* DEFINE FUNCTION ********************* */
#define ASSERT_NON_FATAL( cond, str, ... )                              \
  if( cond )                                                            \
    printf( "\e[1;5;38;5;178mBravo !\e[0;1;33m You've found a non critical bug !\e[0m:%s:%d: %s(): " str "\n", __FILE__, __LINE__, __FUNCTION__, ##__VA_ARGS__ );

#define COLOR( no_color, color )  ( app->Flags & APP_FLG_NO_COLORS ? no_color : color )

#define FIXME( str )             LOCK_MUTEX( app->LockPrint ); { static int flg = 0; if( !flg ) { printf( COLOR( "FIXME: %s : %u: %s(): %s\n", "\e[1;38;5;178mFIXME\e[0m: %s : \e[33m%u\e[0m: \e[36m%s\e[0m(): %s\n" ), __FILE__, __LINE__, __FUNCTION__, str ); flg = 1 ; }} UNLOCK_MUTEX( app->LockPrint )

#define VERBOSE( ... )						\
  if( app->Flags & APP_FLG_VERBOSE ) {				\
    time_t timer; char timestamp[18];				\
    struct tm *ltime;						\
    time( &timer );						\
    ltime = localtime( &timer );				\
    strftime( timestamp, 15, "%d-%m %H:%M:%S", ltime );		\
    LOCK_MUTEX( app->LockPrint );				\
    printf( "[%s] ", timestamp );				\
    printf( __VA_ARGS__ );					\
    UNLOCK_MUTEX( app->LockPrint );				\
  }

#if 0
#  define CALLOC(ptr, nb, size)						\
  if((  ptr = calloc( nb, size )) == NULL )                             \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */ \
  printf( "%s : %s : %u : %s = CALLOC( %p )\n", __FILE__, __FUNCTION__, __LINE__, #ptr, ptr );

#  define MALLOC(ptr, size)						\
  if(( ptr = malloc( size )) == NULL )                                  \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */ \
  printf( "%s : %s : %u : %s = MALLOC( %p, %u )\n", __FILE__, __FUNCTION__, __LINE__, #ptr, ptr, (unsigned int)(size) );

#  define REALLOC(ptr, size)						\
  printf( "%s : %s : %u : %s = REALLOC( %p, %u )\n", __FILE__, __FUNCTION__, __LINE__, #ptr, ptr, (unsigned int)(size) ); \
  if(( ptr = realloc( ptr, size )) == NULL )                            \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */ \
  printf( "%s : %s : %u : %s = REALLOC( %p, %u )\n", __FILE__, __FUNCTION__, __LINE__, #ptr, ptr, (unsigned int)(size) );

#  define STRDUP(ptr, str)						\
  if(( ptr = strdup( str )) == NULL )                                   \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */ \
  printf( "%s : %s : %u : %s = STRDUP(%p)\n", __FILE__, __FUNCTION__, __LINE__, #ptr, ptr );

#else
#  define CALLOC(ptr, nb, size)						\
  if((  ptr = calloc( nb, size )) == NULL )                             \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */

#  define MALLOC(ptr, size)						\
  if(( ptr = malloc( size )) == NULL )                                  \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */

#  define REALLOC(ptr, size)						\
  if(( ptr = realloc( ptr, size )) == NULL )                            \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */

#  define STRDUP(ptr, str)                                              \
  if(( ptr = strdup( str )) == NULL )                                   \
    Exit_with_msg( app->Strs[1], 1 ); /* "Couldn't allocate enough memory" */
#endif

#define WARN_APP(str, ...) printf( COLOR("%s: "str"\n", "\e[1m\e[38;5;178m%s\e[0m: "str"\n"), app->Strs[76], ##__VA_ARGS__ ) /* "Warning" */
#define ERR_APP(str, ...) fprintf( stderr, COLOR("%s: "str"\n", "\e[1m\e[31m%s\e[0m: "str"\n"), app->Strs[0], ##__VA_ARGS__)  /* "Error" */
#define ERR_SYS(str, ...) fprintf( stderr, COLOR("%s: "str": %s\n", "\e[1m\e[31m%s\e[0m: "str": %s\n"), app->Strs[0], ##__VA_ARGS__, strerror( errno )) /* "Error" */

/* ********************** STRUCTURE ************************ */
typedef struct         s_Source
{
  const char           *Mem;             /* Pointer to the available memory */
  unsigned int         Length;           /* Length of the available memory */
  unsigned int         Pos;              /* Position in the memory */
}                      s_src;

typedef struct         s_Destination
{
  char                 *Mem;             /* Pointer to the available memory */
  unsigned int         Length;           /* Length of the available memory */
  unsigned int         Pos;              /* Position in the memory */
}                      s_dest;

typedef struct         s_CacheFile
{
  s_src                Source;
  time_t               LastUpdate;       /* Timestamp of the last update of the file */
}                      s_chf;

typedef enum Type_operation {
  OP_TYPE_AFTER_LINE = 0           /* After line containing: |____| */
  , OP_TYPE_AFTER_WORD             /* After the |_| word |_| */
  , OP_TYPE_IN_LINE                /* In line |_| containing |____| */
  , OP_TYPE_IN_SECTION             /* In section |_| delimited with |_| */
  , OP_TYPE_IN_FIELD               /* In field |_| cut with |_| */
  , OP_TYPE_IN_WORD                /* In word |_| */
} e_op;

enum {
  OP_WIDGET_AFTER_LINE_HBOX = 0    /* Hbox of the block */
  , OP_WIDGET_AFTER_LINE_PATTERN   /* Entry of the pattern */

  , OP_WIDGET_AFTER_LINE_COUNT     /* MUST BE THE LAST ELEMENT - Count of widgets in the table */
};

enum {
  OP_WIDGET_AFTER_WORD_HBOX = 0    /* Hbox of the block */
  , OP_WIDGET_AFTER_WORD_PATTERN   /* Entry of the pattern */
  , OP_WIDGET_AFTER_WORD_NB        /* SpinButton of the word number */

  , OP_WIDGET_AFTER_WORD_COUNT     /* MUST BE THE LAST ELEMENT - Count of widgets in the table */
};

enum {
  OP_WIDGET_IN_LINE_HBOX = 0       /* Hbox of the block */
  , OP_WIDGET_IN_LINE_NB           /* SpinButton of the line number */
  , OP_WIDGET_IN_LINE_CHECK        /* CheckButton of the pattern */
  , OP_WIDGET_IN_LINE_PATTERN      /* Entry of the pattern */

  , OP_WIDGET_IN_LINE_COUNT        /* MUST BE THE LAST ELEMENT - Count of widgets in the table */
};

enum {
  OP_WIDGET_IN_SECTION_HBOX = 0    /* Hbox of the block */
  , OP_WIDGET_IN_SECTION_NB        /* SpinButton of the line number */
  , OP_WIDGET_IN_SECTION_CHARS     /* Entry containing the 2 opening and closing characters of the section */

  , OP_WIDGET_IN_SECTION_COUNT     /* MUST BE THE LAST ELEMENT - Count of widgets in the table */
};

enum {
  OP_WIDGET_IN_FIELD_HBOX = 0      /* Hbox of the block */
  , OP_WIDGET_IN_FIELD_NB          /* SpinButton of the field number */
  , OP_WIDGET_IN_FIELD_CHAR        /* Entry containing the characters splitting the fields */

  , OP_WIDGET_IN_FIELD_COUNT       /* MUST BE THE LAST ELEMENT - Count of widgets in the table */
};

enum {
  OP_WIDGET_IN_WORD_HBOX = 0       /* Hbox of the block */
  , OP_WIDGET_IN_WORD_NB           /* SpinButton of the word number */
  , OP_WIDGET_IN_WORD_CHECK        /* CheckButton of the 'cut with' part */

  , OP_WIDGET_IN_WORD_COUNT        /* MUST BE THE LAST ELEMENT - Count of widgets in the table */
};

typedef struct        s_Operation
{
  GtkWidget           **GtkTable;        /* Table holding GTK widgets */
  const char          *Pattern;          /* Searched pattern */
  e_op                Type;              /* Type of operation, see OP_TYPE_ */
  unsigned short      Number;            /* Holding a number */
  unsigned short      Position;          /* Holding a position */
  char                Delimiter[2];      /* Holding delimitors */
  _Bool               Activated;         /* Holding a boolean */
  _Bool               Reorder;           /* This operation was set with an obsolete key_name */
  _Bool               Temporary;         /* This operation was temporary allocated and not assigned to a watcher */
  _Bool               Template;          /* Operation is made for template */
}                     s_op;

#define CB_OFFICIAL_VERSION       (1 << 0)    /* Concerns the official version block */
#define CB_TEMPLATE_CREATE        (1 << 1)    /* Concerns the template create block */
#define CB_LOCAL_VERSION          (1 << 2)    /* Concerns the local version block */

#define CB_BLOCK_AFTER_LINE       (1 << 8)    /* Create 'After line' block */
#define CB_BLOCK_AFTER_WORD       (1 << 9)    /* Create 'After word' block */
#define CB_BLOCK_IN_FIELD         (1 << 10)   /* Create 'In field' block */
#define CB_BLOCK_IN_LINE          (1 << 11)   /* Create 'In line' block */
#define CB_BLOCK_IN_SECTION       (1 << 12)   /* Create 'In section' block */
#define CB_BLOCK_IN_WORD          (1 << 13)   /* Create 'In word' block */

typedef struct         s_Template
{
  GtkWidget            *MainBox;         /* Main box of the block */
  GtkWidget            *MenuItem;        /* MenuItem ot the template */
  const char           *Name;            /* Name of template */
  const char           *HeaderOptions;   /* Options for the header request */
  s_op                 **OperationTable; /* Operations table */
  GtkTreeIter          Iter;             /* GtkTreeIter used in preference templates list */
  unsigned char        OperationCount;   /* Count of operation in the table */
}                      s_tpl;

#define FILE_FLG_SEARCH_OFFICIAL  (1 << 0) /* The last key name was about official search */
typedef struct         s_File
{
  /* Mutex protection START : LockFile */
  pthread_mutex_t      LockFile;         /* Mutex protecting following variables */
  char                 *Name;            /* Filename */
  char                 *Mem;             /* Memory containing the entire file content */
  char                 **Error;          /* Table of errors at loading */
  s_tpl                **TemplateTable;  /* Template table */
  s_dest               SectionTarget;    /* Buffer pointing to the target section in s_file.Mem */
  s_dest               SectionReplace;   /* Buffer replacing data pointed by s_file.SectionTarget */
  unsigned int         Line;             /* Current line of parsed file */
  unsigned int         Length;           /* Length of memory available in .Mem */
  unsigned char        ErrorsCount;      /* Count of errors available in .Error */
  unsigned char        TemplateCount;    /* Count of templates in s_file.TemplateTable */
  uint8_t              Flags;            /* Saving some parameters; see FILE_FLG_ */
  /* Mutex protection END   : LockFile */
}                      s_file;

#define WTR_FLG_ATTACHED                     (1 << 0) /* The watcher is attached to the main table */
#define WTR_FLG_DISABLE                      (1 << 1) /* The watcher has been disabled */
#define WTR_FLG_HTTP                         (1 << 2) /* The web resource will be retrieve with HTPP instead of HTTPS */
#define WTR_FLG_NO_AUTO_UPDATE               (1 << 3) /* Something is wrong with official version cache, no automatic update will be performed */
#define WTR_FLG_OPTIONNAL                    (1 << 4) /* The watcher is optionnal, if LocalExec file does not exist, watcher is disabled */
#define WTR_FLG_STDERR                       (1 << 5) /* The version must be read on stderr instead of stdout */
#define WTR_FLG_INIT                         (1 << 6) /* The watcher is in init step */
#define WTR_FLG_VERSION_MISMATCH             (1 << 7) /* Local version mismatch official version */

#define WTR_FLG_LOCAL_AUTO                   (1 << 8) /* Local version is set automatically */
#define WTR_FLG_LOCAL_ERROR                  (1 << 9) /* An error has been detected when retrieving local or official version */
#define WTR_FLG_LOCAL_IN_PROGRESS            (1 << 10) /* Local version is being retrieved */
#define WTR_FLG_LOCAL_REFRESH                (1 << 11) /* The local version or operations to search it have changed */
#define WTR_FLG_AREA_REFRESH                 (1 << 12) /* The draw area needs a refresh */
#define WTR_FLG_OFFICIAL_HEADER              (1 << 13) /* Header option si enable */
#define WTR_FLG_OFFICIAL_REFRESH             (1 << 14) /* The operations to search the official version have changed */

#define WTR_FLG_OFFICIAL_UPDATE              (1 << 16) /* The user requested a resource update, or changed the url */
#define WTR_FLG_OFFICIAL_ERROR               (1 << 17) /* An error has been detected when retrieving local or official version */
#define WTR_FLG_OFFICIAL_IN_PROGRESS         (1 << 18) /* Official version is being retrieved */
#define WTR_FLG_TRIGGER_MISMATCH             (1 << 19) /* Trigger on mismatch is enable */
#define WTR_FLG_TRIGGER_ERROR                (1 << 20) /* Trigger on error is enable */
typedef struct        s_Watcher
{
  /* core part */
  char                *Name;                 /* Name of watched program */
  char                *LocalArg;             /* Argument for executable */
  char                *LocalExec;            /* Executable file */
  char                *LocalVersion;         /* Retrieved local version */

  s_op                **LocalOperationTable; /* Operations table to search local version */
  unsigned char       LocalOperationCount;   /* Count of operation in the table for local version */

  unsigned char       OfficialOperationCount;   /* Count of operation in the table */
  s_op                **OfficialOperationTable; /* Operations table to search official version */

  char                *OfficialURL;          /* URL to the page containing the official version */
  char                *OfficialHeader;       /* List of option in HTTP header */
  char                *OfficialVersion;      /* Retrieved official version */

  s_chf               CacheFile;             /* Last retrieved/mapped  informations about the cache file */

  char                *TemplateName;         /* Name of template used */
  char                *ImgPath;              /* Path to the icon file */

  char                *Trigger_on_error;     /* Table containing data about the trigger when an error is detected */
  char                *Trigger_on_mismatch;  /* Table containing data about the trigger when a mismatch is detected */
  char                Trigger_xor[SIZE_BUFFER_XOR]; /* XOR between versions to preserve trigger to be executed more than once */

  unsigned char       NameLength;            /* Length of the watcher name */
  unsigned int        OfficialTimer;         /* Delay between 2 checks of the official version, in minutes */
  unsigned int        Pos;                   /* Position into app->WatcherTable */
  unsigned int        SavePos;               /* Position saved into file */
  uint32_t            Flags;                 /* Saving some parameters; see WTR_FLG_ */

  /* gui part */
  GtkWidget           *BoxNextUpdate;              /* Box containing 'Next update' infos  */
  GtkWidget           *DrawArea;                   /* Drawable area */
  GdkPixbuf           *OriginPixbuf;               /* Origin pixbuf from loaded picture file */
  GdkPixbuf           *CurrentPixbuf;              /* Current pixbuf used in GUI */
  //#todo opti on infoview uniq instead of per watcher
  GtkWidget           *InfosWindow;                /* Informations window */
  GtkWidget           *LabelFrame;                 /* Label for the frame of the infos window */
  GtkWidget           *LabelLocalVersion;          /* Label "Local version: " */
  GtkWidget           *LabelLocalVersionResult;    /* Label widget to show local version */
  GtkWidget           *LabelOfficialVersion;       /* Label "Official version: " */
  GtkWidget           *LabelOfficialVersionResult; /* Label widget to show official version */
  GtkWidget           *LabelState;                 /* Label "Up-to-date " */
  GtkWidget           *LabelNextUpdate;            /* Label "Next update: " */
  GtkTreeIter         Iter;                        /* GtkTreeIter used in preference / watcher / list */
  unsigned int        InfoSizeX, InfoSizeY;        /* Dimensions of s_wtr.InfosWindow */
  unsigned int        VisiblePos;                  /* Position into the main window */
}                     s_wtr;

typedef struct        s_SearchOperation
{
  GtkEntry            *Result;           /* Target result entry */
  s_wtr               *Watcher;          /* Pointer to updated watcher */
  char                *Url;              /* URL of the resource if a download is required */
  const char          *Header;           /* HTTP Header options */
  char                *CacheFile;        /* Target cache file path/name if a download is required */
  char                *TemplateName;     /* Set if a template is used */
  s_op                **OperationTable;  /* Operations table to search version */
  unsigned char       OperationCount;    /* Count of operation in the table */
}                     s_so;

typedef struct         s_WatcherKeyName
{
  const char           *Name;                   /* Name of key */
  void                 (*Func)(s_file*, s_src*, s_wtr*, s_src*);  /* Function to modify a watcher with a parameter */
  unsigned short       Length;                  /* Length of s_wkn.Name */
}                      s_wkn;

typedef struct         s_TemplateKeyName
{
  const char           *Name;                   /* Name of key */
  void                 (*Func)(s_file*, s_src*, s_src*, s_tpl*);  /* Function to modify a template with a parameter */
  unsigned short       Length;                  /* Length of s_tkn.Name */
}                      s_tkn;

typedef struct         s_PreferencesKeyName
{
  const char           *Name;                   /* Name of key */
  void                 (*Func)(s_file*, s_src*, s_src*);  /* Function to modify a setting */
  unsigned short       Length;                  /* Length of s_pkn.Name */
}                      s_pkn;

#define GUI_CANCEL_FLG_V_EXPAND              (1 << 0) /* Check box 'Vertical expand' has been modified */
#define GUI_CANCEL_FLG_INFOS_VIEW            (1 << 3) /* Radio 'Infos view expension' has been modified */
#define GUI_CANCEL_FLG_LOCK_MAIN             (1 << 4) /* Check box 'Lock position' */
#define GUI_CANCEL_FLG_SYSTRAY               (1 << 5) /* Check box 'Show systray' */
#define GUI_CANCEL_FLG_ALWAYS                (1 << 6) /* Radio 'Always on' */
#define GUI_CANCEL_FLG_MAIN_COLOR            (1 << 7) /* ColorButton for main background has been modified */

#define GUI_CANCEL_FLG_CACHE                 (1 << 8) /* Entry 'cache_folder' */
#define GUI_CANCEL_FLG_WATCHER_ORDER         (1 << 9) /* Order of watchers has been modified */
#define GUI_CANCEL_FLG_PREFERENCES           (1 << 10) /* User has modified a parameter impacting preferences */

#define GUI_CANCEL_INFOS_TO_BOTTOM           0 /* Saved 'Infos view expansion'; see ccl.InfosView */
#define GUI_CANCEL_INFOS_TO_TOP              1 /* Infos window will be shown to the left of the watcher icon; see ccl.InfosView */
#define GUI_CANCEL_INFOS_TO_LEFT             2 /* Infos window will be shown to the left of the watcher icon; see ccl.InfosView */
#define GUI_CANCEL_INFOS_TO_RIGHT            3 /* Infos window will be shown to the right of the watcher icon; see ccl.InfosView */
typedef struct        s_CancelPref
{
  unsigned int        MaxXWatcher;           /* Saving MaxXWatcher */
  unsigned int        MaxYWatcher;           /* Saving MaxYWatcher */
  unsigned int        SizeIcon;              /* Saving the icon size */
#if GTK_MAJOR_VERSION >= 3
  GdkRGBA             MainColor;             /* Saving color of main background */
#else
  float               MainOpacity;           /* Saving opacity of main background */
  GdkColor            MainColor;             /* Saving color of main background */
#endif
  unsigned char       InfosView;             /* Saving Flags for cancelling; see GUI_CANCEL_INFOS_ */
  uint16_t            Flags;                 /* Saving which option was modified; see GUI_CANCEL_FLG_ */
  _Bool               vexpand;               /* Saving vexpand state */
  _Bool               LockPosition;          /* Saving lock position state */
  _Bool               Systray;               /* Saving systray state */
  _Bool               Always;                /* Saving 'Always on' */
}                     s_clp;

#define SAVE_FLG_WATCHER_ORDER               (1 << 0) /* Order of watchers has changed */
typedef struct        s_Save
{
  uint16_t            Flags;                 /* Saving states, see SAVE_FLG_ */
}                     s_save;

#define GUI_FLG_LOCK_MAIN          (1 << 0) /* Watchers window is locked */
#define GUI_FLG_ON_TOP             (1 << 1) /* Watchers windows is always on top, else always on bottom */
#define GUI_FLG_VERTICAL_EXPAND    (1 << 2) /* Watchers list is expanding vertically then horizontally */
#define GUI_FLG_MOVING             (1 << 3) /* Watchers window is being moved */
#define GUI_FLG_INFOS_TO_TOP       (1 << 4) /* Infos window will be shown to the top of the watcher icon */
#define GUI_FLG_INFOS_TO_LEFT      (1 << 5) /* Infos window will be shown to the left of the watcher icon */
#define GUI_FLG_INFOS_TO_RIGHT     (1 << 6) /* Infos window will be shown to the right of the watcher icon */
#define GUI_FLG_INFOS_MASK         (GUI_FLG_INFOS_TO_TOP | GUI_FLG_INFOS_TO_LEFT | GUI_FLG_INFOS_TO_RIGHT) /* Infos window mask */
#define GUI_FLG_NO_WATCHER         (1 << 7) /* No watcher is visible on the watchers window */

#define GUI_FLG_FORCE_MOVE         (1 << 8) /* Is trying to put the window back to its position */
#define GUI_FLG_MOVE_FORCED        (1 << 9) /* The step to force the window position is finish */
#define GUI_FLG_FIRST_INIT         (1 << 10) /* The step to force the window position is finish */
#define GUI_FLG_WATCHER_MODIFY     (1 << 11) /* A watcher is being modified by user, the watcher settings window is opened */
#define GUI_FLG_TEMPLATE_MODIFY    (1 << 12) /* A template is being modified by user, the template settings window is opened */
#define GUI_FLG_NEW_WATCHER        (1 << 13) /* A new watcher is being configured */
typedef struct        s_GraphicUserInterface
{
#if GTK_MAJOR_VERSION >= 3
  GtkApplication      *App;                  /* GtkApplication */
#endif
  GtkWidget           *Main[MW_WIDGET_COUNT]; /* Widget table of the main window */
  GtkWidget           *PreferenceSettings[PS_WIDGET_COUNT]; /* Widget table of the settings window */
  GtkWidget           *TemplateSettings[TS_WIDGET_COUNT];   /* Widget table of the template settings window */
  GtkWidget           *WatcherSettings[WS_WIDGET_COUNT];    /* Widget table of the window to add/modify a watcher */

  GtkWidget           *ChooseFile;           /* Dialog window to choose a single file */

  GtkListStore        *ListWatchers;         /* List of watchers */
  GtkTreeSelection    *SelectionWatcher;     /* Selection for watchers list */
  s_wtr               *HoveredWatcher;       /* Current watcher set/unset when enter/leave event on the DrawArea of a watcher */
  s_wtr               *SelectedWatcher;      /* Current watcher set when right click occured on the DrawArea of a watcher */

  GtkListStore        *ListTemplates;        /* List of templates */
  GtkTreeSelection    *SelectionTemplate;    /* Selection for templates list */

  cairo_t             *MainCairo;            /* Cairo context to draw main window */

  char                **ImgFile;             /* Table of absolute or relative path of the image used for the watcher */
  GdkPixbuf           **ImgTable;            /* Table of images */
  unsigned int        ImgCount;              /* Count of elements in s_gui.ImgTable and s_gui.ImgFile */

  s_save              Save;                  /* Keeping preferences changes, to be able to cancel them */
  s_clp               CancelPref;            /* Keeping preferences changes, to be able to cancel them */
  s_tpl               TmpTemplate;           /* Temporary template; used when user is adding/modifying a template */
  s_wtr               TmpWatcher;            /* Temporary watcher; used when user is adding/modifying a watcher */
  s_file              File;                  /* File corresponding to gui */

  /* Mutex protection START : LockWatcher */
  pthread_mutex_t      LockWatchers;         /* Mutex protecting following variables */
  s_wtr                **WatcherTable;       /* Table of watcher structures */
  unsigned int         WatcherTableSize;     /* Number of watchers s_app.WatcherTable can hold */
  unsigned int         WatcherCount;         /* Count of watchers in s_app.WatcherTable */
  /* Mutex protection END   : LockWatchers */

  struct timeval      ClickTimestamp;        /* Last timestamp of button 1 click on watchers window */
  int                 MainX, MainY;          /* Position of watchers window */
  int                 SaveX, SaveY;          /* Position of watchers window last time it was saved */
  int                 MoveX, MoveY;          /* Position of cursor in window; used only when moving */
  unsigned int        MaxXWatcher;           /* Max count of watchers per lines */
  unsigned int        MaxYWatcher;           /* Max count of watchers per columns */

#if GTK_MAJOR_VERSION >= 3
  GdkRGBA             MainColor;             /* Color of the watchers window */
#else
  float               MainOpacity;           /* Opacity value of the watchers window */
  GdkColor            MainColor;             /* Color of the watchers window */
#endif

  unsigned char       WatcherSizeX;          /* Width of watcher icon */
  unsigned char       WatcherSizeY;          /* Height of watcher icon */
  uint16_t            Flags;                 /* Saving some parameters; see GUI_FLG_ */
}                     s_gui;

#define THD_FLG_RELOAD           (1 << 0) /* Thread needs to takes actions after a reload */

#define APP_FLG_NO_COLORS        (1 << 0) /* Terminal colors are disabled */
#define APP_FLG_BLINK            (1 << 1) /* Systray is blinking with watcher icon */
#define APP_FLG_VERBOSE          (1 << 3) /* Verbose mode has been activated */
#define APP_FLG_INIT             (1 << 4) /* Application is being initialized */
#define APP_FLG_NO_ALPHA         (1 << 5) /* Alpha channel is not supported */
#define APP_FLG_NO_COMPOSITE     (1 << 6) /* Window composite not available */
#define APP_FLG_SYSTRAY          (1 << 7) /* Systray is enable */
#define APP_FLG_PHONE            (1 << 8) /* The current device is a phone */

#define APP_FLG_FORCE_PHONE      (1 << 9) /* The parameter 'phone' was in the config file */
typedef struct         s_Application
{
  const char           **Strs;           /* Containing all printable string of the software */
  const char           *UserHome;        /* Containing the current user home path */
  char                 *LastFolder;      /* Last folder used by user; can be null */

  /* Mutex protection START : LockCache */
  pthread_mutex_t      LockCache;        /* Mutex protecting following variables */
  char                 *CacheFolder;     /* folder where web resources will be stored */
  /* Mutex protection END   : LockCache */

  pthread_mutex_t      LockPrint;        /* Mutex to protect printing on stdout / stderr */

  GtkCellRenderer      *CellRender[COL_COUNT];   /* CellRender for the columns of the watchers list */
  GtkWidget            *About[ABW_WIDGET_COUNT]; /* Widget table of the about window */
  GtkWidget            *Error[ERW_WIDGET_COUNT]; /* Widget table of the error window */
  GtkStatusIcon        *Systray;         /* Systray icon */
  s_gui                Gui;              /* GUI structure */

  char                 ConfigSeparator;  /* Separator character into config file; 0 means not found */

  pthread_attr_t       Attr;             /* Threads attributs */
  pthread_cond_t       TimerCond;        /* Thread condition */

  /* Mutex protection START : LockCond */
  pthread_mutex_t      LockCond;         /* Mutex to lock condition and protecting following variables */
  s_wtr                **ReloadTable;    /* Table containing old watchers */
  unsigned char        *ThreadFlags;     /* Table containing the thread flags; see THD_FLG_ */
  unsigned int         CountReload;      /* Count of watchers in s_app.ReloadTable */
  /* Mutex protection END   : LockCond */

  pthread_t            *IdWorkers;       /* Table of thread ID */
  int                  CoreCount;        /* Number of available CPU cores */

  unsigned int         UserHomeLength;   /* Length of s_app.UserHome */
  uint16_t             Flags;            /* Saving some parameters, see APP_FLG_ */

#if TRACE
  unsigned int         TraceLevel;
#endif /* TRACE */
}                      s_app;

/* *********************** EXTERN ************************** */
extern s_app           * restrict app;
extern const char      *English[];

/* ********************* DECLARATION *********************** */
_Bool               Add_error(s_file *file, char *error);
_Bool               Add_local_operation_to_watcher(s_wtr *watcher, s_op *operation);
_Bool               Add_official_operation_to_watcher(s_wtr *watcher, s_op *operation);
_Bool               Add_operation_to_template(s_tpl *template, s_op *operation);
void                Add_template_to_file(s_file *file, s_tpl *template);
void                Add_template_to_menu(GtkWidget *menu, s_tpl *template);
void                Add_watcher_to_gui(s_gui *gui, s_wtr *watcher);
_Bool               Apply_search_operation(s_so *search, s_src *source, char *error);
char                *Build_cachefile_name(const char *file_name, unsigned int file_len);
void                Build_xor(const s_wtr *watcher, char *xor);
void                Callback_check_to_sensitive(GtkWidget *button, GtkWidget *checkbox);
#if GTK_MAJOR_VERSION < 3
gboolean            Callback_draw_main(GtkWidget *widget, GdkEvent *event, gpointer data);
#else
gboolean            Callback_draw_main(GtkWidget *widget, cairo_t *cr, gpointer data);
#endif
void                Callback_edit_watcher(GtkWidget *unused1, s_gui *gui);
gboolean            Callback_on_delete(GtkWidget *widget, gpointer window);
void                Callback_on_destroy(GtkWidget *widget, gpointer window);
gboolean            Callback_on_expose_watcher(GtkWidget *widget, GdkEvent *event, gpointer data);
void                Callback_show_preferences(GtkWidget *widget, s_gui *gui);
_Bool               Check_file(s_file *file, const char *file_name, _Bool warn);
unsigned int        Convert_source_to_boolean(s_file *file, s_src *source, _Bool *ret_bool);
unsigned int        Convert_source_to_number(s_file *file, s_src *source, unsigned int *ret_nb);
_Bool               Copy_file(const char *source, const char *dest);
void                Copy_gui_after_line(s_op *operation);
void                Copy_gui_after_word(s_op *operation);
void                Copy_gui_in_field(s_op *operation);
void                Copy_gui_in_line(s_op *operation);
void                Copy_gui_in_section(s_op *operation);
void                Copy_gui_in_word(s_op *operation);
void                Create_del_button(GtkWidget *container);
void                Create_menu_search_block(GtkWidget *vbox_main, uint8_t block);
void                Create_menu_watcher();
void                Create_new_gui_watcher(s_wtr *item);
void                Create_preferences_tab_templates(GtkWidget *tab);
void                Create_preferences_tab_watchers(GtkWidget *tab);
void                Create_systray();
void                Create_watchers();
void                Create_window_about();
void                Create_window_error();
void                Create_window_main();
void                Create_window_preferences();
void                Create_window_template_settings();
void                Create_window_watcher_settings();
void                Del_official_operation_from_watcher(s_wtr *watcher, GtkWidget *hbox);
_Bool               Domain_is_valid(const char *domain);
unsigned int        Download_resource_to_cache(char *url, const char *header, const char *dest_file);
void                Empty_search_block_list(s_gui *gui);
char                *Error_to_xml(const char *str);
void                Exit_with_msg(const char *msg, char flg);
void                Free_struct_operation(s_op *operation, _Bool lock_gtk);
void                Free_struct_template(s_tpl *template, _Bool lock_gtk);
void                Free_struct_search(s_so *search, _Bool lock_gtk);
void                Free_struct_watcher(s_wtr *watcher, _Bool free_ptr, _Bool lock_gtk);
_Bool               Generate_preferences_section(s_gui *gui, s_dest *dest);
_Bool               Generate_template_section(s_tpl *template, s_dest *dest);
_Bool               Generate_watcher_section(s_wtr *watcher, s_dest *dest);
s_tpl               *Get_current_selected_template(s_gui *gui);
s_wtr               *Get_current_selected_watcher(s_gui *gui);
char                *Get_default_cache_folder_name();
_Bool               Get_preferences(s_file *file, s_src *source);
unsigned int        Get_next_value(s_file *file, s_src *source, s_src *value);
const s_wkn         *Get_obsolete_watchers_properties(s_file *file, s_src *source, s_wtr *watcher);
const s_pkn         *Get_obsolete_preferences_properties(s_file *file, s_src *source);
void                Get_options(int ac, char **av);
_Bool               Get_watcher_settings(s_file *file, s_src *source, char *name, unsigned char length);
unsigned char       Get_gui_after_line(s_op *operation, GtkWidget *lbl_error);
unsigned char       Get_gui_after_word(s_op *operation, GtkWidget *lbl_error);
unsigned char       Get_gui_in_field(s_op *operation, GtkWidget *lbl_error);
unsigned char       Get_gui_in_line(s_op *operation, GtkWidget *lbl_error);
unsigned char       Get_gui_in_section(s_op *operation, GtkWidget *lbl_error);
unsigned char       Get_gui_in_word(s_op *operation, GtkWidget *lbl_error);
_Bool               Get_template(s_file *file, s_src *source, unsigned int start, unsigned int length);
unsigned int        Goto_next_section(s_src *source);
void                key_always_on(s_file *file, s_src *source, s_src *value);
void                Load_file_config(s_file *file, char *name);
void                Load_file_lang(const char *file);
void                Modify_local_version(s_wtr *watcher);
s_op                *New_operation_after_line(const char *pattern);
s_op                *New_operation_after_word(unsigned int nb_word, const char *pattern);
s_op                *New_operation_in_line(unsigned int nb_line, _Bool containing, const char *pattern);
s_op                *New_operation_in_section(unsigned int nb_field, char open_field, char close_field);
s_op                *New_operation_in_field(unsigned int nb_section, char split);
s_op                *New_operation_in_word(unsigned int nb_word, _Bool from_end);
s_tpl               *New_template(const char *name);
_Bool               Normalize_version(s_src *source, char *buffer);
unsigned int        Open_file(const char *file, s_chf *dest);
s_so                *Operations_to_search(s_op **table, unsigned char count);
void                Pop_error(s_file *file, const char *section_name, const char *key_name, const char *err);
void                Quit_app();
void                Refresh_official_version(s_wtr *watcher);
void                Remove_section_from_file(const char *name);
void                Retrieve_local_version(s_wtr *watcher);
void                Retrieve_official_version_from_cache(s_wtr *watcher, const char *file, _Bool lock_gui);
_Bool               Search_after_line(s_src *source, const char *pattern);
unsigned int        Search_after_word(s_src *source, unsigned int target_pattern, const char *pattern);
unsigned int        Search_in_field(s_src *source, unsigned int target_field, char split);
unsigned int        Search_in_line(s_src *source, unsigned int target_line, const char *pattern);
unsigned int        Search_in_section(s_src *source, unsigned int target_section, char section_open, char section_close);
unsigned int        Search_in_word(s_src *source, unsigned int target_word, _Bool from_end);
_Bool               Search_section2(s_file *file, s_dest *target, const char *section);
void                Set_icon_to_button(GtkWidget *button, GdkPixbuf *pix);
void                Set_window_properties(GtkWidget *win, GtkWidget *main_box);
void                Show_window(GtkWidget *widget, gpointer data);
void                Systray_stop_blink();
void                *th_Update_from_search(void *update);
void                *th_Retrieve_local_version(void *param);
void                *th_Refresh_official_version(void *param);
void                *th_Retrieve_both_versions(void *param);
void                *th_Save_modifications_to_file(void *param);
void                *th_Updater(void *param);
_Bool               Try_create_folder(char *folder, unsigned int length);
void                Update_in_progress(s_wtr *watcher, GtkLabel *label, _Bool from_thread);
void                Update_info_position(s_wtr *item);
void                Update_main_geometry();
void                Update_official_result();
void                Update_template_result();
void                Update_timer(s_wtr *watcher, time_t now, _Bool from_thread);
void                Update_watcher_area(s_wtr *watcher, _Bool lock_gtk);
void                Update_watcher_infosview(s_wtr *watcher);
_Bool               Verify_gui_in_word(s_op *operation, GtkWidget *lbl_error);

#endif /* INC_APP_H */
