/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
   Purpose: Resetting file structure
*/
static void             Reset_struct_file(s_file *file)
{
  unsigned short        a;

  if( file->ErrorsCount ) {
    for( a = 0; a < file->ErrorsCount; ++a )
      free( file->Error[a] );
    free( file->Error );
    file->ErrorsCount = 0;
  }

  file->Line = 0;
}

/*
  Purpose: Adding an error to the errors table of the file
  - 'error' must be allocated
*/
_Bool                    Add_error(s_file *file, char *error)
{
  char                   **tab;

  ASSERT( file->ErrorsCount >= MAX_COUNT_FILE_ERROR, "file->ErrorsCount >= MAX_COUNT_FILE_ERROR" );
  if(( tab = realloc( file->Error, ( file->ErrorsCount + 1 ) * sizeof(*file->Error) )) == NULL ) {
    free( error );
    return( 1 );
  }
  tab[file->ErrorsCount++] = error;
  file->Error = tab;
  return( 0 );
}

/*
  Purpose: Printing on terminal errors, and buffering them for the error window
*/
void                     Pop_error(s_file *file, const char *watcher_name, const char *key_name, const char *err)
{
  char                   *tmp_err, tmp[192];
  int                    length;

  if( app->Flags & APP_FLG_NO_COLORS ) {
    ERR_APP( "%s %u: [%s]: %s: %s", app->Strs[72], file->Line, /* "Line" */
	     watcher_name, key_name, err );
  }
  else {
    ERR_APP( "%s \e[33m%u\e[0m: [\e[36m%s\e[0m]: \e[1;35m%s\e[0m: %s", app->Strs[72],  /* "Line" */
	     file->Line, watcher_name, key_name, err );
  }
  if( file->ErrorsCount < MAX_COUNT_FILE_ERROR ) {
    length = snprintf( tmp, 192, "%s %u: [%s]: %s: %s", app->Strs[72], /* "Line" */
	     file->Line, watcher_name, key_name, err );
    if( length > 0 ) {
      if(( tmp_err = strdup( tmp )) == NULL ) {
	ERR_SYS( "%s", app->Strs[1] ); /* "Couldn't allocate enough memory" */
	return;
      }
      Add_error( file, tmp_err );
    }
  }
}

/*
  Purpose: Handeling a simple error
*/
static void              Handle_simple_error(s_file *file, const char *error)
{
  if( app->Flags & APP_FLG_NO_COLORS ) {
    ERR_APP( "%s %u: %s", app->Strs[72], file->Line, error ) ; /* "Line" */
  }
  else {
    ERR_APP( "%s \e[33m%u\e[0m: %s", app->Strs[72], file->Line, error ) ; /* "Line" */
  }
  if( file->ErrorsCount < MAX_COUNT_FILE_ERROR ) {
    char *err, tmp[100];
    int length;

    length = snprintf( tmp, 100, "%s %u: %s", app->Strs[72], file->Line, error ); /* "Line" */
    if( length > 0 ) {
      if(( err = strdup( tmp )) == NULL ) {
	ERR_SYS( "%s", app->Strs[1] ); /* "Couldn't allocate enough memory" */
	return;
      }
      Add_error( file, err );
    }
  }
}

/*
  Purpose: Adding a official search operation to the watchers
*/
_Bool                     Add_official_operation_to_watcher(s_wtr *watcher, s_op *operation)
{
  if( watcher->OfficialOperationCount == MAX_COUNT_OPERATION ) {
    Free_struct_operation( operation, 0 );
    ERR_APP( "%s:%s%s: %u", watcher->Name, app->Strs[13], /* " Official version " */
             app->Strs[235], MAX_COUNT_OPERATION ); /* "Too much search operation on watcher" */
    return( 1 );
  }
  REALLOC( watcher->OfficialOperationTable, ( watcher->OfficialOperationCount + 1 ) * sizeof(*watcher->OfficialOperationTable) );
  watcher->OfficialOperationTable[watcher->OfficialOperationCount++] = operation;
  return( 0 );
}

/*
  Purpose: Adding a local search operation to the watchers
*/
_Bool                     Add_local_operation_to_watcher(s_wtr *watcher, s_op *operation)
{
  if( watcher->LocalOperationCount == MAX_COUNT_OPERATION ) {
    Free_struct_operation( operation, 0 );
    ERR_APP( "%s: %s: %s: %u", watcher->Name, app->Strs[236], /* "Local version" */
             app->Strs[235], MAX_COUNT_OPERATION ); /* "Too much search operation on watcher" */
    return( 1 );
  }
  REALLOC( watcher->LocalOperationTable, ( watcher->LocalOperationCount + 1 ) * sizeof(*watcher->LocalOperationTable) );
  watcher->LocalOperationTable[watcher->LocalOperationCount++] = operation;
  return( 0 );
}

/*
  Purpose: Cheking the validity of the domain name
  - Validation is mostly following RFC 1034-35
*/
_Bool                    Domain_is_valid(const char *domain)
{
  unsigned int           a, dom_length, label_length;
  char                   c;
  _Bool                  is_number;

  ASSERT( !domain, "domain = 0" );
  dom_length = strlen( domain );

  if( dom_length >= 7 && ! strncmp( "http://", domain, 7 )) {
    domain = &domain[7];
    dom_length -= 7;
  }
  else if( dom_length >= 8 && ! strncmp( "https://", domain, 8 )) {
    domain = &domain[8];
    dom_length -= 8;
  }

  if( dom_length < 3 )
    return( 1 );

  label_length = 0;
  is_number = 0;
  for( a = 0; a < dom_length; ++a ) {
    c = domain[a];
    if( c == '.' ) {
      if( ! label_length ) /* Only root label can have a length of 0  */
        return( 1 );
      label_length = 0;
    }
    else if( c == '/' ) {
      if( ! a )
        return( 1 );
      break;
    }
    else {
      if( ! label_length ) { /* is the first character of the label a number ? */
        if( c > '9' || c < '0' )
          is_number = 0;
        else
          is_number = 1;
      }
      if( !(( c >= 'a' && c <= 'z') || ( c >= 'A' && c <= 'Z' )
            || ( c >= '0' && c <= '9' ) || c == '-' ))
        return( 1 );
      ++label_length;
      if( label_length > 63 ) /* A label can't be longer than 63 octets */
        return( 1 );
    }
  }
  if( is_number ) /* First character of last label must not be number */
    return( 1 );

  if( dom_length < 3 || dom_length > 254 ) /* 254 + '\0' = 255 */
    return( 1 );

  return( 0 );
}


/*
  Purpose: Checking the validity of the file
*/
_Bool                   Check_file(s_file *file, const char *file_name, _Bool warn)
{
  struct stat           st;

  if( stat( file_name, &st )) {
    if( warn ) {
      if( app->Flags & APP_FLG_NO_COLORS ) {
	ERR_SYS( "%s %u: '%s': %s", app->Strs[72], file->Line, file_name, app->Strs[2] ); /* "Line" "Unable to open file" */
      }
      else {
	ERR_SYS( "%s \e[33m%u\e[0m: '%s': %s", app->Strs[72], file->Line, file_name, app->Strs[2] ); /* "Line" "Unable to open file" */
      }
      if( file->ErrorsCount < MAX_COUNT_FILE_ERROR ) {
	char *err, tmp[100];
	int length;

	length = snprintf( tmp, 100, "%s %u: '%s': %s", app->Strs[72], file->Line, file_name, app->Strs[2] ); /* "Line" "Unable to open file" */
	if( length > 0 ) {
	  if(( err = strdup( tmp )) == NULL ) {
	    ERR_SYS( "%s", app->Strs[1] ); /* "Couldn't allocate enough memory" */
	    return( 1 );
	  }
	  Add_error( file, err );
	}
      }
    }
    return( 1 );
  }
  if( ! st.st_size && warn ) {
    if( app->Flags & APP_FLG_NO_COLORS ) {
      ERR_APP( "%s %u: '%s': %s", app->Strs[72], file->Line, file_name, app->Strs[69] ); /* "Line" "File is empty" */
    }
    else {
      ERR_APP( "%s \e[33m%u\e[0m: '%s': %s", app->Strs[72], file->Line, file_name, app->Strs[69] ); /* "Line" "File is empty" */
    }
    if( file->ErrorsCount < MAX_COUNT_FILE_ERROR ) {
      char *err, tmp[100];
      int length;

      length = snprintf( tmp, 100, "%s %u: '%s': %s", app->Strs[72], file->Line, file_name, app->Strs[69] ); /* "Line" "File is empty" */
      if( length > 0 ) {
	if(( err = strdup( tmp )) == NULL ) {
	  ERR_SYS( "%s", app->Strs[1] ); /* "Couldn't allocate enough memory" */
	  return( 1 );
	}
	Add_error( file, err );
      }
    }
    return( 1 );
  }

  return( 0 );
}

/*
  Purpose: Convert source to boolean
*/
unsigned int            Convert_source_to_boolean(s_file *file, s_src *source, _Bool *ret_bool)
{
  switch( source->Length ) {
  case 1:
    if( source->Mem[0] == '1' ) { *ret_bool = 1; return( 0 ); }
    else if( source->Mem[0] == '0' ) { *ret_bool = 0; return( 0 ); }
    break;

  case 2:
    if( ! strncasecmp( source->Mem, "on", 2 )) { *ret_bool = 1; return( 0 ); }
    else if( ! strncasecmp( source->Mem, "no", 2 )) { *ret_bool = 0; return( 0 ); }
    break;

  case 3:
    if( ! strncasecmp( source->Mem, "off", 3 )) { *ret_bool = 0; return( 0 ); }
    else if( ! strncasecmp( source->Mem, "yes", 3 )) { *ret_bool = 1; return( 0 ); }
    break;

  case 4:
    if( ! strncasecmp( source->Mem, "true", 4 )) { *ret_bool = 1; return( 0 ); }
    break;

  case 5:
    if( ! strncasecmp( source->Mem, "false", 5 )) { *ret_bool = 0; return( 0 ); }
    break;

  case 6:
    if( ! strncasecmp( source->Mem, "enable", 6 )) { *ret_bool = 1; return( 0 ); }
    break;

  case 7:
    if( ! strncasecmp( source->Mem, "disable", 7 )) { *ret_bool = 0; return( 0 ); }
    break;
  }

  return( 83 ); /* app->Strs[83] "Unknwon value for parameter" */
}

/*
  Purpose: Convert source to number
*/
unsigned int            Convert_source_to_number(s_file *file, s_src *source, unsigned int *ret_nb)
{
  unsigned int          nb;
  char                  c;

  if( source->Mem[source->Pos] < '0' || source->Mem[source->Pos] > '9' )
    return( 80 ); /* app->Strs[80] "A number is expected" */

  nb = 0;
  for( ; source->Pos < source->Length; ++source->Pos ) {
    c = source->Mem[source->Pos];
    if( c < '0' || c > '9' )
      break;

    if( nb > UINT_MAX / 10 )
      return( 81 ); /* app->Strs[81] "Number is overflowing unsigned int max value" */

    nb *= 10;
    if( nb > UINT_MAX - ( c - '0' ))
      return( 81 ); /* app->Strs[81] "Number is overflowing unsigned int max value" */

    nb += c - '0';
  }

  *ret_nb = nb;
  return( 0 );
}

/*
  Purpose: Searching next section
*/
unsigned int            Goto_next_section(s_src *source)
{
  unsigned int          count = 0;

  while( source->Pos < source->Length ) {
    for( ; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
    if( source->Pos >= source->Length )
      break;
    ++count;
    for( ++source->Pos; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' ); ++source->Pos );
    if( source->Pos >= source->Length || source->Mem[source->Pos] == '[' )
      break;
  }

  return( count );
}

/*
  Purpose: Getting the next value
*/
unsigned int           Get_next_value(s_file *file, s_src *source, s_src *value)
{
  unsigned int         pos_start;
  char                 str_char = 0;

  for( ; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' ); ++source->Pos );
  if( source->Pos >= source->Length || source->Mem[source->Pos] == '\n' )
    return( 79 );  /* app->Strs[79] "Missing value" */

  if( !app->ConfigSeparator ) {
    if( source->Mem[source->Pos] == '=' || source->Mem[source->Pos] == ':' || source->Mem[source->Pos] == '\t') {
      app->ConfigSeparator = source->Mem[source->Pos];
      for( ++source->Pos; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' ); ++source->Pos );
    }
    else
      app->ConfigSeparator = ' ';
  }

  if( source->Mem[source->Pos] == '\'' || source->Mem[source->Pos] == '"' )
    str_char = source->Mem[source->Pos++];

  pos_start = source->Pos;

  for( ; source->Pos < source->Length; ++source->Pos ) {
    if( source->Mem[source->Pos] == '\n' )
      break;
    if( str_char ) {
      if( source->Mem[source->Pos] == str_char ) {
        str_char = 0;
        break;
      }
    }
    else if( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' )
      break;
  }

  if( str_char )
    return( 84 ); /* app->Strs[84] "Missing closing character" */

  if( source->Pos == pos_start )
    return( 79 ); /* app->Strs[79] "Missing value" */

  value->Mem = &source->Mem[pos_start];
  value->Length = source->Pos - pos_start;
  value->Pos = 0;

  /* Next value available */
  return( 0 );
}

/*
  Purpose: Getting an element between '[]' with its attributs
*/
static _Bool             Get_section(s_file *file, s_src *source)
{
  char                   *name;
  unsigned int           start, len;

  for( ++source->Pos ; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' ); ++source->Pos );
  for( start = source->Pos; source->Pos < source->Length && source->Mem[source->Pos] != ']' && source->Mem[source->Pos] != '\n'; ++source->Pos );

  if( source->Pos >= source->Length || source->Mem[source->Pos] != ']') {
    Handle_simple_error( file, app->Strs[74] ); /* "Missing closing ']'" */
    return( 1 );
  }

  len = source->Pos - start;
  if( len )
    for( ; source->Mem[start + len - 1] == ' ' || source->Mem[start + len - 1] == ' '; --len );

  if( ! len ) {
    Handle_simple_error( file, app->Strs[73] ); /* "Missing element name" */
    return( 1 );
  }

  if( len == 11 && ! strncasecmp( &source->Mem[start], "preferences", 11 ))
    return( Get_preferences( file, source ));

  if( len >= 9 && ! strncasecmp( &source->Mem[start], "template:", 9 ))
    return( Get_template( file, source, start, len ));

  if( len > MAX_LENGTH_NAME ) {
    Handle_simple_error( file, app->Strs[75] ); /* "Name is too Long" */
    return( 1 );
  }

  MALLOC( name, len + 1 );
  memcpy( name, &source->Mem[start], len );
  name[len] = '\0';

  if( strchr( name, '[' )) {
    Handle_simple_error( file, app->Strs[183] ); /*  "Invalid character in section name" */
    free( name );
    return( 1 );
  }

  return( Get_watcher_settings( file, source, name, len ));
}

/*
  Purpose: Main loop for parsing the configuration file
*/
static void             Parse_config_file(s_file *file, s_src *source)
{
  char                  c;

  file->Line = 1;
  for( ; source->Pos < source->Length; ++source->Pos ) {
    c = source->Mem[source->Pos];
    switch( c ) {
    case '\n': ++file->Line; break;

    case '[': /* new element */
      if( Get_section( file, source )) {
	file->Line += Goto_next_section( source );
	--source->Pos;
      }
      break;

    case '#': /* skip line */
      for( ++source->Pos; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
      --source->Pos;
      continue;

    case ' ':
    case '\t': /* skip spaces */
      for( ++source->Pos; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' ); ++source->Pos );
      --source->Pos;
      continue;

    default:
      Handle_simple_error( file, app->Strs[71] ) ; /* "Unexpected character into configuration file" */
      file->Line += Goto_next_section( source );
      --source->Pos;
    }
  }
}

/*
  Purpose: Sub-part of the fast sort algorithm
*/
static unsigned int    Fast_sub_sort(s_wtr **tab, int start, int end)
{
  s_wtr                *swp;
  int                  a, compt, pivot;

  compt = start;
  pivot = tab[start]->Pos;
  for( a = start + 1; a <= end; a++) {
    if( tab[a]->Pos <= pivot ) {
      ++compt;
      swp = tab[compt];
      tab[compt] = tab[a];
      tab[a] = swp;
    }
  }
  swp = tab[compt];
  tab[compt] = tab[start];
  tab[start] = swp;
  return( compt );
}

/*
  Purpose: Reorder a watchers table with the '.Pos' of each element
*/
static void             Recursive_fast_sort(s_wtr **tab, int start, int end)
{
  int                   pivot;

  if( start < end ) {
    pivot = Fast_sub_sort( tab, start, end );
    Recursive_fast_sort( tab, start, pivot - 1 );
    Recursive_fast_sort( tab, pivot + 1, end );
  }
}

/*
  Purpose: Loading the content of the file into the s_file structure
  - If content of 'file' is not zeroised, this function will assume its content is valid
*/
void                    Load_file_config(s_file *file, char *filename)
{
  struct stat           st;
  s_src                 src;
  unsigned int          a;
  int                   fd;

  VERBOSE( "%s: '%s'\n", app->Strs[181], filename ); /* "Loading configuration file" */
  if(( fd = open( filename, O_RDONLY )) < 0 ) {
    ERR_SYS( "%s: open( %s )", app->Strs[2], filename ); /* "Unable to open file" */
    goto Load_config_FAIL;
  }
  if( fstat( fd, &st ) < 0 ) {
    ERR_SYS( "%s: fstat( %s )", app->Strs[2], filename ); /* "Unable to open file" */
    close( fd );
    goto Load_config_FAIL;
  }
  if( st.st_size > MAX_LENGTH_CONF ) { /* 256 * 1024 */
    ERR_APP( "%s: %s", filename, app->Strs[70] ); /* "Sorry file is too big for me" */
    close( fd );
    goto Load_config_FAIL;
  }

  if( !st.st_size ) {
    close( fd );
    goto Load_config_FAIL;
  }

  if( file->Length != st.st_size ) {
    file->Length = (unsigned int) st.st_size;
    REALLOC( file->Mem, file->Length );
  }

  if(( read( fd, file->Mem, file->Length )) != file->Length ) {
    ERR_SYS( "%s: %s", filename, app->Strs[175] ); /* "Unable to read file" */
    close( fd );
    goto Load_config_FAIL;
  }
  close( fd );

  Reset_struct_file( file );

  file->Name = filename;

  src.Mem = file->Mem;
  src.Length = file->Length;
  src.Pos = 0;
  Parse_config_file( file, &src );

  if( app->Gui.WatcherCount )
    Recursive_fast_sort( app->Gui.WatcherTable, 0, app->Gui.WatcherCount - 1 );

  for( a = 0; a < app->Gui.WatcherCount; ++a )
    app->Gui.WatcherTable[a]->Pos = a;

  return;

 Load_config_FAIL:
  file->Length = 15;
  MALLOC( file->Mem, file->Length );
  memcpy( file->Mem, "[preferences]\n\n", file->Length );
}
