/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

static const char  *section_name = "template";

/*
  Purpose: Adding a search operation to the template
*/
_Bool                    Add_operation_to_template(s_tpl *template, s_op *operation)
{
  if( template->OperationCount == MAX_COUNT_OPERATION ) {
    Free_struct_operation( operation, 0 );
    ERR_APP( "%s %s: %s", section_name, template->Name, app->Strs[246] ); /* "Too much search operation for template" */
    return( 1 );
  }
  REALLOC( template->OperationTable, ( template->OperationCount + 1 ) * sizeof(*template->OperationTable) );
  operation->Template = 1;
  template->OperationTable[template->OperationCount++] = operation;
  return( 0 );
}

/*
  Purpose: Adding a template to the table
*/
void                     Add_template_to_file(s_file *file, s_tpl *template)
{
  if( file->TemplateCount == MAX_COUNT_TEMPLATE ) {
    Free_struct_template( template, 0 );
    ERR_APP( "%s %s: %s", section_name, template->Name, app->Strs[245] ); /* "Limit count of template reached" */
    return;
  }
  REALLOC( file->TemplateTable, ( file->TemplateCount + 1 ) * sizeof(*file->TemplateTable) );
  file->TemplateTable[file->TemplateCount++] = template;
}

/*
  Purpose: Handeling key name 'official_after_line'
*/
static void             key_official_after_line(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  char                  *pattern;
  s_op                  *operation;

  MALLOC( pattern, value->Length + 1 );
  memcpy( pattern, value->Mem, value->Length );
  pattern[value->Length] = '\0';

  operation = New_operation_after_line( pattern );
  Add_operation_to_template( template, operation );
}

/*
  Purpose: Handeling key name 'official_after_word'
*/
static void             key_official_after_word(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  char                  *pattern;
  s_op                  *operation;
  unsigned int          err, nb_word;

  if(( err =  Convert_source_to_number( file, value, &nb_word ))) {
    Pop_error( file, section_name, "official_after_word", app->Strs[err] );
    return;
  }

  if(( err = Get_next_value( file, source, value ))) {
    Pop_error( file, section_name, "official_after_word", app->Strs[err] );
    return;
  }

  MALLOC( pattern, value->Length + 1 );
  memcpy( pattern, value->Mem, value->Length );
  pattern[value->Length] = '\0';

  operation = New_operation_after_word( nb_word, pattern );
  Add_operation_to_template( template, operation );
}

/*
  Purpose: Handeling key name 'official_http_header'
*/
static void             key_official_http_header(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  char                  *options;

  MALLOC( options, value->Length + 1 );
  memcpy( options, value->Mem, value->Length );
  options[value->Length] = '\0';

  free( (char*) template->HeaderOptions );
  template->HeaderOptions = options;
}

/*
  Purpose: Handeling key name 'official_in_field'
*/
static void             key_official_in_field(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  s_op                  *operation;
  unsigned int          err, nb_field;

  if(( err =  Convert_source_to_number( file, value, &nb_field ))) {
    Pop_error( file, template->Name, "official_in_field", app->Strs[err] );
    return;
  }

  if( ! nb_field || nb_field > 65535 ) {
    Pop_error( file, template->Name, "official_in_field", app->Strs[107] ); /* "Number is out of bounds" */
    return;
  }

  if(( err = Get_next_value( file, source, value ))) {
    Pop_error( file, template->Name, "official_in_field", app->Strs[err] );
    return;
  }

  if( value->Mem[0] < ' ' || value->Mem[0] > '~' ) {
    Pop_error( file, template->Name, "official_in_field", app->Strs[121] ); /* "A delimiter specified is not permitted" */
    return;
  }

  operation = New_operation_in_field( nb_field, value->Mem[0] );
  Add_operation_to_template( template, operation );
}

/*
  Purpose: Handeling key name 'official_in_line'
*/
static void              key_official_in_line(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  char                  *pattern = 0;
  s_op                  *operation;
  unsigned int          err, nb_line;

  if(( err =  Convert_source_to_number( file, value, &nb_line ))) {
    Pop_error( file, template->Name, "official_in_line", app->Strs[err] );
    return;
  }

  if( ! nb_line || nb_line > 65535 ) {
    Pop_error( file, template->Name, "official_line", app->Strs[107] ); /* "Number is out of bounds" */
    return;
  }

  if( !( err = Get_next_value( file, source, value ))) {
    if( value->Length > MAX_LENGTH_PATTERN )
      value->Length = MAX_LENGTH_PATTERN;
    MALLOC( pattern, value->Length + 1 );
    memcpy( pattern, value->Mem, value->Length );
    pattern[value->Length] = '\0';
  }
  else if( err != 79 ) {
    Pop_error( file, template->Name, "official_in_line", app->Strs[err] ); /* "Missing closing character" */
    return;
  }

  operation = New_operation_in_line( nb_line, pattern != 0, pattern );
  Add_operation_to_template( template, operation );
}

/*
  Purpose: Handeling key name 'official_in_section'
*/
static void              key_official_in_section(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  s_op                  *operation;
  unsigned int          err, nb_section;
  char                  open_char = '<', close_char = '>';

  if(( err =  Convert_source_to_number( file, value, &nb_section ))) {
    Pop_error( file, template->Name, "official_in_section", app->Strs[err] );
    return;
  }

  if( !( err = Get_next_value( file, source, value ))) {
    if( value->Length >= 2 ) {
      open_char = value->Mem[0];
      close_char = value->Mem[1];
    }
    else if( value->Length == 1 ) {
      Pop_error( file, template->Name, "official_in_section", app->Strs[189] ); /* "Missing delimiter character" */
      return;
    }
  }
  else if( err == 84 ) {
    Pop_error( file, template->Name, "official_in_section", app->Strs[84] ); /* "Missing closing character" */
    return;
  }

  operation = New_operation_in_section( nb_section, open_char, close_char );
  Add_operation_to_template( template, operation );
}

/*
  Purpose: Handeling key name 'official_in_word'
*/
static void              key_official_in_word(s_file *file, s_src *source, s_src *value, s_tpl *template)
{
  s_op                  *operation;
  unsigned int          err, nb_word, pos = 1;
  char                  delim = 0;
  _Bool                 from_end = 0, obsolete = 0;

  if(( err =  Convert_source_to_number( file, value, &nb_word ))) {
    Pop_error( file, template->Name, "official_in_word", app->Strs[err] );
    return;
  }

  if( ! Get_next_value( file, source, value )) {
    if( value->Mem[value->Pos] == '1' || value->Mem[value->Pos] == '0'
        || value->Mem[value->Pos] == 'o' || value->Mem[value->Pos] == 'y'
        || value->Mem[value->Pos] == 't' || value->Mem[value->Pos] == 'f'
        || value->Mem[value->Pos] == 'e' || value->Mem[value->Pos] == 'd' ) {
      if(( err =  Convert_source_to_boolean( file, value, &from_end )) > 1 ) {
        if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
          Pop_error( file, template->Name, "official_in_word", app->Strs[err] );
        else
          Pop_error( file, template->Name, "local_in_word", app->Strs[err] );
        return;
      }
    }
    else {
      obsolete = 1;
      delim = value->Mem[0];
      if( ! Get_next_value( file, source, value ))
        if(( err =  Convert_source_to_number( file, value, &pos ))) {
          if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
            Pop_error( file, template->Name, "official_in_word", app->Strs[err] );
          else
            Pop_error( file, template->Name, "local_in_word", app->Strs[err] );
          return;
        }
    }
  }

  operation = New_operation_in_word( nb_word, from_end );
  Add_operation_to_template( template, operation );
  if( obsolete ) {
    operation = New_operation_in_field( pos, delim );
    Add_operation_to_template( template, operation );
  }
}

/*
  Purpose: Contains all key names to configure preferences
*/
static const s_tkn     template_key_name[] = {
  { "official_after_line" , key_official_after_line , 19 },
  { "official_after_word" , key_official_after_word , 19 },
  { "official_http_header", key_official_http_header, 20 },
  { "official_in_field"   , key_official_in_field   , 17 },
  { "official_in_line"    , key_official_in_line    , 16 },
  { "official_in_section" , key_official_in_section , 19 },
  { "official_in_word"    , key_official_in_word    , 16 }
};
#define COUNT_TEMPLATE_KEY_NAME     (sizeof(template_key_name) / sizeof(*template_key_name))

/*
  Purpose: Getting template
*/
_Bool                   Get_template(s_file *file, s_src *source, unsigned int start, unsigned int length)
{
  char                  *str;
  const s_tkn           *item;
  s_tpl                 *template;
  s_src                 value;
  unsigned int          a, err;
  char                  c;

  ASSERT( length < 9, "length < 9");

  if( length - 9 == 0 ) {
    Pop_error( file, section_name, "", app->Strs[73] ); /* "Missing element name" */
    return( 1 );
  }

  start += 9;
  length -= 9;
  MALLOC( str, length + 1 );
  memcpy( str, &source->Mem[start], length );
  str[length] = '\0';

  /* Verifying unicity of name */
  for( a = 0; a < app->Gui.File.TemplateCount; ++a)
    if( ! strcmp( str, app->Gui.File.TemplateTable[a]->Name )) {
      Pop_error( file, str, "", app->Strs[97] ); /* "This name is already used" */
      free( str );
      return( 1 );
    }

  /* Going to end of section line*/
  for( ++source->Pos ; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
  ++file->Line;

  template = New_template( str );

  /* Skipping evental empty lines */
  for( ++source->Pos; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' || source->Mem[source->Pos] == '\n'); ++source->Pos )
    if( source->Mem[source->Pos] == '\n' )
      ++file->Line;

  /* Main loop */
  for( ; source->Pos < source->Length; ++source->Pos ) {
    c = source->Mem[source->Pos];
    if( c == '\n' ) {
      ++file->Line;
      for( ++source->Pos ; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' || source->Mem[source->Pos] == '\n'); ++source->Pos )
	if( source->Mem[source->Pos] == '\n' )
	  ++file->Line;
      --source->Pos;
      continue;
    }
    if( c == '[' ) { /* Other section found */
      --source->Pos;
      break;
    }
    if( c == '#' ) { /* Skipping comments */
      for( ++source->Pos; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
      --source->Pos;
      continue;
    }

    for( a = 0; a < COUNT_TEMPLATE_KEY_NAME; ++a ) {
      item = &template_key_name[a];
      if( c == item->Name[0] ) {
	if( source->Pos + item->Length <= source->Length && ! strncmp( item->Name, &source->Mem[source->Pos], item->Length )) {
	  source->Pos += item->Length;
	  /* key_name found */
	  break;
	}
      }
    }

    if( a == COUNT_TEMPLATE_KEY_NAME ) {
      /* key_name not found */
      length = source->Pos;
      for( ; source->Pos < source->Length; ++source->Pos )
	if( source->Mem[source->Pos] == ' ' ||  source->Mem[source->Pos] == '\t'
	    || source->Mem[source->Pos] == '\n' || source->Mem[source->Pos] == app->ConfigSeparator )
	  break;
      length = source->Pos - length;
      MALLOC( str, length + 1 );
      if( length )
	memcpy( str, &source->Mem[source->Pos - length], length );
      str[length] = '\0';
      Pop_error( file, section_name, str, app->Strs[78] ); /* "Unknown key name" */
      free( str );
      for( ++source->Pos ; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
      ++file->Line;
    }
    else {
      if(( err = Get_next_value( file, source, &value ))) {
	if( err == 2 )
	  Pop_error( file, section_name, item->Name, app->Strs[84] ); /* "Missing closing character" */
	else
	  Pop_error( file, section_name, item->Name, app->Strs[79] ); /* "Missing value" */
	Free_struct_template( template, 0 );
	return( 1 );
      }

      ASSERT( !item->Func, "Function for '%s' key, has not been set", item->Name );
      item->Func( file, source, &value, template );
    }
    for( ; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
    ++file->Line;
  }

  if( ! template->OperationCount ) {
    Pop_error( file, section_name, template->Name, app->Strs[77] ); /* "Missing parameter in section" */
    Free_struct_template( template, 0 );
    return( 1 );
  }

  Add_template_to_file( file, template );

  return( 0 );
}
