/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Handeling key name 'arg'
*/
static void             key_arg(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                  *arg;

  if( watcher->LocalVersion ) {
    Pop_error( file, watcher->Name, "arg", app->Strs[100] ); /* "'local_version' has been previously used, skipping parameter" */
    return;
  }

  MALLOC( arg, value->Length + 1 );
  memcpy( arg, value->Mem, value->Length );
  arg[value->Length] = '\0';

  free( watcher->LocalArg );
  watcher->LocalArg = arg;
}

/*
  Purpose: Handeling key name 'disable'
*/
static void              key_disable(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  unsigned int           err;
  _Bool                  state;

  if(( err =  Convert_source_to_boolean( file, value, &state )) > 1 ) {
    Pop_error( file, watcher->Name, "disable", app->Strs[err] );
    return;
  }

  if( state )
    watcher->Flags |= WTR_FLG_DISABLE;
  else
    watcher->Flags &= ~( WTR_FLG_DISABLE );
}

/*
  Purpose: Handeling key name 'exec'
*/
static void              key_exec(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  if( watcher->LocalVersion ) {
    Pop_error( file, watcher->Name, "exec", app->Strs[100] ); /* "'local_version' has been previously used, skipping parameter" */
    return;
  }

  if( value->Mem[0] != '/' ) {
    Pop_error( file, watcher->Name, "exec", app->Strs[154] ); /* "An absolute path is required" */
    return;
  }

  MALLOC( watcher->LocalExec, value->Length + 1 ); /* + '\0' */
  memcpy( watcher->LocalExec, value->Mem, value->Length );
  watcher->LocalExec[value->Length] = '\0';
  watcher->Flags |= WTR_FLG_LOCAL_AUTO;
}

/*
  Purpose: Handeling key name 'icon'
*/
static void               key_icon(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  const char              *home = 0;
  char                    *icon_file;
  GdkPixbuf               *pix;
  GError                  *err = 0;
  unsigned int            home_length = 0;

  if( value->Mem[0] == '~' ) {
    home = app->UserHome;
    home_length = app->UserHomeLength;
  }

  MALLOC( icon_file, home_length + value->Length + (home_length == 0) ); /* '\0' */
  if( home_length ) {
    memcpy( icon_file, home, home_length );
    memcpy( &icon_file[home_length], &value->Mem[1], value->Length - 1 );
    icon_file[home_length + value->Length - 1] = '\0';
  }
  else {
    memcpy( icon_file, value->Mem, value->Length );
    icon_file[value->Length] = '\0';
  }

  if( Check_file( file, icon_file, 1 )) {
    free( icon_file );
    return;
  }

  pix = gdk_pixbuf_new_from_file( icon_file, &err );
  if( ! pix ) {
    if( app->Flags & APP_FLG_NO_COLORS ) {
      ERR_APP( "%s %u: %s: %s", app->Strs[72], file->Line, icon_file, app->Strs[135] ); /* "Line" "Unable to load image file" */
    }
    else {
      ERR_APP( "%s \e[33m%u\e[0m: %s: %s", app->Strs[72], file->Line, icon_file, app->Strs[135] ); /* "Line" "Unable to load image file" */
    }
  }
  else
    watcher->OriginPixbuf = pix;
  watcher->ImgPath = icon_file;
}

/*
  Purpose: Handeling key name 'local_version'
*/
static void              key_local_version(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  if( watcher->LocalExec || watcher->LocalArg ) {
    Pop_error( file, watcher->Name, "local_version", app->Strs[99] ); /* "'exec' or 'arg' has already been called, skipping parameter" */
    return;
  }
  free( watcher->LocalVersion );

  if( value->Length > MAX_LENGTH_VERSION ) {
    MALLOC( watcher->LocalVersion, MAX_LENGTH_VERSION + 4 ); /* + '...' + '\0' */
    memcpy( watcher->LocalVersion, value->Mem, MAX_LENGTH_VERSION );
    watcher->LocalVersion[MAX_LENGTH_VERSION] = '.';
    watcher->LocalVersion[MAX_LENGTH_VERSION + 1] = '.';
    watcher->LocalVersion[MAX_LENGTH_VERSION + 2] = '.';
    watcher->LocalVersion[MAX_LENGTH_VERSION + 3] = '\0';
  }
  else {
    MALLOC( watcher->LocalVersion, value->Length + 1 );
    memcpy( watcher->LocalVersion, value->Mem, value->Length );
    watcher->LocalVersion[value->Length] = '\0';
  }
  return;
}

/*
  Purpose: Handeling key name 'official|local_after_line'
*/
static void             key_search_after_line(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                  *pattern;
  s_op                  *operation;

  MALLOC( pattern, value->Length + 1 );
  memcpy( pattern, value->Mem, value->Length );
  pattern[value->Length] = '\0';

  operation = New_operation_after_line( pattern );
  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
    Add_official_operation_to_watcher( watcher, operation );
  else
    Add_local_operation_to_watcher( watcher, operation );
}

/*
  Purpose: Handeling key name 'official|local_after_word'
*/
static void             key_search_after_word(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                  *pattern;
  s_op                  *operation;
  unsigned int          err, nb_word;

  if(( err =  Convert_source_to_number( file, value, &nb_word ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_after_word", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_after_word", app->Strs[err] );
    return;
  }

  if(( err = Get_next_value( file, source, value ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_after_word", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_after_word", app->Strs[err] );
    return;
  }

  MALLOC( pattern, value->Length + 1 );
  memcpy( pattern, value->Mem, value->Length );
  pattern[value->Length] = '\0';

  operation = New_operation_after_word( nb_word, pattern );
  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
    Add_official_operation_to_watcher( watcher, operation );
  else
    Add_local_operation_to_watcher( watcher, operation );
}

/*
  Purpose: Handeling key name 'official|local_in_section'
*/
static void             key_search_in_section(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  s_op                  *operation;
  unsigned int          err, nb_section;
  char                  open_char = '<', close_char = '>';

  if(( err =  Convert_source_to_number( file, value, &nb_section ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_section", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_in_section", app->Strs[err] );
    return;
  }

  if( !( err = Get_next_value( file, source, value ))) {
    if( value->Length >= 2 ) {
      open_char = value->Mem[0];
      close_char = value->Mem[1];
    }
    else if( value->Length == 1 ) {
      if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
	Pop_error( file, watcher->Name, "official_in_section", app->Strs[189] ); /* "Missing delimiter character" */
      else
	Pop_error( file, watcher->Name, "local_in_section", app->Strs[189] ); /* "Missing delimiter character" */
      return;
    }
  }
  else if( err == 84 ) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_section", app->Strs[84] ); /* "Missing closing character" */
    else
      Pop_error( file, watcher->Name, "local_in_section", app->Strs[84] ); /* "Missing closing character" */
    return;
  }

  operation = New_operation_in_section( nb_section, open_char, close_char );
  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
    Add_official_operation_to_watcher( watcher, operation );
  else
    Add_local_operation_to_watcher( watcher, operation );
}

/*
  Purpose: Handeling key name 'official|local_in_word'
*/
static void             key_search_in_word(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  s_op                  *operation;
  unsigned int          err, nb_word, pos = 1;
  char                  delim = 0;
  _Bool                 from_end, obsolete = 0;

  if(( err =  Convert_source_to_number( file, value, &nb_word ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_word", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_in_word", app->Strs[err] );
    return;
  }

  if( ! Get_next_value( file, source, value )) {
    if( value->Mem[value->Pos] == '1' || value->Mem[value->Pos] == '0'
	|| value->Mem[value->Pos] == 'o' || value->Mem[value->Pos] == 'y'
	|| value->Mem[value->Pos] == 't' || value->Mem[value->Pos] == 'f'
	|| value->Mem[value->Pos] == 'e' || value->Mem[value->Pos] == 'd' ) {
      if(( err =  Convert_source_to_boolean( file, value, &from_end )) > 1 ) {
	if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
	  Pop_error( file, watcher->Name, "official_in_word", app->Strs[err] );
	else
	  Pop_error( file, watcher->Name, "local_in_word", app->Strs[err] );
	return;
      }
    }
    else {
      obsolete = 1;
      delim = value->Mem[0];
      if( ! Get_next_value( file, source, value ))
	if(( err =  Convert_source_to_number( file, value, &pos ))) {
	  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
	    Pop_error( file, watcher->Name, "official_in_word", app->Strs[err] );
	  else
	    Pop_error( file, watcher->Name, "local_in_word", app->Strs[err] );
	  return;
	}
    }
  }

  operation = New_operation_in_word( nb_word, from_end );
  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
    Add_official_operation_to_watcher( watcher, operation );
  else
    Add_local_operation_to_watcher( watcher, operation );
  if( obsolete ) {
    operation = New_operation_in_field( pos, delim );
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Add_official_operation_to_watcher( watcher, operation );
    else
      Add_local_operation_to_watcher( watcher, operation );
  }
}

/*
  Purpose: Handeling key name 'official|local_in_line'
*/
static void             key_search_in_line(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                  *pattern = 0;
  s_op                  *operation;
  unsigned int          err, nb_line;

  if(( err =  Convert_source_to_number( file, value, &nb_line ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_line", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_in_line", app->Strs[err] );
    return;
  }

  if( ! nb_line || nb_line > 65535 ) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_line", app->Strs[107] ); /* "Number is out of bounds" */
    else
      Pop_error( file, watcher->Name, "local_line", app->Strs[107] ); /* "Number is out of bounds" */
    return;
  }

  if( !( err = Get_next_value( file, source, value ))) {
    if( value->Length > MAX_LENGTH_PATTERN )
      value->Length = MAX_LENGTH_PATTERN;
    MALLOC( pattern, value->Length + 1 );
    memcpy( pattern, value->Mem, value->Length );
    pattern[value->Length] = '\0';
  }
  else if( err != 79 ) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_line", app->Strs[err] ); /* "Missing closing character" */
    else
      Pop_error( file, watcher->Name, "local_in_line", app->Strs[err] ); /* "Missing closing character" */
    return;
  }

  operation = New_operation_in_line( nb_line, pattern != 0, pattern );
  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
    Add_official_operation_to_watcher( watcher, operation );
  else
    Add_local_operation_to_watcher( watcher, operation );
}

/*
  Purpose: Handeling key name 'official|local_in_field'
*/
static void             key_search_in_field(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  s_op                  *operation;
  unsigned int          err, nb_field;

  if(( err =  Convert_source_to_number( file, value, &nb_field ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_field", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_in_field", app->Strs[err] );
    return;
  }

  if( ! nb_field || nb_field > 65535 ) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_field", app->Strs[107] ); /* "Number is out of bounds" */
    else
      Pop_error( file, watcher->Name, "local_in_field", app->Strs[107] ); /* "Number is out of bounds" */
    return;
  }

  if(( err = Get_next_value( file, source, value ))) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_field", app->Strs[err] );
    else
      Pop_error( file, watcher->Name, "local_in_field", app->Strs[err] );
    return;
  }

  if( value->Mem[0] < ' ' || value->Mem[0] > '~' ) {
    if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
      Pop_error( file, watcher->Name, "official_in_field", app->Strs[121] ); /* "A delimiter specified is not permitted" */
    else
      Pop_error( file, watcher->Name, "local_in_field", app->Strs[121] ); /* "A delimiter specified is not permitted" */
    return;
  }

  operation = New_operation_in_field( nb_field, value->Mem[0] );
  if( file->Flags & FILE_FLG_SEARCH_OFFICIAL )
    Add_official_operation_to_watcher( watcher, operation );
  else
    Add_local_operation_to_watcher( watcher, operation );
}

/*
  Purpose: Handeling key name 'official_http_header'
*/
static void             key_official_http_header(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                  *options;

  MALLOC( options, value->Length + 1 );
  memcpy( options, value->Mem, value->Length );
  options[value->Length] = '\0';

  free( watcher->OfficialHeader );
  watcher->OfficialHeader = options;
  watcher->Flags |= WTR_FLG_OFFICIAL_HEADER;
}

/*
  Purpose: Handeling key name 'official_timer'
*/
static void             key_official_timer(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  unsigned int          err, nb;

  if(( err =  Convert_source_to_number( file, value, &nb ))) {
    Pop_error( file, watcher->Name, "official_timer", app->Strs[err] );
    return;
  }

  if( nb < 60 || nb > 525600 ) {
    Pop_error( file, watcher->Name, "official_timer", app->Strs[117] ); /* "Invalid value" */
    return;
  }

  watcher->OfficialTimer = nb;
}

/*
  Purpose: Handeling key name 'position'
*/
static void             key_position(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  unsigned int          err, nb;

  if(( err =  Convert_source_to_number( file, value, &nb ))) {
    Pop_error( file, watcher->Name, "position", app->Strs[err] );
    return;
  }

  watcher->Pos = nb;
  watcher->SavePos = nb;
}

/*
  Purpose: Handeling key name 'url'
*/
static void              key_url(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                   *url;

  if( value->Mem[0] == 'h' && value->Length >= 7 && ! strncasecmp( value->Mem, "http://", 7 )) {
    watcher->Flags |= WTR_FLG_HTTP;
    MALLOC( url, value->Length + 1 );
    memcpy( url, value->Mem, value->Length );
    url[value->Length] = '\0';
    value->Pos = 7;
  }
  else {
    watcher->Flags &= ~( WTR_FLG_HTTP );
    if( value->Mem[0] != 'h' || value->Length < 8 || strncasecmp( value->Mem, "https://", 8 )) {
      MALLOC( url, 8 + value->Length + 1 ); /* "https://" + '\0' */
      memcpy( url, "https://", 8 );
      memcpy( &url[8], value->Mem, value->Length );
      url[value->Length + 8] = '\0';
    }
    else {
      MALLOC( url, value->Length + 1 );
      memcpy( url, value->Mem, value->Length );
      url[value->Length] = '\0';
    }
    value->Pos = 8;
  }

  if( Domain_is_valid( &url[value->Pos] )) {
    Pop_error( file, watcher->Name, "url", app->Strs[120] ); /* "The URL has an unrecognized format" */
    free( url );
    return;
  }

  free( watcher->OfficialURL );
  watcher->OfficialURL = url;
}

/*
  Purpose: Handeling key name 'optionnal'
*/
static void              key_optionnal(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  unsigned int           err;
  _Bool                  state;

  if(( err =  Convert_source_to_boolean( file, value, &state )) > 1 ) {
    Pop_error( file, watcher->Name, "optionnal", app->Strs[err] );
    return;
  }

  if( state )
    watcher->Flags |= WTR_FLG_OPTIONNAL;
  else
    watcher->Flags &= ~( WTR_FLG_OPTIONNAL );
}

/*
  Purpose: Handeling key name 'stderr'
*/
static void              key_stderr(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  unsigned int           err;
  _Bool                  state;

  if(( err =  Convert_source_to_boolean( file, value, &state )) > 1 ) {
    Pop_error( file, watcher->Name, "stderr", app->Strs[err] );
    return;
  }

  if( state )
    watcher->Flags |= WTR_FLG_STDERR;
  else
    watcher->Flags &= ~( WTR_FLG_STDERR );
}

/*
  Purpose: Handeling key name 'trigger_on_mismatch'
*/
static void              key_trigger_on_mismatch(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                   *path;

  if( value->Mem[0] != '/' ) {
    Pop_error( file, watcher->Name, "trigger_on_mismatch", app->Strs[154] ); /* "An absolute path is required" */
    return;
  }

  MALLOC( path, value->Length + 1 );
  memcpy( path, value->Mem, value->Length );
  path[value->Length] = '\0';

  watcher->Trigger_on_mismatch = path;
  watcher->Flags |= WTR_FLG_TRIGGER_MISMATCH;
}

/*
  Purpose: Handeling key name 'trigger_on_error'
*/
static void              key_trigger_on_error(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                   *path;

  if( value->Mem[0] != '/' ) {
    Pop_error( file, watcher->Name, "trigger_on_error", app->Strs[154] ); /* "An absolute path is required" */
    return;
  }

  MALLOC( path, value->Length + 1 );
  memcpy( path, value->Mem, value->Length );
  path[value->Length] = '\0';

  watcher->Trigger_on_error = path;
  watcher->Flags |= WTR_FLG_TRIGGER_ERROR;
}

/*
  Purpose: Handeling key name 'use_template'
*/
static void             key_use_template(s_file *file, s_src *source, s_wtr *watcher, s_src *value)
{
  char                  *name;

  MALLOC( name, value->Length + 1 );
  memcpy( name, value->Mem, value->Length );
  name[value->Length] = '\0';

  free( watcher->TemplateName );
  watcher->TemplateName = name;
}

/*
  Purpose: Contains search key names to configure a watcher
*/
static const s_wkn      watcher_key_name[] = {
  { "arg"                    , key_arg                  , 3 },
  { "disable"                , key_disable              , 7 },
  { "exec"                   , key_exec                 , 4 },
  { "icon"                   , key_icon                 , 4 },
  { "local_version"          , key_local_version        , 13 },
  { "local_after_line"       , key_search_after_line    , 16 },
  { "local_after_word"       , key_search_after_word    , 16 },
  { "local_in_field"         , key_search_in_field      , 14 },
  { "local_in_line"          , key_search_in_line       , 13 },
  { "local_in_section"       , key_search_in_section    , 16 },
  { "local_in_word"          , key_search_in_word       , 13 },
  { "official_after_line"    , key_search_after_line    , 19 },
  { "official_after_word"    , key_search_after_word    , 19 },
  { "official_in_field"      , key_search_in_field      , 17 },
  { "official_in_line"       , key_search_in_line       , 16 },
  { "official_in_section"    , key_search_in_section    , 19 },
  { "official_in_word"       , key_search_in_word       , 16 },
  { "official_http_header"   , key_official_http_header , 20 },
  { "official_timer"         , key_official_timer       , 14 },
  { "optionnal"              , key_optionnal            , 9 },
  { "position"               , key_position             , 8 },
  { "stderr"                 , key_stderr               , 6 },
  { "trigger_on_error"       , key_trigger_on_error     , 16 },
  { "trigger_on_mismatch"    , key_trigger_on_mismatch  , 19 },
  { "url"                    , key_url                  , 3 },
  { "use_template"           , key_use_template         , 12 }
};
#define COUNT_KEY_NAME    (sizeof(watcher_key_name) / sizeof(*watcher_key_name))

/*
  Purpose: Adding the watcher to the main table
*/
void                    Add_watcher_to_gui(s_gui *gui, s_wtr *watcher)
{
  if( gui->WatcherCount + 1 >= gui->WatcherTableSize ) {
    gui->WatcherTableSize += STEP_SIZE_WATCHER_TABLE;
    REALLOC( gui->WatcherTable, gui->WatcherTableSize * sizeof(*gui->WatcherTable) );
    bzero( &gui->WatcherTable[gui->WatcherCount], STEP_SIZE_WATCHER_TABLE * sizeof(*gui->WatcherTable) );
  }
  gui->WatcherTable[gui->WatcherCount++] = watcher;
}

/*
  Purpose: Checking the validity of the watcher
*/
static _Bool            Check_validity_watcher(s_file *file, s_wtr *watcher)
{
  if( ! watcher->LocalVersion && ! watcher->LocalExec ) {
    Pop_error( file, watcher->Name, "'local_version' / 'exec'", app->Strs[128] ); /* "Missing mandatory element" */
    return( 1 );
  }

  if( watcher->LocalExec )
    if( Check_file( file, watcher->LocalExec, 0 ))
      if( watcher->Flags & WTR_FLG_OPTIONNAL )
	watcher->Flags |= WTR_FLG_DISABLE;

  if( ! watcher->OfficialURL || ! watcher->OfficialURL[0] ) {
    Pop_error( file, watcher->Name, "url", app->Strs[128] ); /* "Missing mandatory element" */
    return( 1 );
  }
  return( 0 );
}

/*
  Purpose: Getting the properties of the watcher from the source
*/
static void             Get_keyname_parameters(s_file *file, s_src *source, s_wtr *watcher, const s_wkn *item)
{
  unsigned int          err;
  s_src                 value;

  source->Pos += item->Length;
  if(( err = Get_next_value( file, source, &value ))) {
    Pop_error( file, watcher->Name, item->Name, app->Strs[err] );
    return;
  }

  if( item->Name[0] == 'o' ) file->Flags |= FILE_FLG_SEARCH_OFFICIAL;
  else file->Flags &= ~( FILE_FLG_SEARCH_OFFICIAL );

  ASSERT( !item->Func, "Function for '%s' key, has not been set", item->Name );
  item->Func( file, source, watcher, &value );
}

/*
  Purpose: Getting the properties of the watcher from the source
*/
static void             Get_watcher_properties(s_file *file, s_src *source, s_wtr *watcher)
{
  const s_wkn           *item;
  unsigned int          a;
  char                  c, c2;

  for( ; source->Pos < source->Length; ++source->Pos ) {
    c = source->Mem[source->Pos];
    if( c == '\n' ) {
      ++file->Line;
      for( ++source->Pos ; source->Pos < source->Length && ( source->Mem[source->Pos] == ' ' || source->Mem[source->Pos] == '\t' || source->Mem[source->Pos] == '\n'); ++source->Pos )
	if( source->Mem[source->Pos] == '\n' )
	  ++file->Line;
      --source->Pos;
      continue;
    }
    if( c == '[' ) { /* Other section found */
      --source->Pos;
      break;
    }
    if( c == '#' ) { /* Skipping comments */
      for( ++source->Pos; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
      --source->Pos;
      continue;
    }

    for( a = 0; a < COUNT_KEY_NAME; ++a ) {
      item = &watcher_key_name[a];
      if( c == item->Name[0] && source->Pos + item->Length < source->Length
	  && ! strncmp( item->Name, &source->Mem[source->Pos], item->Length )
	  && source->Pos + item->Length + 1 < source->Length ) {
	c2 = source->Mem[source->Pos + item->Length];
	if( c2 == ' ' || c2 == '\t' || c2 == app->ConfigSeparator ) {
	  Get_keyname_parameters( file, source, watcher, item );
	  break;
	}
      }
    }

    if( a == COUNT_KEY_NAME ) {
      if(( item = Get_obsolete_watchers_properties( file, source, watcher )))
	Get_keyname_parameters( file, source, watcher, item );
      else {
	char str[31];
	for( a = 0; a < 30 && source->Pos + a < source->Length; ++a) {
	  if( source->Mem[source->Pos + a] == ' ' || source->Mem[source->Pos + a] == '\n' || source->Mem[source->Pos + a] == '\t' )
	    break;
	  if( a > 27 ) {
	    str[a++] = '.';
	    str[a++] = '.';
	    str[a++] = '.';
	    break;
	  }
	  str[a] = source->Mem[source->Pos + a];
	}
	str[a] = '\0';
	Pop_error( file, watcher->Name, str, app->Strs[78] ); /* "Unknown key name" */
      }
    }
    for( ; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
    --source->Pos;
  }
}

/*
  Purpose: Creating a new watcher from source
*/
static void             Reorder_obsolete_operation(s_op **table, unsigned int count)
{
  s_op                  *operation, *swp;
  unsigned int          a, last_pos = 0;

  /* obsolete 'After line' is first */
  for( a = 0; a < count; ++a ) {
    operation = table[a];
    if( operation->Type == OP_TYPE_AFTER_LINE && operation->Reorder ) {
      swp = table[0];
      table[0] = operation;
      table[a] = swp;
      break;
    }
  }

  /* obsolete 'In line' is before 'In section' */
  for( a = 0; a < count; ++a ) {
    operation = table[a];
    if( operation->Type == OP_TYPE_IN_SECTION && operation->Reorder ) {
      last_pos = a;
      ++a;
      break;
    }
  }
  for( ; a < count; ++a ) {
    operation = table[a];
    if( operation->Type == OP_TYPE_IN_LINE ) {
      swp = table[last_pos];
      table[last_pos] = operation;
      table[a] = swp;
      break;
    }
  }

  /* obsolete 'In field' is last */
  for( a = 0; a < count; ++a ) {
    operation = table[a];
    if( operation->Type == OP_TYPE_IN_FIELD && operation->Reorder ) {
      if( a != count - 1 ) {
	for( ; a + 1 < count; ++a )
	  table[a] = table[a + 1];
	table[a] = operation;
      }
      break;
    }
  }
}

/*
  Purpose: Creating a new watcher from source
  - 'name' is freed if something came wrong during the process
*/
_Bool                   Get_watcher_settings(s_file *file, s_src *source, char *name, unsigned char length)
{
  s_wtr                 *watcher;
  unsigned int          a;

  /* Verifying section is not empty */
  for( ++source->Pos ; source->Pos < source->Length && source->Mem[source->Pos] != '\n'; ++source->Pos );
  if( source->Pos >= source->Length ) {
    Pop_error( file, name, "", app->Strs[77] ); /* "Missing parameter in section" */
    free( name );
    return( 1 );
  }
  ++source->Pos;
  ++file->Line;

  /* Verifying unicity of name */
  for( a = 0; a < app->Gui.WatcherCount; ++a)
    if( ! strcmp( name, app->Gui.WatcherTable[a]->Name )) {
      Pop_error( file, name, "", app->Strs[97] ); /* "This name is already used" */
      free( name );
      return( 1 );
    }

  CALLOC( watcher, 1, sizeof(*watcher) );
  watcher->Name = name;
  watcher->NameLength = length;
  watcher->OfficialTimer = DEFAULT_TIMER;
  watcher->Pos = app->Gui.WatcherCount;
  watcher->SavePos = app->Gui.WatcherCount;
  watcher->OriginPixbuf = app->Gui.ImgTable[IMG_DEFAULT_WATCHER];

  Get_watcher_properties( file, source, watcher );

  if( Check_validity_watcher( file, watcher )) {
    Free_struct_watcher( watcher, 1, 0 );
    return( 1 );
  }

  Reorder_obsolete_operation( watcher->OfficialOperationTable, watcher->OfficialOperationCount );

  Add_watcher_to_gui( &app->Gui, watcher );
  return( 0 );
}
