/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Retrieving output from the child into dest
*/
static _Bool        Retrieve_from_std(s_wtr *watcher, int fd, char *buffer)
{
  struct timeval    time;
  s_dest            dest;
  s_src             source;
  s_so              search;
  fd_set            slrd;
  unsigned int      count = 0;
  int               rd, ret;

  bzero( &dest, sizeof(dest) );
  while( 1 ) {
    FD_ZERO( &slrd );
    FD_SET( fd, &slrd );
    time.tv_sec = 10;
    time.tv_usec = 0;

    if(( ret = select( fd + 1, &slrd, NULL, NULL, &time )) < 0 ) {
      snprintf( buffer, SIZE_BUFFER_ERROR, "select(): %s", strerror( errno ));
      goto error;
    }
    if( ! ret ) {
      if( count++ < 2 ) /* 2x 10 sec should be more than enough */
	continue;
      snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[67] ); /* "Command timed out" */
      goto error;
    }

    if( FD_ISSET( fd, &slrd )) {
    next_buffer:
      if( dest.Pos >= MAX_STD_OUPUT_LENGTH - STEP_SIZE_BUFFER_READ ) {
	WARN_APP( "%s", app->Strs[199] ); /* "Command output too large" */
	break;
      }

      if( dest.Pos + STEP_SIZE_BUFFER_READ >= dest.Length ) {
	dest.Length += STEP_SIZE_BUFFER_READ;
	REALLOC( dest.Mem, dest.Length );
      }

      rd = read( fd, &dest.Mem[dest.Pos], STEP_SIZE_BUFFER_READ );
      if( rd < 0 ) {
	snprintf( buffer, SIZE_BUFFER_ERROR, "read(): %s", strerror( errno ));
	goto error;
      }

      if( ! rd )
	break;
      dest.Pos += rd;

      goto next_buffer;
    }
  }

  if( ! dest.Pos ) {
    snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[165] ); /* "No output available from executable" */
    goto error;
  }

  source.Mem = dest.Mem;
  source.Length = dest.Pos;
  source.Pos = 0;

  bzero( &search, sizeof(search) );
  search.OperationTable = watcher->LocalOperationTable;
  search.OperationCount = watcher->LocalOperationCount;

  if( Apply_search_operation( &search, &source, buffer ))
    goto error;

  if( Normalize_version( &source, buffer ))
    goto error;

  free( dest.Mem );
  return( 0 );

 error:
  ERR_APP( "%s: %s: %s", watcher->Name, app->Strs[136], buffer ); /* "Update local version" */
  return( 1 );
}

/*
  Purpose: Retrieving local version from the output of the watcher's executable
*/
static _Bool        ExecuteCommand(s_wtr *watcher, char *buffer)
{
  pid_t             pid;
  int               std[2];

  if( watcher->LocalVersion ) {
    free( watcher->LocalVersion );
    watcher->LocalVersion = 0;
  }

  if( Check_file( &app->Gui.File, watcher->LocalExec, 0 )) {
    snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[173] ); /* "Executable file is invalid" */
    ERR_APP( "%s: %s: '%s'", watcher->Name, buffer, watcher->LocalExec );
    return( 1 );
  }

  VERBOSE( "%s: %s: '%s %s'\n", watcher->Name, app->Strs[139], watcher->LocalExec, watcher->LocalArg ); /* "Retrieving local version from" */

  if( pipe( std ) < 0 ) {
    snprintf( buffer, SIZE_BUFFER_ERROR, "pipe(): %s", strerror( errno ));
    ERR_APP( "%s", buffer );
    return( 1 );
  }

  if(( pid = fork()) < 0 ) {
    snprintf( buffer, SIZE_BUFFER_ERROR, "fork(): %s", strerror( errno ));
    ERR_APP( "%s", buffer );
    close( std[0] );
    close( std[1] );
    return( 1 );
  }
  else if( ! pid ) {
    char *tab[3];

    close( std[0] );
    close( 0 );
    if( watcher->Flags & WTR_FLG_STDERR ) {
      close( 1 );
      dup2( std[1], 2 );
    }
    else {
      close( 2 );
      dup2( std[1], 1 );
    }
    tab[0] = watcher->LocalExec;
    tab[1] = watcher->LocalArg;
    tab[2] = 0;
    execv( tab[0], tab );
    ERR_SYS( "%s: %s", tab[0], app->Strs[63] ); /* "Unable to launch command, execv() failed" */
    exit( EXIT_FAILURE );
  }

  close( std[1] );

  if( Retrieve_from_std( watcher, std[0], buffer )) {
    close( std[0] );
    return( 1 );
  }
  close( std[0] );
  STRDUP( watcher->LocalVersion, buffer );
  return( 0 );
}

/*
  Purpose: Retrieving local version for watcher
*/
void                Retrieve_local_version(s_wtr *watcher)
{
  char              *str;

  LOCK_GTK();
  gtk_label_set_label( GTK_LABEL(watcher->LabelLocalVersionResult), app->Strs[59] ); /* "In progress" */
  UNLOCK_GTK();

  if( watcher->Flags & WTR_FLG_LOCAL_AUTO ) {
    ASSERT( ! watcher->LocalExec, "%s: watcher->LocalExec == 0", watcher->Name );
    char buffer[SIZE_BUFFER_ERROR];
    if( ExecuteCommand( watcher, &buffer[0] )) {
      str = Error_to_xml( &buffer[0] );
      LOCK_GTK();
      gtk_label_set_markup( GTK_LABEL(watcher->LabelLocalVersionResult), str );
      UNLOCK_GTK();
      free( str );
      watcher->Flags |= WTR_FLG_LOCAL_ERROR;
      watcher->Flags &= ~( WTR_FLG_LOCAL_IN_PROGRESS | WTR_FLG_VERSION_MISMATCH | WTR_FLG_LOCAL_REFRESH );
      return;
    }
    ASSERT( !watcher->LocalVersion, "watcher->LocalVersion == 0" );
    VERBOSE( "%s: %s: %s\n", watcher->Name, app->Strs[167], watcher->LocalVersion ); /* "Local version found" */
  }
  else {
    ASSERT( !watcher->LocalVersion, "watcher->LocalVersion == 0" );
    VERBOSE( "%s:%s%s\n", watcher->Name, app->Strs[12], watcher->LocalVersion ); /* " Local version (manual): " */
  }

  LOCK_GTK();
  gtk_label_set_label( GTK_LABEL(watcher->LabelLocalVersionResult), watcher->LocalVersion );
  UNLOCK_GTK();
  watcher->Flags &= ~( WTR_FLG_LOCAL_IN_PROGRESS | WTR_FLG_VERSION_MISMATCH | WTR_FLG_LOCAL_REFRESH | WTR_FLG_LOCAL_ERROR );
}
