/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Creating trigger table to be executed
  - This function has potential memory leaks, to be used only within an exiting child created with fork()
*/
static _Bool            Trigger_to_table(s_wtr *watcher, char *trigger, char ***table)
{
  char                   **tab, **tmp;
  unsigned int           a, bkp;
  unsigned char          count;
  char                   c;

  tab = 0;
  count = 0;
  for( a = 0; trigger[a]; ++a ) {
    switch( trigger[a] ) {
    case '"':
    case '\'':
      c = trigger[a++];

      for( bkp = a; trigger[a] && trigger[a] != c; ++a );

      if(( tmp = realloc( tab, ( count + 1 ) * sizeof(char*) )) == NULL ) {
	ERR_SYS( "%s", app->Strs[2] ); /* "Couldn't allocate enough memory" */
	return( 1 ); /* don't care about freeing */
      }
      tab = tmp;

      tab[count] = &trigger[bkp];
      trigger[a] = '\0';
      if( *tab[count] == '@' ) {
	if( !strcmp( tab[count], "@local" ))
	  tab[count] = watcher->LocalVersion;
	else if( !strcmp( tab[count], "@official" ))
	  tab[count] = watcher->OfficialVersion;
	else if( !strcmp( tab[count], "@name" ))
	  tab[count] = watcher->Name;
      }
      ++count;
      break;

    case ' ':
    case '\t':
      for( ++a; trigger[a] && ( trigger[a] == ' ' || trigger[a] == '\t' ); ++a );
      --a;
      break;

    default:
      for( bkp = a; trigger[a] && trigger[a] != ' ' && trigger[a] != '\t'; ++a );

      if(( tmp = realloc( tab, ( count + 1 ) * sizeof(char*) )) == NULL ) {
	ERR_SYS( "%s", app->Strs[2] ); /* "Couldn't allocate enough memory" */
	return( 1 ); /* don't care about freeing */
      }
      tab = tmp;

      tab[count] = &trigger[bkp];
      if( !trigger[a] )
	--a;
      else
	trigger[a] = '\0';

      if( *tab[count] == '@' ) {
	if( !strcmp( tab[count], "@local" ))
	  tab[count] = watcher->LocalVersion;
	else if( !strcmp( tab[count], "@official" ))
	  tab[count] = watcher->OfficialVersion;
	else if( !strcmp( tab[count], "@name" ))
	  tab[count] = watcher->Name;
      }
      ++count;
      break;
    }
    if( count >= 255 )
      break;
  }

  if(( tmp = realloc( tab, ( count + 1 ) * sizeof(char*) )) == NULL ) {
    ERR_SYS( "%s", app->Strs[2] ); /* "Couldn't allocate enough memory" */
    return( 1 ); /* don't care about freeing */
  }
  tab = tmp;
  tab[count] = 0;

  *table = tab;
  return( 0 );
}

/*
 Purpose: Launching executable for the trigger on mismatch
*/
static void         Trigger_execute(s_wtr *watcher, _Bool on_error)
{
  pid_t             pid;
  char              *trigger;

  if( on_error ) {
    if( ! watcher->Trigger_on_error )
      return;
    trigger = watcher->Trigger_on_error;
    VERBOSE( "%s: %s: '%s'\n", watcher->Name, app->Strs[222], trigger ); /* "Trigger on error activated" */
  }
  else if( ! watcher->Trigger_on_mismatch )
    return;
  else {
    trigger = watcher->Trigger_on_mismatch;
    VERBOSE( "%s: %s: '%s'\n", watcher->Name, app->Strs[221], trigger ); /* "Trigger on mismatch activated" */
  }

  if(( pid = fork()) < 0 ) {
    ERR_SYS( "%s", app->Strs[62] ); /* "Unable to launch command, fork() failed" */
  }
  else if( ! pid ) {
    char **tab;
    if( ! Trigger_to_table( watcher, trigger, &tab )) {
      execv( tab[0], tab );
      ERR_SYS( "%s ...: %s", tab[0], app->Strs[63] ); /* "Unable to launch command, execv() failed" */
    }
    exit( EXIT_FAILURE );
  }
}

/*
  Purpose: Checking the network state, and fill err in case of network down
*/
static unsigned int       Check_network_state(char *url)
{
  char                   *port;
  struct addrinfo        *res, *tmp;
  struct addrinfo        hints;
  int                    fd_udp, ret;
  unsigned char          a, flg = 0;

  /* Domain names to test DNS */
  const char *fqdn[] = {
    "eff.org", "gnu.org",
    "kernel.org", "openstreetmap.org",
    "wikipedia.org"
  };
#define SIZE_TABLE_FQDN  (sizeof(fqdn) / sizeof(*fqdn))

  bzero( &hints, sizeof(hints) );
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_family = PF_UNSPEC;

  port = "80";
  ret = strlen( url );
  if( ret >= 7 && ! strncmp( "http://", url, 7 ))
    url = &url[7];
  else if( ret >= 8 && ! strncmp( "https://", url, 8 )) {
    port = "443";
    url = &url[8];
  }

  for( a = 0; url[a] != '\0'; ++a )
    if( url[a] == '/' ) {
      url[a] = '\0';
      flg = 1;
      break;
    }

  if(( ret = getaddrinfo( url, port, &hints, &res )) != 0 ) {
    if( ret != EAI_NONAME ) {
      ERR_APP( "getaddrinfo(%s): %d", url, ret );
    }
    if( flg )
      url[a] = '/';

    /* We try some other DNS calls */
    for( a = 0; a < SIZE_TABLE_FQDN; ++a )
      if( ! getaddrinfo( fqdn[a], "80", &hints, &res ))
	break;

    if( a == SIZE_TABLE_FQDN ) /* None were found = no available DNS or... we're pretty fu*$! */
      return( 251 ); /* app->Strs[251] "No network response (DNS error)" */
    return( 254 ); /* app->Strs[254] "URL not resolved (DNS error)" */
  }

  if( flg )
    url[a] = '/';

  for( tmp = res; tmp; tmp = tmp->ai_next ) {
    if(( fd_udp = socket( AF_INET, SOCK_DGRAM, 0 )) == -1 )
      continue;
    /*
       We try a UDP connection which doesn't send any packet
       But triggers network stack which will tells if local network is down
    */
    if( ! connect( fd_udp, tmp->ai_addr, tmp->ai_addrlen )) {
      close( fd_udp );
      freeaddrinfo( res );
      /* local network is up  */
      return( 0 );
    }
    close( fd_udp );
  }

  freeaddrinfo( res );
  /* local network is down  */
  return( 250 ); /* app->Strs[250] "Network link is down" */
}


/*
   Purpose: Convert line to table
*/
static char       **String_to_table(const char *line, unsigned int *count)
{
  const char      *start, *found;
  char            **tab = 0;
  unsigned int    len, length, nb = 1;

  length = strlen( line );
  start = line;

  while(( found = memmem( start, length, ",", 1 ))) {
    len = found - start;

    REALLOC( tab, nb * sizeof(char*) );
    MALLOC( tab[nb - 1], 9 + len + 1 );
    memcpy( tab[nb - 1], "--header=", 9 );
    memcpy( &tab[nb - 1][9], start, len );
    tab[nb - 1][9 + len] = '\0';
    ++nb;

    start = found + 1;
    length -= ( len + 1 );
  }

  REALLOC( tab, nb * sizeof(char*) );
  MALLOC( tab[nb - 1], 9 + length + 1 );
  memcpy( tab[nb - 1], "--header=", 9 );
  memcpy( &tab[nb - 1][9], start, length );
  tab[nb - 1][9 + length] = '\0';

  *count = nb;
  return( tab );
}

/*
  Purpose: Retrieving the web resource from the watcher's URL
*/
unsigned int         Download_resource_to_cache(char *url, const char *header, const char *dest_file)
{
  pid_t              pid;
  unsigned int       err;

  ASSERT( !url, "url = 0" );

  if( Domain_is_valid( url )) {
    ERR_APP( "%s - %s", app->Strs[120], url ); /* "The URL has an unrecognized format" */
    return( 120 );
  }

  if(( err = Check_network_state( url ))) {
    ERR_APP( "%s - %s", app->Strs[err], url );
    return( err );
  }

  if(( pid = fork()) < 0 ) {
    ERR_SYS( "%s", app->Strs[62] ); /* "Unable to launch command, fork() failed" */
    return( 252 ); /* app->Strs[252] "A system call failed" */
  }
  else if( !pid ) {
    char **tab, **tab_header = 0;
    unsigned int count = 0, nb = 0;

    if( header )
      tab_header = String_to_table( header, &nb );

    MALLOC( tab, ( 12 + nb ) * sizeof(*tab) );
    close( 0 );
    close( 1 );

    tab[count++] = "/usr/bin/wget";
    tab[count++] = "--user-agent";
    tab[count++] = "Updates-warner";
    tab[count++] = "-q";
    tab[count++] = (char*) url;
    tab[count++] = "-O";
    tab[count++] = (char*) dest_file;
    tab[count++] = "-T";
    tab[count++] = "20";
    tab[count++] = "-t";
    tab[count++] = "3";
    if( tab_header ) {
      memcpy( &tab[count], tab_header, nb * sizeof(char*) );
      count += nb;
    }
    tab[count++] = 0;
    execv( tab[0], tab );
    ERR_SYS( "'%s %s': %s", tab[0], tab[1], app->Strs[63] ); /* "Unable to launch command, execv() failed" */
    exit( EXIT_FAILURE );
  }
  waitpid( pid, NULL, 0 );
  return( 0 );
}

/*
  Purpose: Prepare the error to xml, replacong non-xml-complient characters, if neeeded
*/
char                *Error_to_xml(const char *str)
{
  char              *tmp = 0;
  unsigned int      a, pos, count, length;
  char              c;

  ASSERT( !str, "str = 0" );

  for( a = count = 0; str[a] ; ++a) {
    c = str[a];
    if( c == '<' || c == '>' )
      ++count;
  }

  length = strlen( str );
  MALLOC( tmp, 30 + count * 4 + length + 11 + 1 );
  memcpy( tmp, "<span foreground='#FF0000'><b>", 30 );
  pos = 30;

  for( a = 0; str[a] ; ++a ) {
    c = str[a];
    if( c == '<' ) {
      tmp[pos++] = '&';
      tmp[pos++] = 'l';
      tmp[pos++] = 't';
      tmp[pos++] = ';';
    }
    else if( c == '>' ) {
      tmp[pos++] = '&';
      tmp[pos++] = 'g';
      tmp[pos++] = 't';
      tmp[pos++] = ';';
    }
    else
      tmp[pos++] = c;
  }

  memcpy( &tmp[pos], "</b></span>", 11 );
  pos += 11;

  tmp[pos++] = '\0';
  return( tmp );
}

/*
  Purpose: Computing the XOR result
*/
void                Build_xor(const s_wtr *watcher, char *xor)
{
  unsigned int      a;

  bzero( xor, SIZE_BUFFER_XOR );
  if( watcher->LocalVersion )
    for( a = 0; watcher->LocalVersion[a]; ++a )
      xor[a % SIZE_BUFFER_XOR] ^= watcher->LocalVersion[a];

  if( watcher->OfficialVersion )
    for( a = 0; watcher->OfficialVersion[a]; ++a )
      xor[a % SIZE_BUFFER_XOR] ^= watcher->OfficialVersion[a];
}

/*
  Purpose: Checking the trigger xor and launching trigger thread if needed
*/
static char         Check_trigger_xor(s_wtr *watcher, _Bool on_error)
{
  char              xor[SIZE_BUFFER_XOR];

  if( watcher->Flags & WTR_FLG_INIT )
    return( 0 );

  Build_xor( watcher, xor );

  if( memcmp( xor, watcher->Trigger_xor, SIZE_BUFFER_XOR )) {
    memcpy( watcher->Trigger_xor, xor, SIZE_BUFFER_XOR );
    Trigger_execute( watcher, on_error );
    return( 1 );
  }

  return( 0 );
}

/*
  Purpose: Retrieving the official version from the cache file
*/
static _Bool        Get_official_version_from_file(s_wtr *watcher, const char *file, char *buffer)
{
  s_src             source;
  s_so             search;
  unsigned int      err = 0;

  watcher->Flags |= WTR_FLG_NO_AUTO_UPDATE;

  VERBOSE( "%s: %s: '%s'\n", watcher->Name, app->Strs[140], file ); /* "Retrieving official version from" */

  if(( err = Open_file( file, &watcher->CacheFile ))) {
    snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[err] );
    goto error;
  }

  if( watcher->OfficialVersion ) {
    free( watcher->OfficialVersion );
    watcher->OfficialVersion = 0;
  }

  source.Mem = watcher->CacheFile.Source.Mem;
  source.Length = watcher->CacheFile.Source.Length;
  source.Pos = 0;

  bzero( &search, sizeof(search) );
  search.OperationTable = watcher->OfficialOperationTable;
  search.OperationCount = watcher->OfficialOperationCount;
  search.TemplateName = watcher->TemplateName;

  if( Apply_search_operation( &search, &source, buffer )) {
    if( !( watcher->Flags & WTR_FLG_INIT )) {
      if( watcher->Name ) {
	WARN_APP( "%s: %s", watcher->Name, buffer );
      }
      else {
	WARN_APP( "%s", buffer );
      }
    }
    goto error;
  }

  if( Normalize_version( &source, buffer ))
    goto error;

  if( watcher->CacheFile.Source.Mem )
    munmap( (char*) watcher->CacheFile.Source.Mem, watcher->CacheFile.Source.Length );
  return( 0 );

 error:
  if( watcher->CacheFile.Source.Mem )
    munmap( (char*) watcher->CacheFile.Source.Mem, watcher->CacheFile.Source.Length );
  return( 1 );
}

/*
  Purpose: Retrieving the official version from the cache file
*/
void                Retrieve_official_version_from_cache(s_wtr *watcher, const char *file, _Bool lock_gui)
{
  char              *tmp, buffer[SIZE_BUFFER_ERROR];

  if( Get_official_version_from_file( watcher, file, &buffer[0] )) {
    free( watcher->OfficialVersion );
    watcher->OfficialVersion = 0;

    tmp = Error_to_xml( buffer );
    if( lock_gui ) { LOCK_GTK(); }
    gtk_label_set_markup( GTK_LABEL(watcher->LabelOfficialVersionResult), tmp );
    if( lock_gui ) { UNLOCK_GTK(); }
    free( tmp );
    watcher->Flags &= ~( WTR_FLG_OFFICIAL_IN_PROGRESS | WTR_FLG_VERSION_MISMATCH );
    watcher->Flags |= WTR_FLG_OFFICIAL_ERROR;

    Check_trigger_xor( watcher, 1 );
    return;
  }


  free( watcher->OfficialVersion );
  STRDUP( watcher->OfficialVersion, buffer );

  VERBOSE( "%s: %s: '%s'\n", watcher->Name, app->Strs[144], watcher->OfficialVersion ); /* "Official version found" */

  if( lock_gui ) { LOCK_GTK(); }
  gtk_label_set_label( GTK_LABEL(watcher->LabelOfficialVersionResult), watcher->OfficialVersion );
  if( lock_gui ) { UNLOCK_GTK(); }

  watcher->Flags &= ~( WTR_FLG_OFFICIAL_ERROR | WTR_FLG_OFFICIAL_IN_PROGRESS | WTR_FLG_VERSION_MISMATCH | WTR_FLG_NO_AUTO_UPDATE | WTR_FLG_OFFICIAL_REFRESH );
}

/*
  Purpose: Copying a file
*/
_Bool               Copy_file(const char *source, const char *dest)
{
  char              *err;
  struct stat       st;
  char              buf[8192];
  ssize_t           ret;
  int               fd_src, fd_dest;

  if(( fd_src = open( source, O_RDONLY )) < 0 ) {
    err = "open";
    goto error;
  }

  if( fstat( fd_src, &st ) < 0 ) {
    err = "fstat";
    goto error_src;
  }

  if(( fd_dest = open( dest, O_CREAT | O_WRONLY, S_IRWXU )) < 0 ) {
    err = "open";
    goto error_src;
  }

  if( ftruncate( fd_dest, st.st_size )) {
    err = "ftruncate";
    goto error_all;
  }

  posix_fadvise( fd_src, 0, 0, POSIX_FADV_SEQUENTIAL );

  while(( ret = read( fd_src, buf, 8192 )) > 0 )
    if( write( fd_dest, buf, ret ) != ret ) {
      err = "write";
      goto error_all;
    }

  if( ret < 0 ) {
    err = "read";
    goto error_all;
  }

  close( fd_src );
  close( fd_dest );
  return( 0 );

 error_all:
  close( fd_dest );
 error_src:
  close( fd_src );
 error:
  ERR_SYS( "'%s': %s: %s()", source, app->Strs[160], err ); /* "A problem occured when moving file" */
  return( 1 );
}

/*
  Purpose: Updating the 'Official result' entry after any modification by the user
*/
void                     Update_official_result()
{
  char                   *tmp;
  s_wtr                  *watcher = &app->Gui.TmpWatcher;
  s_wtr                  *selected = app->Gui.SelectedWatcher;
  s_so                   *search;
  pthread_t              tmp_id;

  search = Operations_to_search( watcher->OfficialOperationTable, watcher->OfficialOperationCount );
  if( ! search )
    return;

  search->Watcher = watcher;
  search->CacheFile = Build_cachefile_name( "tmp_watcher_result", 18 );

  if( selected && !( app->Gui.Flags & GUI_FLG_NEW_WATCHER )) {
    tmp = Build_cachefile_name( selected->Name, selected->NameLength );
    Copy_file( tmp, search->CacheFile );
    free( tmp );
  }

  search->TemplateName = watcher->TemplateName;
  search->Url = watcher->OfficialURL;
  if( watcher->OfficialHeader )
    search->Header = watcher->OfficialHeader;
  else if( search->TemplateName ) {
    unsigned int a;
    for( a = 0; a < app->Gui.File.TemplateCount; ++a )
      if( ! strcmp( search->TemplateName, app->Gui.File.TemplateTable[a]->Name )) {
	search->Header = app->Gui.File.TemplateTable[a]->HeaderOptions;
	break;
      }
  }
  watcher->OfficialURL = 0;

  search->Result = GTK_ENTRY(app->Gui.WatcherSettings[WS_ETY_OFFI_RESULT]);

  if( pthread_create( &tmp_id, &app->Attr, th_Update_from_search, search )) {
    ERR_APP( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    Free_struct_search( search, 0 );
  }
}

/*
  Purpose: Updating the timer in the infoview of the watcher
*/
void               Update_timer(s_wtr *watcher, time_t now, _Bool from_thread)
{
  char             buffer[50];
  unsigned int     ongoing;

  if( watcher->CacheFile.LastUpdate > now )
    now = watcher->CacheFile.LastUpdate;

  ongoing = (unsigned int) ( now - watcher->CacheFile.LastUpdate ) / 60;

  if( ongoing >= watcher->OfficialTimer )
    snprintf( buffer, 50, "0 %s", app->Strs[123] ); /* "minutes" */
  else {
    ongoing = watcher->OfficialTimer - ongoing;
    if( ongoing >= 1440 ) {
      if( ongoing % 60 == 0 ) {
	if( ( ongoing % 1440 ) / 60 == 0 )
	  snprintf( buffer, 50, "%u %s", ongoing / 1440, app->Strs[184] );  /* "days" */
	else
	  snprintf( buffer, 50, "%u %s - %02u %s", ongoing / 1440, app->Strs[184],  /* "days" */
		    ( ongoing % 1440 ) / 60, app->Strs[185] );  /* "hours" */
      }
      else if( ( ongoing % 1440 ) / 60 == 0 )
	snprintf( buffer, 50, "%u %s - %02u %s", ongoing / 1440, app->Strs[184],  /* "days" */
		  ongoing % 60, app->Strs[123] ); /* "minutes" */
      else
	snprintf( buffer, 50, "%u %s - %02u %s - %02u %s", ongoing / 1440, app->Strs[184],  /* "days" */
		  ( ongoing % 1440 ) / 60, app->Strs[185],  /* "hours" */
		  ongoing % 60, app->Strs[123] ); /* "minutes" */
    }
    else if( ongoing > 60 ) {
      if( ongoing % 60 == 0 )
	snprintf( buffer, 50, "%u %s", ongoing / 60, app->Strs[185] );  /* "hours" */
      else
	snprintf( buffer, 50, "%u %s - %02u %s", ongoing / 60, app->Strs[185],  /* "hours" */
		  ongoing % 60, app->Strs[123] ); /* "minutes" */
    }
    else
      snprintf( buffer, 50, "%u %s", ongoing, app->Strs[123] ); /* "minutes" */
  }

  if( from_thread ) { LOCK_GTK(); }
  gtk_label_set_label( GTK_LABEL(watcher->LabelNextUpdate), buffer );
  if( from_thread ) { UNLOCK_GTK(); }
}

/*
  Purpose: Updating the draw area of the watcher
*/
void               Update_watcher_area(s_wtr *watcher, _Bool lock_gtk)
{
  struct timeval   now;

  watcher->Flags &= ~( WTR_FLG_AREA_REFRESH );
  if( !( watcher->Flags & ( WTR_FLG_LOCAL_IN_PROGRESS | WTR_FLG_OFFICIAL_IN_PROGRESS ))) {
    /* here, both version are retrieved */
    if( ! watcher->LocalVersion || ! watcher->OfficialVersion ) {
      SNO( watcher->Flags & WTR_FLG_VERSION_MISMATCH, "LocalVersion : %u - OfficialVersion : %u",
	   watcher->LocalVersion != 0, watcher->OfficialVersion != 0 );
      if( !( app->Flags & APP_FLG_PHONE )) {
#ifndef LIB_HANDY
	if( lock_gtk ) { LOCK_GTK(); }
	if( !( app->Flags & APP_FLG_BLINK ) && !( watcher->Flags & WTR_FLG_INIT )) {
	  app->Flags |= APP_FLG_BLINK;
	  gtk_status_icon_set_from_pixbuf( app->Systray, watcher->CurrentPixbuf );
#  if GTK_MAJOR_VERSION < 3
	  gtk_status_icon_set_blinking( app->Systray, 1 );
#  endif
#  if ( GTK_MAJOR_VERSION == 2 && GTK_MINOR_VERSION >= 18 ) || ( GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION < 14 )
	  gtk_status_icon_set_title( app->Systray, watcher->Name );
#  endif
	}
	if( lock_gtk ) { UNLOCK_GTK(); }
#endif /* LIB_HANDY */
      }
    }
    else {
      SNO( ! strlen( watcher->LocalVersion ) || ! strlen( watcher->OfficialVersion ), "LocalVersion : %u - OfficialVersion : %u",
	   (unsigned int)strlen(watcher->LocalVersion), (unsigned int)strlen(watcher->OfficialVersion) );
      if( strcmp( watcher->LocalVersion, watcher->OfficialVersion )) {
	watcher->Flags |= WTR_FLG_VERSION_MISMATCH;

	if( Check_trigger_xor( watcher, 0 ) && !( app->Flags & APP_FLG_PHONE )) {
#ifndef LIB_HANDY
	  if( lock_gtk ) { LOCK_GTK(); }
	  if( !( app->Flags & APP_FLG_BLINK )) {
	    app->Flags |= APP_FLG_BLINK;
	    gtk_status_icon_set_from_pixbuf( app->Systray, watcher->CurrentPixbuf );
#  if GTK_MAJOR_VERSION < 3
	    gtk_status_icon_set_blinking( app->Systray, 1 );
#  endif
#  if ( GTK_MAJOR_VERSION == 2 && GTK_MINOR_VERSION >= 18 ) || ( GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION < 14 )
	    gtk_status_icon_set_title( app->Systray, watcher->Name );
#  endif
	  }
	  if( lock_gtk ) { UNLOCK_GTK(); }
#endif /* LIB_HANDY */
	}
      }
      else
	Build_xor( watcher, watcher->Trigger_xor );
    }
  }

  if( lock_gtk ) { LOCK_GTK(); }
  Update_watcher_infosview( watcher );

  Callback_on_expose_watcher( watcher->DrawArea, 0, watcher );

#if GTK_MAJOR_VERSION < 3
  gdk_flush();
#else
  gdk_display_flush( gtk_widget_get_display( watcher->DrawArea ));
#endif
  if( lock_gtk ) { UNLOCK_GTK(); }

  gettimeofday( &now, NULL );
  Update_timer( watcher, now.tv_sec, lock_gtk );
}

/*
  Purpose: Determining whether or not a refresh of the web resource is required
*/
static unsigned char    Official_refresh_required(s_wtr *watcher, const char *file)
{
  struct stat           st;
  struct timeval        now;

  if( watcher->Flags & WTR_FLG_NO_AUTO_UPDATE )
    return( 0 );

  gettimeofday( &now, NULL );
  if( stat( file, &st )) {
    /* here, cache file is missing */
    watcher->CacheFile.LastUpdate = now.tv_sec - watcher->OfficialTimer * 60;
    switch( errno ) {
    case ENOENT:
      VERBOSE( "%s: %s\n", watcher->Name, app->Strs[143] ); /* "The cache file is not available" */
      break;

    default:
      ERR_SYS( "Official_refresh_required(): stat( %s ): %s\n", file, strerror( errno ));
      break;
    }
    watcher->Flags |= WTR_FLG_OFFICIAL_UPDATE;
    return( 1 );
  }

  watcher->CacheFile.LastUpdate = st.st_ctim.tv_sec;
  if( ! st.st_size ) {
    VERBOSE( "%s: %s\n", watcher->Name, app->Strs[142] ); /* "The downloaded file is empty" */
    watcher->Flags |= WTR_FLG_OFFICIAL_UPDATE;
    return( 1 );
  }
  if( now.tv_sec - st.st_ctim.tv_sec > watcher->OfficialTimer * 60 ) {
    VERBOSE( "%s: %s\n", watcher->Name, app->Strs[141] ); /* "The update timer has expired" */
    watcher->Flags |= WTR_FLG_OFFICIAL_UPDATE;
    return( 1 );
  }
  if( watcher->Flags & WTR_FLG_OFFICIAL_UPDATE ) {
    VERBOSE( "%s: %s\n", watcher->Name, app->Strs[198] ); /* "Updating official version, manually requested" */
    return( 1 );
  }

  return( 0 );
}

/*
  Purpose: Creating the full path of the cache file
  - returned value should be free'd after use
*/
char                *Build_cachefile_name(const char *file_name, unsigned int file_len)
{
  char              *file;
  unsigned int      cache_len;

  LOCK_MUTEX( app->LockCache );
  cache_len = strlen( app->CacheFolder );
  MALLOC( file, cache_len + file_len + 2 ); /* '/' + '\0' */
  memcpy( file, app->CacheFolder, cache_len );
  UNLOCK_MUTEX( app->LockCache );

  if( file[cache_len - 1] != '/' )
    file[cache_len++] = '/';

  memcpy( &file[cache_len], file_name, file_len );
  file[cache_len + file_len] = '\0';

  return( file );
}

/*
  Purpose: Updating official version in a thread
*/
void                *th_Refresh_official_version(void *param)
{
  s_wtr             *watcher = param;

  if( !( watcher->Flags & WTR_FLG_OFFICIAL_IN_PROGRESS )) {
    watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE | WTR_FLG_OFFICIAL_ERROR );
    watcher->Flags |= WTR_FLG_OFFICIAL_IN_PROGRESS;
    Update_watcher_area( watcher, 1 );
    Refresh_official_version( watcher );
  }

  return( 0 );
}

static _Bool        Get_resource(s_wtr *watcher, const char *file)
{
  const char        *header = 0;
  char              *tmp;
  unsigned int      a, err;

  if( watcher->TemplateName )
    for( a = 0; a < app->Gui.File.TemplateCount; ++a )
      if( ! strcmp( watcher->TemplateName, app->Gui.File.TemplateTable[a]->Name )) {
	header = app->Gui.File.TemplateTable[a]->HeaderOptions;
	break;
      }
  if( watcher->OfficialHeader )
    header = watcher->OfficialHeader;
  if(( err = Download_resource_to_cache( watcher->OfficialURL, header, file ))) {
    tmp = Error_to_xml( app->Strs[err] );
    LOCK_GTK();
    gtk_label_set_markup( GTK_LABEL(watcher->LabelOfficialVersionResult), tmp );
    UNLOCK_GTK();
    free( tmp );
    watcher->Flags &= ~( WTR_FLG_INIT | WTR_FLG_OFFICIAL_IN_PROGRESS );
    watcher->Flags |= WTR_FLG_OFFICIAL_ERROR;
    Update_watcher_area( watcher, 1 );
    return( 1 );
  }
  watcher->Flags &= ~( WTR_FLG_OFFICIAL_UPDATE | WTR_FLG_INIT );
  return( 0 );
}

/*
  Purpose: Refreshing info of the official version
  - This function MUST NOT be called from the main GTK thread
*/
void                Refresh_official_version(s_wtr *watcher)
{
  char              *file;
  _Bool             skip = 0;

  /*
    Note: Build_cachefile_name() is called each minute for each enabled watcher, in this way of doing I'm wasting CPU but saving memory
    The opposite way would be to save the path for each watcher, and modify the path of each watcher only when user changes the cache folder
    I'm thinking about an other 'in between' solution
  */
  file = Build_cachefile_name( watcher->Name, watcher->NameLength );
  if( Official_refresh_required( watcher, file )) {
    if( watcher->Flags & WTR_FLG_OFFICIAL_UPDATE )
      skip = Get_resource( watcher, file );
    if( ! skip ) {
      Retrieve_official_version_from_cache( watcher, file, 1 );
      Update_watcher_area( watcher, 1 );
    }
  }
  else if( watcher->Flags & ( WTR_FLG_LOCAL_IN_PROGRESS | WTR_FLG_OFFICIAL_IN_PROGRESS | WTR_FLG_OFFICIAL_REFRESH )) {
    watcher->Flags &= ~( WTR_FLG_LOCAL_IN_PROGRESS | WTR_FLG_OFFICIAL_IN_PROGRESS);
    Retrieve_official_version_from_cache( watcher, file, 1 );
    Update_watcher_area( watcher, 1 );
  }
  else {
    struct timeval now;
    gettimeofday( &now, NULL );
    Update_timer( watcher, now.tv_sec, 1 );
  }
  free( file );

  watcher->Flags &= ~( WTR_FLG_INIT );
}
