/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Searching the option line. The positions of the first and last characters
           of the line are returned, the position of the first character of the
	   value is also returned.
  - returned end = 0 means option not found
  - returned value = 0 means value not found
*/
static void             Search_option(const char *option, unsigned int length, s_dest *source,
				      unsigned int *start, unsigned int *value, unsigned int *end)
{
  unsigned int          a, max;
  char                  c;

  *start = 0;
  *value = 0;
  *end = 0;
  if( source->Length <= length )
    return;
  max = source->Length - length;
  c = *option;
  for( a = 0; a < max; ++a ) {
    for( *start = a; a < max && ( source->Mem[a] == ' ' || source->Mem[a] == '\t' ); ++a );
    if( c == source->Mem[a] && !strncmp( &source->Mem[a], option, length ) ) {
      a += length;
      if( a + 1 >= source->Length ) {
	*end = a;
	return;
      }

      for( ; a < source->Length && ( source->Mem[a] == ' ' || source->Mem[a] == '\t' ); ++a );

      if( source->Mem[a] == '=' || source->Mem[a] == ':' )
	for( ++a; a < source->Length && ( source->Mem[a] == ' ' || source->Mem[a] == '\t' ); ++a );

      if( a < source->Length && source->Mem[a] != '\n' ) {
	*value = a;
	for( ; a < source->Length && source->Mem[a] != '\n'; ++a );
      }
      *end = a;
      return;
    }
    for( ; a < max && source->Mem[a] != '\n'; ++a );
  }
  return;
}

/*
  Purpose: Adding a string value to the destination
*/
static _Bool            Add_string(s_dest *dest, const char *value, unsigned int val_len)
{
  char                  *tmp;
  unsigned int          length, new_length;
  unsigned char         quote = 0;

  ASSERT( !value, "value = 0" );
  ASSERT( !val_len, "val_len = 0" );

  if( memchr( value, '\'', val_len ))
    quote = '"';
  else if( memchr( value, '"', val_len ) || memchr( value, ' ', val_len ) || memchr( value, '\t', val_len ))
    quote = '\'';

  length = val_len + 2 * (quote != 0) + 1;

  if( dest->Pos + length >= dest->Length ) {
    new_length = STEP_BUFFER_FILE;
    if( length > STEP_BUFFER_FILE )
      new_length = length + STEP_BUFFER_FILE;
    if( dest->Length > UINT_MAX - new_length )
      return( 1 );
    dest->Length += new_length;
    if(( tmp = realloc( dest->Mem, dest->Length )) == NULL )
      return( 1 );
    dest->Mem = tmp;
  }

  if( quote )
    dest->Mem[dest->Pos++] = quote;
  memcpy( &dest->Mem[dest->Pos], value, val_len );
  dest->Pos += val_len;
  if( quote )
    dest->Mem[dest->Pos++] = quote;
  return( 0 );
}

/*
  Purpose: Adding a number value to the destination
*/
static _Bool            Add_number(s_dest *dest, unsigned int nb)
{
  char                  buffer[10], *tmp;
  unsigned char         pos, length;

  pos = 10;
  if( nb ) {
    while( nb ) {
      buffer[--pos] = '0' + ( nb % 10 );
      nb /= 10;
    }
    length = 11 - pos;
  }
  else
    length = 2;

  if( dest->Pos + length >= dest->Length ) {
    if( dest->Length > UINT_MAX - STEP_BUFFER_FILE )
      return( 1 );
    dest->Length += STEP_BUFFER_FILE;
    if(( tmp = realloc( dest->Mem, dest->Length )) == NULL )
      return( 1 );
    dest->Mem = tmp;
  }

  if( pos < 10 ) {
    memcpy( &dest->Mem[dest->Pos], &buffer[pos], 10 - pos );
    dest->Pos += 10 - pos;
  }
  else
    dest->Mem[dest->Pos++] = '0';

  return( 0 );
}

/*
   Purpose: Removing option from file buffer
*/
static void             Remove_option(const char *option, unsigned int len_opt, s_dest *dest)
{
  unsigned int          start, pos_val, end, length;

  Search_option( option, len_opt, dest, &start, &pos_val, &end );

  if( end ) {
    if( end < dest->Length && dest->Mem[end] == '\n' )
      ++end;
    length = end - start;
    memmove( &dest->Mem[start], &dest->Mem[end], dest->Length - start - length);
    dest->Length -= length;
  }
}

/*
  Purpose: Adding raw data in destination
*/
static _Bool           Add_raw_buffer(s_dest *dest, const char *buffer, unsigned int length)
{
  char                 *tmp;
  unsigned int         new_length;

  if( dest->Pos + length >= dest->Length ) {
    new_length = STEP_BUFFER_FILE;
    if( length > STEP_BUFFER_FILE )
      new_length = length + STEP_BUFFER_FILE;
    if( dest->Length > UINT_MAX - new_length )
      return( 1 );
    dest->Length += new_length;
    if(( tmp = realloc( dest->Mem, dest->Length )) == NULL )
      return( 1 );
    dest->Mem = tmp;
  }

  memcpy( &dest->Mem[dest->Pos], buffer, length );
  dest->Pos += length;
  return( 0 );
}

/*
  Purpose: Adding an search operation into the destination
*/
static _Bool            Add_official_operation(s_dest *dest, s_op *operation)
{
  switch( operation->Type ) {
  case OP_TYPE_AFTER_LINE:
    if( operation->Reorder ) {
      Remove_option( "official_after_contains", 23, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "official_after_line ", 20 ) || Add_string( dest, operation->Pattern, strlen( operation->Pattern )))
      return( 1 );
    break;

  case OP_TYPE_AFTER_WORD:
    if( Add_raw_buffer( dest, "official_after_word ", 20 ) || Add_number( dest, operation->Number ))
      return( 1 );
    dest->Mem[dest->Pos++] = ' ';
    if( Add_string( dest, operation->Pattern, strlen( operation->Pattern )))
      return( 1 );
    break;

  case OP_TYPE_IN_LINE:
    if( operation->Reorder ) {
      Remove_option( "official_line", 13, dest );
      Remove_option( "official_line_contains", 22, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "official_in_line ", 17 ) || Add_number( dest, operation->Number ))
      return( 1 );
    if( operation->Activated ) {
      dest->Mem[dest->Pos++] = ' ';
      if( Add_string( dest, operation->Pattern, strlen( operation->Pattern )))
	return( 1 );
    }
    break;

  case OP_TYPE_IN_SECTION:
    if( operation->Reorder ) {
      Remove_option( "official_field", 14, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "official_in_section ", 20 ) || Add_number( dest, operation->Number ))
      return( 1 );
    dest->Mem[dest->Pos++] = ' ';
    if( Add_string( dest, operation->Delimiter, 2 ))
      return( 1 );
    break;

  case OP_TYPE_IN_FIELD:
    if( operation->Reorder ) {
      Remove_option( "official_cut_with", 17, dest );
      Remove_option( "official_cut_pos", 16, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "official_in_field ", 18 ) || Add_number( dest, operation->Number ))
      return( 1 );
    dest->Mem[dest->Pos++] = ' ';
    if( Add_string( dest, operation->Delimiter, 1 ))
      return( 1 );
    break;

  case OP_TYPE_IN_WORD:
    if( Add_raw_buffer( dest, "official_in_word ", 17 ) || Add_number( dest, operation->Number ))
      return( 1 );
    if( operation->Activated ) {
      dest->Mem[dest->Pos++] = ' ';
      if( Add_raw_buffer( dest, "1", 1 ))
	return( 1 );
    }
    break;

  default:
    ASSERT( 1, "Unknown operation->Type = %u", operation->Type );
  }

  dest->Mem[dest->Pos++] = '\n';
  return( 0 );
}

/*
  Purpose: Adding an local search operation into the destination
*/
static _Bool            Add_local_operation(s_dest *dest, s_op *operation)
{
  switch( operation->Type ) {
  case OP_TYPE_AFTER_LINE:
    if( operation->Reorder ) {
      Remove_option( "local_after_contains", 20, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "local_after_line ", 17 ) || Add_string( dest, operation->Pattern, strlen( operation->Pattern )))
      return( 1 );
    break;

  case OP_TYPE_AFTER_WORD:
    if( Add_raw_buffer( dest, "local_after_word ", 17 ) || Add_number( dest, operation->Number ))
      return( 1 );
    dest->Mem[dest->Pos++] = ' ';
    if( Add_string( dest, operation->Pattern, strlen( operation->Pattern )))
      return( 1 );
    break;

  case OP_TYPE_IN_LINE:
    if( operation->Reorder ) {
      Remove_option( "local_line", 10, dest );
      Remove_option( "local_line_contains", 19, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "local_in_line ", 14 ) || Add_number( dest, operation->Number ))
      return( 1 );
    if( operation->Activated ) {
      dest->Mem[dest->Pos++] = ' ';
      if( Add_string( dest, operation->Pattern, strlen( operation->Pattern )))
	return( 1 );
    }
    break;

  case OP_TYPE_IN_SECTION:
    if( operation->Reorder ) {
      Remove_option( "local_field", 11, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "local_in_section ", 17 ) || Add_number( dest, operation->Number ))
      return( 1 );
    dest->Mem[dest->Pos++] = ' ';
    if( Add_string( dest, operation->Delimiter, 2 ))
      return( 1 );
    break;

  case OP_TYPE_IN_FIELD:
    if( operation->Reorder ) {
      Remove_option( "local_cut_with", 14, dest );
      Remove_option( "local_cut_pos", 13, dest );
      operation->Reorder = 0;
    }
    if( Add_raw_buffer( dest, "local_in_field ", 15 ) || Add_number( dest, operation->Number ))
      return( 1 );
    dest->Mem[dest->Pos++] = ' ';
    if( Add_string( dest, operation->Delimiter, 1 ))
      return( 1 );
    break;

  case OP_TYPE_IN_WORD:
    if( Add_raw_buffer( dest, "local_in_word ", 14 ) || Add_number( dest, operation->Number ))
      return( 1 );
    if( operation->Activated ) {
      dest->Mem[dest->Pos++] = ' ';
      if( Add_raw_buffer( dest, "1", 1 ))
	return( 1 );
    }
    break;

  default:
    ASSERT( 1, "Unknown operation->Type = %u", operation->Type );
  }

  dest->Mem[dest->Pos++] = '\n';
  return( 0 );
}

/*
   Purpose: Saving data to config file
*/
static _Bool            Write_file(s_file *file)
{
  int                   fd;

#ifdef MODE_DEV
  #if 0
  LOCK_MUTEX( app->LockPrint );
  printf("************************ WRITING FILE [%u] ************************\n", file->Length );
  fd = write( 1, file->Mem, file->Length );
  printf("*****************************************************************\n");
  UNLOCK_MUTEX( app->LockPrint );
  #endif
#endif /* MODE_DEV */

  if(( fd = open( file->Name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR )) < 0 ) {
    ERR_SYS( "%s: open( %s )", app->Strs[2], file->Name ); /* "Unable to open file" */
    return( 1 );
  }

  if( ftruncate( fd, file->Length )) {
    ERR_SYS( "%s: ftruncate( %u )", app->Strs[147], file->Length ); /* "An error occured during writing into config file" */
    close( fd );
    return( 1 );
  }

  if( write( fd, file->Mem, file->Length ) != file->Length ) {
    ERR_SYS( "%s: write( %u )", app->Strs[147], file->Length ); /* "An error occured during writing into config file" */
    close( fd );
    return( 1 );
  }

  if( fsync( fd )) {
    char tmp[4096];
    snprintf( tmp, 4096, "  %s  \n  %s\n\n fsync(): %s\n",
	      app->Strs[223], app->Gui.File.Name, strerror(errno) ); /* "Something went wrong when saving changes into configuration file" */
    LOCK_GTK();
    gtk_label_set_label( GTK_LABEL(app->Error[ERW_LBL_MESSAGE]), tmp );
    gtk_widget_show( app->Error[ERW_WIN_MAIN] );
    UNLOCK_GTK();
  }

  close( fd );
  return( 0 );
}

/*
  Purpose: Searching the positions of the first character of the next section name
           and first character of the last line of the next section in the app->Gui.File.Mem
  - returned pos = UINT_MAX means not found
*/
static void             Search_section_pos(unsigned int start, unsigned int *pos, unsigned int *end)
{
  unsigned int          a, bkp;

  *pos = UINT_MAX;
  *end = 0;

  for( a = start; a < app->Gui.File.Length; ++a ) {
    if( app->Gui.File.Mem[a] == '[' ) {
      bkp = a;
      if( a ) { /* checking if's the first non space character */
	for( --a; a && ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); --a );
	if( app->Gui.File.Mem[a] != '\n' && app->Gui.File.Mem[a] != ' ' && app->Gui.File.Mem[a] != '\t' ) {
	  a = bkp;
	  continue;
	}
      }

      for( a = bkp + 1; a < app->Gui.File.Length && ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); ++a );
      *pos = a;

      /* start position found, now searching end of section */
      for( ; a < app->Gui.File.Length; ++a ) {
	if( app->Gui.File.Mem[a] == '[' ) {
	  bkp = a;
	  for( --a; ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); --a );
	  if( app->Gui.File.Mem[a] != '\n' && app->Gui.File.Mem[a] != ' ' && app->Gui.File.Mem[a] != '\t' ) {
	    a = bkp;
	    continue;
	  }
	  *end = bkp;
	  return;
	}
      }
      *end = a;
      return;
    }
  }
}

/*
  Purpose: Searching the positions of the first and last line of the section in the app->Gui.File.Mem
*/
static _Bool            Search_section(const char *section, unsigned int *pos, unsigned int *end)
{
  unsigned int          a, max, length, bkp, new_pos;
  char                  c;

  *pos = UINT_MAX;
  *end = 0;
  length = strlen( section );
  if( length >= app->Gui.File.Length )
    return( 1 ); /* SNO */

  max = app->Gui.File.Length - length;
  c = *section;

  for( a = 0; a < max; ++a ) {
    //#todo skip comments
    if( app->Gui.File.Mem[a] == '[' ) {
      bkp = a;
      if( a ) { /* checking it's the first non space character */
	for( --a; a && ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); --a );
	if( app->Gui.File.Mem[a] != '\n' && app->Gui.File.Mem[a] != ' ' && app->Gui.File.Mem[a] != '\t' ) {
	  a = bkp;
	  continue;
	}
      }
      for( a = bkp + 1; a < max && ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); ++a );
      if( app->Gui.File.Mem[a] == c && !strncmp( &app->Gui.File.Mem[a], section, length )) {
	for( a += length; a < app->Gui.File.Length && ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); ++a );
	if( a < app->Gui.File.Length && app->Gui.File.Mem[a] == ']' ) {
	  for( ++a ; a < app->Gui.File.Length && app->Gui.File.Mem[a] != '\n'; ++a );
	  if( a >= app->Gui.File.Length )
	    return( 1 );
	  new_pos = bkp;
	  /* start position found, now searching end of section */
	  for( ; a < app->Gui.File.Length; ++a ) {
	    if( app->Gui.File.Mem[a] == '[' ) {
	      bkp = a;
	      for( --a; ( app->Gui.File.Mem[a] == ' ' || app->Gui.File.Mem[a] == '\t' ); --a );
	      if( app->Gui.File.Mem[a] != '\n' && app->Gui.File.Mem[a] != ' ' && app->Gui.File.Mem[a] != '\t' ) {
		a = bkp;
		continue;
	      }
	      *pos = new_pos;
	      *end = bkp;
	      /* End of section found before an other section */
	      return( 0 );
	    }
	  }
	  *pos = new_pos;
	  *end = a;
	  /* End of section found at end of file */
	  return( 0 );
	}
      }
    }
  }
  /* Section not found */
  return( 1 );
}

/*
  Purpose: Searching the positions of the first and last line of the section in the app->Gui.File.Mem
*/
_Bool                   Search_section2(s_file *file, s_dest *target, const char *section_name)
{
  unsigned int          end;

  bzero( target, sizeof(*target) );

  if( Search_section( section_name, &target->Pos, &end)) {
    target->Pos = 0;
    return( 1 );
  }

  target->Mem = file->Mem;
  target->Length = end;
  return( 0 );
}

/*
   Purpose: Replacing content in a position of the buffer
*/
_Bool                   Replace_content(const char *data, unsigned int len_data, unsigned int available, s_dest *dest)
{
  char                  *tmp;
  unsigned int          length;

  if( len_data > available ) {
    length = len_data - available;
    if(( tmp = realloc( dest->Mem, dest->Length + length )) == NULL )
      return( 1 );
    dest->Mem = tmp;
    memmove( &dest->Mem[dest->Pos + len_data], &dest->Mem[dest->Pos + available], dest->Length - ( dest->Pos + available ));
    dest->Length += length;
  }
  else if( len_data < available ) {
    memmove( &dest->Mem[dest->Pos + len_data], &dest->Mem[dest->Pos + available], dest->Length - ( dest->Pos + available ));
    dest->Length -= ( available - len_data );
  }

  memcpy( &dest->Mem[dest->Pos], data, len_data );
  return( 0 );
}

/*
   Purpose: Adding option + value at the dest->Pos which should be the end of a section
*/
#define MISSING_EOL     (1 << 0)
#define FINISH_WITH_EOL (1 << 1)
_Bool                   Insert_option(const char *option, unsigned int len_opt, const char *value, unsigned int len_val, char quote, s_dest *dest)
{
  char                  *tmp;
  unsigned int          start, length;
  unsigned char         flg;

  flg = MISSING_EOL;
  start = dest->Pos;
  if( dest->Mem[start - 1] == '\n' ) {
    flg = 0;
    if( dest->Mem[start - 2] != '\n' )
      flg |= FINISH_WITH_EOL;
    else
      --start;
  }

  length = len_opt + 1 + len_val + 1; /* ' ' + '\n' */
  if( flg )
    ++length;

  if( app->ConfigSeparator && app->ConfigSeparator != ' ' )
    length += 2; /* + separator + ' ' */

  dest->Length += length;
  if(( tmp = realloc( dest->Mem, dest->Length )) == NULL ) {
    ERR_SYS( "%s", app->Strs[193] ); /* "A problem occured when applying modifications" */
    return( 1 );
  }
  dest->Mem = tmp;

  memmove( &dest->Mem[start + length], &dest->Mem[start], dest->Length - start - length );

  if( flg & MISSING_EOL )
    dest->Mem[start++] = '\n';

  memcpy( &dest->Mem[start], option, len_opt );
  start += len_opt;

  if( app->ConfigSeparator && app->ConfigSeparator != ' ' ) {
    dest->Mem[start++] = ' ';
    dest->Mem[start++] = app->ConfigSeparator;
  }
  dest->Mem[start++] = ' ';

  if( quote )
    dest->Mem[start++] = quote;

  memcpy( &dest->Mem[start], value, len_val );
  start += len_val;

  if( quote )
    dest->Mem[start++] = quote;

  dest->Mem[start++] = '\n';
  if( flg & FINISH_WITH_EOL )
    dest->Mem[start++] = '\n';

  return( 0 );
}

/*
  Purpose: Removing a watcher from the config file and freeing resources
*/
void                    Remove_section_from_file(const char *name)
{
  unsigned int          pos, end, length;

  Search_section( name, &pos, &end );
  if( pos == UINT_MAX || pos >= end ) {
    LOCK_MUTEX( app->LockPrint );
    pos =  write( 1, app->Gui.File.Mem, app->Gui.File.Length );
    ASSERT( 1, "[%s]: section not found (%u / %u)", name, pos, end );
  }

  length = end - pos;
  memmove( &app->Gui.File.Mem[pos], &app->Gui.File.Mem[end], app->Gui.File.Length - pos - length );
  app->Gui.File.Length -= length;
}

/*
  Purpose: Saving position of watchers
  //#todo shitty function to rewrite
*/
static void             Save_watchers_position()
{
  char                  *tmp;
  s_dest                dest;
  unsigned int          a, start, pos, end, len_title, len_tmp, count;
  unsigned int          pos_opt, pos_val, end_opt;
  char                  nb[14]; /* 2^32 + ' ' + separator + ' ' + '\0' */

  SNO( !app->Gui.WatcherCount, "Save_watchers_position() called, but no watcher available %u", 0 );

  count = 0;
  for( start = 0; start < app->Gui.File.Length; start += end - start ) {
    Search_section_pos( start, &pos, &end );
    if( pos == UINT_MAX || pos >= end )
      break;

    for( len_title = 0; pos + len_title < app->Gui.File.Length && app->Gui.File.Mem[pos + len_title] != ']'; ++len_title );
    if( len_title > 1 ) {
      if( app->Gui.File.Mem[pos + len_title] == ' ' || app->Gui.File.Mem[pos + len_title] == '\t' )
	for( --len_title; len_title && ( app->Gui.File.Mem[pos + len_title] == ' ' || app->Gui.File.Mem[pos + len_title] == '\t' ); ++len_title );

      for( a = 0; a < app->Gui.WatcherCount; ++a ) {
	tmp = app->Gui.WatcherTable[a]->Name;
	len_tmp = strlen( tmp ); //#todo: opti
	if( len_title == len_tmp && ! strncmp( tmp, &app->Gui.File.Mem[pos], len_title )) {
	  if( app->Gui.File.Mem[pos + len_title] == ']' || app->Gui.File.Mem[pos + len_title] == ' '
	      || app->Gui.File.Mem[pos + len_title] == '\t' ) {
	    if( a != app->Gui.WatcherTable[a]->SavePos ) {
	      dest.Mem = &app->Gui.File.Mem[pos];
	      dest.Pos = 0;
	      dest.Length = end - pos;

	      Search_option( "position", 8, &dest, &pos_opt, &pos_val, &end_opt );

	      dest.Mem = app->Gui.File.Mem;
	      dest.Length = app->Gui.File.Length;
	      if( end_opt ) {
		if( pos_val ) {
		  len_tmp = snprintf( nb, 11, "%u", a );
		  dest.Pos = pos + pos_val;
		}
		else {
		  dest.Pos = pos + pos_opt + 8; /* "position" */
		  pos_val = pos_opt + 8;
		  len_tmp = 0;
		  if( app->ConfigSeparator && app->ConfigSeparator != ' ' ) {
		    nb[len_tmp++] = ' ';
		    nb[len_tmp++] = app->ConfigSeparator;
		  }
		  nb[len_tmp++] = ' ';
		  len_tmp += snprintf( &nb[len_tmp], 11, "%u", a );
		}
		Replace_content( nb, len_tmp, end_opt - pos_val, &dest );
	      }
	      else {
		len_tmp = snprintf( nb, 11, "%u", a );
		dest.Mem = app->Gui.File.Mem;
		dest.Length = app->Gui.File.Length;
		dest.Pos = end;
		Insert_option( "position", 8, nb, len_tmp, 0, &dest );
	      }
	      app->Gui.File.Mem = dest.Mem;
	      app->Gui.File.Length = dest.Length;
	    }
	    app->Gui.WatcherTable[a]->SavePos = a;
	    ++count;
	    break;
	  }
	}
      }
    }
  }
}

/*
  Purpose: Generating a section's name
*/
static _Bool           Generate_section_name(s_dest *dest, const char *name)
{
  unsigned int         length;

  ASSERT( !name, "name = 0" );
  length = strlen( name );

  dest->Pos = 0;
  dest->Length = STEP_BUFFER_FILE; /* small opti to avoid some realloc() */
  dest->Length += length + 3; /* '[' + ']' + '\n' */

  if( !( dest->Mem = malloc( dest->Length )))
    return( 1 );

  dest->Mem[dest->Pos++] = '[';
  memcpy( &dest->Mem[dest->Pos], name, length );
  dest->Pos += length;
  dest->Mem[dest->Pos++] = ']';
  dest->Mem[dest->Pos++] = '\n';
  return( 0 );
}

/*
  Purpose: Compare the current cache folder and the default one
*/
static _Bool            Diff_default_cache_folder()
{
  char                  *defaut;

  if( !( defaut = Get_default_cache_folder_name() ))
    return( 1 );

  if( strcmp( defaut, app->CacheFolder )) {
    free( defaut );
    return( 1 );
  }

  free( defaut );
  return( 0 );
}

/*
  Purpose: Generating a section from the settings
*/
_Bool                   Generate_preferences_section(s_gui *gui, s_dest *dest)
{
  char                  nb[24];

  if( Generate_section_name( dest, "preferences" ))
    goto error;

  if( gui->MaxXWatcher != DEFAULT_COUNT_WIDTH ) {
    if( Add_raw_buffer( dest, "max_width ", 10 ) || Add_number( dest, gui->MaxXWatcher ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( gui->MaxYWatcher != DEFAULT_COUNT_HEIGHT ) {
    if( Add_raw_buffer( dest, "max_height ", 11 ) || Add_number( dest, gui->MaxYWatcher ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( gui->Flags & GUI_FLG_VERTICAL_EXPAND )
    if( Add_raw_buffer( dest, "vertical_expansion yes\n", 23 ))
      goto error;

  if( gui->Flags & GUI_FLG_LOCK_MAIN )
    if( Add_raw_buffer( dest, "lock_position yes\n", 18 ))
      goto error;

  if( !( gui->Flags & GUI_FLG_ON_TOP ))
    if( Add_raw_buffer( dest, "always_on bottom\n", 17 ))
      goto error;

#if GTK_MAJOR_VERSION >= 3
  if( gui->MainColor.red != DEFAULT_COLOR_RED || gui->MainColor.green != DEFAULT_COLOR_GREEN
      || gui->MainColor.blue != DEFAULT_COLOR_BLUE || gui->MainColor.alpha != DEFAULT_COLOR_ALPHA ) {
    unsigned int red, green, blue, alpha;
    red = gui->MainColor.red * 256;
    green = gui->MainColor.green * 256;
    blue = gui->MainColor.blue * 256;
    alpha = gui->MainColor.alpha * 256;
    snprintf( nb, 9, "%.2x%.2x%.2x%.2x", red, green, blue, alpha );
#else
  if( gui->MainColor.red != DEFAULT_COLOR_RED || gui->MainColor.green != DEFAULT_COLOR_GREEN
      || gui->MainColor.blue != DEFAULT_COLOR_BLUE || gui->MainOpacity != DEFAULT_OPACITY ) {
    snprintf( nb, 9, "%.2x%.2x%.2x%.2x", (unsigned int) gui->MainColor.red / 256,
	      (unsigned int) gui->MainColor.green / 256, (unsigned int) gui->MainColor.blue / 256,
	      ( (unsigned int) ( gui->MainOpacity * 255 )));
#endif
    if( Add_raw_buffer( dest, "bg_color ", 9 ) || Add_string( dest, nb, 8 ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  switch( app->Gui.Flags & GUI_FLG_INFOS_MASK ) {
  case GUI_FLG_INFOS_TO_TOP:
    if( Add_raw_buffer( dest, "infosview_expansion top\n", 24 ))
      goto error;
    break;

  case GUI_FLG_INFOS_TO_LEFT:
    if( Add_raw_buffer( dest, "infosview_expansion left\n", 25 ))
      goto error;
    break;

  case GUI_FLG_INFOS_TO_RIGHT:
    if( Add_raw_buffer( dest, "infosview_expansion right\n", 26 ))
      goto error;
    break;
  }

  if( gui->WatcherSizeX != DEFAULT_ICON_WIDTH ) {
    if( Add_raw_buffer( dest, "size_icon ", 10 ) || Add_number( dest, gui->WatcherSizeX ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( !( app->Flags & APP_FLG_SYSTRAY ))
    if( Add_raw_buffer( dest, "systray no\n", 11 ))
      goto error;

  if( app->Flags & APP_FLG_FORCE_PHONE )
    if( Add_raw_buffer( dest, "phone yes\n", 10 ))
      goto error;


  if( Diff_default_cache_folder()) {
    if( Add_raw_buffer( dest, "cache_folder ", 13 ) || Add_string( dest, app->CacheFolder, strlen( app->CacheFolder )))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  snprintf( nb, 24, "%d,%d", app->Gui.MainX, app->Gui.MainY );
  if( Add_raw_buffer( dest, "main_coords ", 12 ) || Add_string( dest, nb, strlen( nb )))
    goto error;
  dest->Mem[dest->Pos++] = '\n';

  if( dest->Pos + 1 >= dest->Length ) {
    char *tmp;
    ++dest->Length;
    if(( tmp = realloc( dest->Mem, dest->Length )) == NULL )
      return( 1 );
    dest->Mem = tmp;
  }
  dest->Mem[dest->Pos++] = '\n';

#ifdef MODE_DEV
  //  printf("Generate_section(): %u / %u\n", dest->Pos, dest->Length );
  //  printf("**** CONTENT **** \n%.*s\n**********\n", dest->Pos, dest->Mem );
#endif
  dest->Length = dest->Pos;
  dest->Pos = 0;
  return( 0 );

 error:
  free( dest->Mem ) ;
  return( 1 );
}

/*
  Purpose: Generating a section from a template
*/
_Bool                   Generate_template_section(s_tpl *template, s_dest *dest)
{
  char                  *name = 0;
  unsigned int          a, length;

  ASSERT( !template, "template = 0" );
  ASSERT( !template->Name, "template->Name = 0" );

  length = strlen( template->Name );
  MALLOC( name, 9 + length + 1 ); /* 'template:' + '\0' */
  memcpy( name, "template:", 9 );
  memcpy( &name[9], template->Name, length );
  name[9 + length] = '\0';

  if( Generate_section_name( dest, name ))
    goto error;
  free( name );

  if( template->HeaderOptions ) {
    if( Add_raw_buffer( dest, "official_http_header ", 21 ) || Add_string( dest, template->HeaderOptions, strlen(template->HeaderOptions) ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  for( a = 0; a < template->OperationCount; ++a )
    if( Add_official_operation( dest, template->OperationTable[a] ))
      goto error;
  dest->Mem[dest->Pos++] = '\n';

  dest->Length = dest->Pos;
  dest->Pos = 0;
  return( 0 );

 error:
  free( name );
  free( dest->Mem ) ;
  return( 1 );
}

/*
  Purpose: Generating a section from a watcher
*/
_Bool                   Generate_watcher_section(s_wtr *watcher, s_dest *dest)
{
  unsigned int          a;

  if( Generate_section_name( dest, watcher->Name ))
    goto error;

  if( Add_raw_buffer( dest, "position ", 9 ) || Add_number( dest, watcher->Pos ))
    goto error;
  dest->Mem[dest->Pos++] = '\n';

  if( watcher->Flags & WTR_FLG_DISABLE )
    if( Add_raw_buffer( dest, "disable yes\n", 12 ))
      goto error;

  if( watcher->ImgPath ) {
    if( Add_raw_buffer( dest, "icon ", 5 ) || Add_string( dest, watcher->ImgPath, strlen(watcher->ImgPath) ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  ASSERT( !watcher->OfficialURL, "[%s]: watcher->OfficialURL = 0", watcher->Name );
  if( Add_raw_buffer( dest, "url ", 4 ) || Add_string( dest, watcher->OfficialURL, strlen(watcher->OfficialURL) ))
    goto error;
  dest->Mem[dest->Pos++] = '\n';

  if( watcher->Flags & WTR_FLG_OFFICIAL_HEADER ) {
    ASSERT( !watcher->OfficialHeader, "[%s]: watcher->OfficialHeader = 0", watcher->Name );
    if( Add_raw_buffer( dest, "official_http_header ", 21 ) || Add_string( dest, watcher->OfficialHeader, strlen(watcher->OfficialHeader) ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( watcher->OfficialTimer != DEFAULT_TIMER ) {
    if( Add_raw_buffer( dest, "official_timer ", 15 ) || Add_number( dest, watcher->OfficialTimer ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( watcher->TemplateName ) {
    if( Add_raw_buffer( dest, "use_template ", 13 ) || Add_string( dest, watcher->TemplateName, strlen( watcher->TemplateName )))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }
  else
    for( a = 0; a < watcher->OfficialOperationCount; ++a )
      if( Add_official_operation( dest, watcher->OfficialOperationTable[a] ))
	goto error;

  if( watcher->Flags & WTR_FLG_LOCAL_AUTO ) {
    if( watcher->Flags & WTR_FLG_OPTIONNAL )
      if( Add_raw_buffer( dest, "optionnal yes\n", 14 ))
	goto error;

    ASSERT( !watcher->LocalExec, "[%s]: watcher->LocalExec = 0", watcher->Name );
    if( Add_raw_buffer( dest, "exec ", 5 ) || Add_string( dest, watcher->LocalExec, strlen(watcher->LocalExec) ))
      goto error;
    dest->Mem[dest->Pos++] = '\n';

    if( watcher->LocalArg ) {
      if( Add_raw_buffer( dest, "arg ", 4 ) || Add_string( dest, watcher->LocalArg, strlen(watcher->LocalArg) ))
	goto error;
      dest->Mem[dest->Pos++] = '\n';
    }

    if( watcher->Flags & WTR_FLG_STDERR ) {
      if( Add_raw_buffer( dest, "stderr yes\n", 11 ))
	goto error;
      dest->Mem[dest->Pos++] = '\n';
    }

    for( a = 0; a < watcher->LocalOperationCount; ++a )
      if( Add_local_operation( dest, watcher->LocalOperationTable[a] ))
	goto error;
  }
  else {
    ASSERT( !watcher->LocalVersion, "[%s]: watcher->LocalVersion = 0", watcher->Name );
    if( Add_raw_buffer( dest, "local_version ", 14 ) || Add_string( dest, watcher->LocalVersion, strlen( watcher->LocalVersion )))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( watcher->Flags & WTR_FLG_TRIGGER_MISMATCH ) {
    ASSERT( !watcher->Trigger_on_mismatch, "[%s]: watcher->Trigger_on_mismatch = 0", watcher->Name );
    if( Add_raw_buffer( dest, "trigger_on_mismatch ", 20 ) || Add_string( dest, watcher->Trigger_on_mismatch, strlen( watcher->Trigger_on_mismatch )))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( watcher->Flags & WTR_FLG_TRIGGER_ERROR ) {
    ASSERT( !watcher->Trigger_on_error, "[%s]: watcher->Trigger_on_error = 0", watcher->Trigger_on_error );
    if( Add_raw_buffer( dest, "trigger_on_error ", 17 ) || Add_string( dest, watcher->Trigger_on_error, strlen( watcher->Trigger_on_error )))
      goto error;
    dest->Mem[dest->Pos++] = '\n';
  }

  if( dest->Pos + 1 >= dest->Length ) {
    char *tmp;
    ++dest->Length;
    if(( tmp = realloc( dest->Mem, dest->Length )) == NULL )
      return( 1 );
    dest->Mem = tmp;
  }
  dest->Mem[dest->Pos++] = '\n';

#ifdef MODE_DEV
  //  printf("Generate_section(): %u / %u\n", dest->Pos, dest->Length );
  //  printf("**** CONTENT **** \n%.*s\n**********\n", dest->Pos, dest->Mem );
#endif
  dest->Length = dest->Pos;
  dest->Pos = 0;
  return( 0 );

 error:
  free( dest->Mem ) ;
  return( 1 );
}

/*
  Purpose: Adding or replacing section into buffer's file
*/
_Bool                   Save_section(s_file *file, s_dest *old_section, s_dest *new_section)
{
  char                  *tmp;
  unsigned int          old_length, new_length;

  old_length = old_section->Length - old_section->Pos;
  new_length = new_section->Length - new_section->Pos;

  if( new_length ) {
    if( old_length ) {/* replacing section */
      if( new_length != old_length ) {
	if( new_length > old_length ) {
	  if(( tmp = realloc( file->Mem, file->Length + new_length - old_length )) == NULL )
	    return( 1 );
	  file->Mem = tmp;
	}
	memmove( &file->Mem[old_section->Pos + new_length], &file->Mem[old_section->Pos + old_length], file->Length - ( old_section->Pos + old_length ));

	file->Length += new_length;
	file->Length -= old_length;
      }
      memcpy( &file->Mem[old_section->Pos], &new_section->Mem[new_section->Pos], new_section->Length - new_section->Pos );
    }
    else { /* adding section */
      if( file->Length ) {
	if( file->Mem[file->Length - 1] != '\n' )
	  ++old_length;
	if( file->Length > 1 && file->Mem[file->Length - 2] != '\n' )
	  ++old_length;
      }
      if(( tmp = realloc( file->Mem, file->Length + new_length )) == NULL )
	return( 1 );
      file->Mem = tmp;
      if( old_length ) {
	file->Mem[file->Length++] = '\n';
	if( old_length > 1 )
	  file->Mem[file->Length++] = '\n';
      }
      memcpy( &file->Mem[file->Length], &new_section->Mem[new_section->Pos], new_length );
      file->Length += new_length;
    }
  }
  else { /* removing section */
    ASSERT( !old_length, "old_length = 0" );
    memmove( &old_section->Mem[old_section->Pos], &old_section->Mem[old_section->Pos + old_length], file->Length - ( old_section->Pos + old_length ));
    file->Length -= old_length;
  }

#ifdef MODE_DEV
  //  printf("**** FILE END (%u) **** \n%.*s\n**********\n", file->Length, file->Length, file->Mem );
#endif
  return( 0 );
}

/*
  Purpose: Saving modifications to config file
*/
void                    *th_Save_modifications_to_file(void *param)
{
  s_file                *file = param;
  _Bool                 write_file = 0;

  ASSERT( !file, "file = 0" );

  LOCK_MUTEX( file->LockFile );
  if( ! Save_section( file, &file->SectionTarget, &file->SectionReplace ))
    write_file = 1;

  if( app->Gui.Save.Flags & SAVE_FLG_WATCHER_ORDER ) {
    Save_watchers_position();
    write_file = 1;
  }

  if( write_file )
    Write_file( file );
  //#todo gui error

  free( file->SectionReplace.Mem );
  bzero( &file->SectionReplace, sizeof(file->SectionReplace) );
  bzero( &file->SectionTarget, sizeof(file->SectionTarget) );

  UNLOCK_MUTEX( file->LockFile );

  bzero( &app->Gui.CancelPref, sizeof(app->Gui.CancelPref) );
  bzero( &app->Gui.Save, sizeof(app->Gui.Save) );
  return( 0 );
}
