/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Positionning the source at the beginning of the line following the line containing the desired pattern
  - source : The content to analyse
  - pattern : Pattern used to count targeted lines
*/
_Bool               Search_after_line(s_src *source, const char *pattern)
{
  unsigned int      a, last_pos, length;

  ASSERT( !source, "source = 0" );
  ASSERT( !pattern, "pattern = 0" );

  length = strlen( pattern );
  for( a = last_pos = source->Pos; a < source->Length; ++a )
    if( source->Mem[a] == '\n' || a + 1 == source->Length ) {
      if( memmem( &source->Mem[last_pos], a - last_pos, pattern, length )) {
        /* Pattern found, going after end of line */
	source->Pos = last_pos + length;
	for( ; source->Pos < source->Length ; ++source->Pos )
	  if( source->Mem[source->Pos] == '\n' ) {
	    ++source->Pos;
	    break;
	  }
        return( 0 );
      }
      last_pos = a + 1;
    }
  /* Pattern not found */
  return( 1 );
}

/*
  Purpose: Positionning the source at the first character of the nth pattern
  - source : The content to analyse
  - target_pattern : The pattern number targeted
  - pattern : Pattern used to count targeted number
*/
unsigned int        Search_after_word(s_src *source, unsigned int target_pattern, const char *pattern)
{
  const char        *last_start, *start = 0;
  unsigned int      count, length, pos;

  ASSERT( !source, "source = 0" );
  ASSERT( !target_pattern, "target_pattern = 0" );
  ASSERT( !pattern, "pattern = 0" );

  length = strlen( pattern );

  last_start = &source->Mem[source->Pos];
  pos = source->Pos;
  for( count = 0; count < target_pattern; ++count ) {
    if(( start = memmem( last_start, source->Length - pos, pattern, length ))) {
      pos += ( start - last_start ) + length;
      last_start = start + length;
    }
    else break;
  }

  if( count != target_pattern )
    /* Pattern not found enough times */
    return( 1 );

  source->Pos = start - source->Mem + length;
  /* Pattern found enough times */
  return( 0 );
}

/*
  Purpose: Positionning the source at the beginning of the nth line, optionnaly containing the pattern
           And shorten the source length to end of line
  - source : The content to analyse
  - target_line : The line number targeted
  - pattern : (optionnal) Pattern used to count targeted lines
*/
unsigned int        Search_in_line(s_src *source, unsigned int target_line, const char *pattern)
{
  unsigned int      a, line, last_pos, length;

  ASSERT( !source, "source = 0" );

  length = pattern ? strlen( pattern ) : 0;

  line = 1;
  for( a = last_pos = source->Pos; a < source->Length; ++a )
    if( source->Mem[a] == '\n' || a + 1 == source->Length ) {
      if( length ) {
        if( ! memmem( &source->Mem[last_pos], a - last_pos, pattern, length )) {
          last_pos = a + 1;
	  /* Pattern not found */
          continue;
        }
	/* Pattern found */
      }

      if( line == target_line ) {
	source->Pos = last_pos;
	source->Length = a;
        return( 0 ); /* Line found */
      }

      ++line;
      last_pos = a + 1;
    }

  /* Line not found */
  return( line );
}

/*
  Purpose: Positionning the source at the first character of the nth field
           And shorten the source length to end of field
  - source : The content to analyse
  - target_field : The field number targeted
  - field_open   : character opening the fields
  - field_close  : closing the fields
*/
unsigned int        Search_in_section(s_src *source, unsigned int target_section, char section_open, char section_close)
{
  unsigned int      a, last_pos, section_pos;

  ASSERT( !source, "source = 0" );
  ASSERT( !target_section, "target_section = 0" );
  ASSERT( !section_open, "section_open = 0" );
  ASSERT( !section_close, "section_close = 0" );

  a = last_pos = source->Pos;

  if( target_section == 1 )
    goto section_found;

  for( section_pos = 1; a < source->Length; ++a ) {
    if( source->Mem[a] == section_open || source->Mem[a] == section_close ) {
      if( section_pos == target_section )
	goto section_found;

      ++section_pos;
      last_pos = a + 1;
    }
  }

  if( section_pos == target_section )
    goto section_found;

  /* Section not found */
  return( section_pos );

 section_found:
  source->Pos = last_pos;
  for( ; a < source->Length; ++a )
    if( source->Mem[a] == section_open || source->Mem[a] == section_close )
      break;
  source->Length = a;
  /* Section found */
  return( 0 );
}

/*
  Purpose: Positionning the source after the nth chosen character in the current line
  - source : The content to analyse
  - target_field : The field number targeted
  - split : The character used to split the line
*/
unsigned int        Search_in_field(s_src *source, unsigned int target_field, char split)
{
  const char        *last_start, *start;
  unsigned int      count, pos, length;

  ASSERT( !source, "source = 0" );
  ASSERT( !target_field, "target_field = 0" );
  ASSERT( !split, "char = 0" );

  for( pos = source->Pos; pos < source->Length ; ++pos )
    if( source->Mem[pos] == '\n' )
      break;
  length = pos;

  last_start = &source->Mem[source->Pos];
  pos = source->Pos;
  for( count = 1; 1; ++count )
    if(( start = memchr( last_start, split, length - pos ))) {
      if( count == target_field ) {
	length = pos + start - last_start;
	break;
      }
      pos += start - last_start + 1;
      last_start = start + 1;
    }
    else break;

  if( count != target_field )
    /* Pattern not found enough times */
    return( count );

  source->Pos = pos;
  source->Length = length;
  /* Pattern found enough times */
  return( 0 );
}

/*
  Purpose: Positionning the source at the nth word in the current line
  - source : The content to analyse
  - target_word : The word number targeted
*/
unsigned int        Search_in_word(s_src *source, unsigned int target_word, _Bool from_end)
{
  unsigned int      count, pos, length;

  ASSERT( !source, "source = 0" );
  ASSERT( !target_word, "target_word = 0" );

  for( pos = source->Pos; pos < source->Length; ++pos )
    if( source->Mem[pos] == '\n' )
      break;
  length = pos;

  count = 1;
  if( from_end ) {
    if( source->Mem[pos] == '\n' ) --pos;
    for( ; pos; --pos ) {
      /* Skipping spaces */
      for( ; pos; --pos )
	if( !( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' ))
	  break;

      if( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' || source->Mem[pos] == '\n' )
	break;

      if( count == target_word ) {
	/* Word found */
	source->Length = pos + 1;
	/* Searching beginning of word */
	for( ; pos; --pos )
	  if( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' || source->Mem[pos] == '\n' )
	    break;
	source->Pos = pos + 1;
	return( 0 );
      }
      ++count;

      /* Skipping word */
      for( ; pos; --pos )
	if( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' || source->Mem[pos] == '\n' )
	  break;
      ++pos;
    }
  }
  else {
    for( pos = source->Pos; pos < length;  ) {
      /* Skipping spaces */
      for( ; pos < length; ++pos )
	if( !( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' ))
	  break;

      if( source->Mem[pos] == '\n' || pos == length )
	break;

      if( count == target_word ) {
	/* Word found */
	source->Pos = pos;
	/* Searching end of word */
	for( ; pos < length; ++pos )
	  if( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' || source->Mem[pos] == '\n' )
	    break;
	source->Length = pos;
	return( 0 );
      }
      ++count;

      /* Skipping word */
      for( ; pos < length; ++pos )
	if( source->Mem[pos] == ' ' || source->Mem[pos] == '\t' || source->Mem[pos] == '\n' )
	  break;
    }
  }

  /* Not enough word found */
  return( count );
}
