/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/* Purpose: Retrieving the local version, the last cached official version, and creating the XOR buffer */
static void        Initialize_watcher_info(unsigned int id)
{
  char             *file;
  s_wtr            *watcher;

  for( ; id < app->Gui.WatcherCount; id += app->CoreCount ) {
    watcher = app->Gui.WatcherTable[id];
    if( watcher && !( watcher->Flags & WTR_FLG_DISABLE )) {
      watcher->Flags |= WTR_FLG_INIT;
      Retrieve_local_version( watcher );
      file = Build_cachefile_name( watcher->Name, watcher->NameLength );
      Retrieve_official_version_from_cache( watcher, file, 1 );
      free( file );
      Build_xor( watcher, watcher->Trigger_xor );
      watcher->Flags |= WTR_FLG_OFFICIAL_IN_PROGRESS;
      Update_watcher_area( watcher, 1 );
      watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE );
    }
  }
}

/* Purpose: Handeling the reload of the configuration file asked by the user */
static void        Release_reload_resources(unsigned int id)
{
  unsigned int     a;

  LOCK_MUTEX( app->Gui.LockWatchers );
  app->ThreadFlags[id] &= ~( THD_FLG_RELOAD );
  for( a = 0; a < app->CoreCount; ++a )
    if( app->ThreadFlags[a] & THD_FLG_RELOAD )
      break;
  UNLOCK_MUTEX( app->Gui.LockWatchers );

  /* The last thread will release the old watchers resources and reactivate the reload button */
  if( a == app->CoreCount ) {
    for( a = 0; a < app->CountReload; ++a )
      Free_struct_watcher( app->ReloadTable[a], 1, 1 );
    free( app->ReloadTable );
    app->ReloadTable = 0;
    LOCK_GTK();
    gtk_widget_set_sensitive( app->Gui.PreferenceSettings[PS_BUT_RELOAD], 1 );
#if GTK_MAJOR_VERSION < 3
    gdk_flush();
#endif
    UNLOCK_GTK();
  }
}

static void        Print_error()
{
  int              length;
  unsigned int     a;
  char             tmp[2048];

  length = snprintf( tmp, 2048, "  %s  \n  %s\n\n", app->Strs[218], app->Gui.File.Name ); /* "Some errors have been found in the configuration file" */
  for( a = 0; a < app->Gui.File.ErrorsCount; ++a )
    if( length > 0 && length < 2048 )
      length += snprintf( &tmp[length], 2048 - length, "  - %s \n", app->Gui.File.Error[a] );

  if( length > 0 ) {
    LOCK_GTK();
    /* if( !( app->Flags & APP_FLG_PHONE )) */
    /*   gtk_widget_set_size_request( app->Error[ERW_WIN_MAIN], 10, 10 ); */
    gtk_label_set_label( GTK_LABEL(app->Error[ERW_LBL_MESSAGE]), tmp );
    gtk_widget_show( app->Error[ERW_WIN_MAIN] );
    UNLOCK_GTK();
  }
}

void               *th_Updater(void *param)
{
  s_wtr            *watcher;
  unsigned int     a, id;

  usleep( 750000 ); /* kind of ugly, we MUST let some time for main gtk_main() to start */

  for( a = 0; a < app->CoreCount; ++a )
    if( &app->IdWorkers[a] == param )
      break;
  id = a;

  /* here, the first thread will print the file loading errors */
  if( a == app->CoreCount - 1 && app->Gui.File.ErrorsCount )
    Print_error();

  Initialize_watcher_info( id );

  while( 1 ) {
    while( 1 ) {
      /* here, we check the timer of the watcher, and if it ran out, we update the informations */
      for( a = id; a < app->Gui.WatcherCount; a += app->CoreCount ) {
	watcher = app->Gui.WatcherTable[a];
	if( app->ThreadFlags[id] & THD_FLG_RELOAD )
	  break;
	if( watcher && !( watcher->Flags & WTR_FLG_DISABLE ))
	  Refresh_official_version( watcher );
      }

      /* here, we wait the next cond_broadcast.
       Spurious wakeups are not handled as they are not a problem */
      LOCK_MUTEX( app->LockCond );
      if( !( app->ThreadFlags[id] & THD_FLG_RELOAD ))
	pthread_cond_wait( &app->TimerCond, &app->LockCond );

      if( app->ThreadFlags[id] & THD_FLG_RELOAD ) {
	Release_reload_resources( id );
	break;
      }
      UNLOCK_MUTEX( app->LockCond );
    }
    UNLOCK_MUTEX( app->LockCond );

    Initialize_watcher_info( id );
  }
}
