/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Checking if the source is empty
*/
static _Bool        Check_source_is_empty(s_src *source)
{
  unsigned int      a;

  for( a = source->Pos; a < source->Length; ++a )
    if( !( source->Mem[a] == ' ' || source->Mem[a] == '\t' || source->Mem[a] == '\r' || source->Mem[a] == '\n' ))
      break;

  if( a == source->Length )
    return( 1 );
  return( 0 );
}

/*
  Purpose: Applying operation to source
*/
_Bool               Apply_search_operation(s_so *search, s_src *source, char *error)
{
  const char        *key_name = 0;
  s_op              *operation, **table;
  unsigned int      a, max_found, count;

  count = search->OperationCount;
  table = search->OperationTable;
  if( search->TemplateName ) {
    for( a = 0; a < app->Gui.File.TemplateCount; ++a )
      if( ! strcmp( search->TemplateName, app->Gui.File.TemplateTable[a]->Name )) {
        count = app->Gui.File.TemplateTable[a]->OperationCount;
        table = app->Gui.File.TemplateTable[a]->OperationTable;
        break;
      }
    if( a == app->Gui.File.TemplateCount ) {
      snprintf( error, SIZE_BUFFER_ERROR, "%s: '%s'",
                app->Strs[249], search->TemplateName ); /* "Template not found" */
      return( 1 );
    }
  }

  ASSERT( !table && count, "table = 0" );
  ASSERT_NON_FATAL( table && !count, "count = 0" );

  for( a = 0; a < count; ++a ) {
    operation = table[a];
    switch( operation->Type ) {
    case OP_TYPE_AFTER_LINE:
      if( ! operation->Pattern )
        continue;
      key_name = "after_line";
      if( Search_after_line( source, operation->Pattern )) {
        snprintf( error, SIZE_BUFFER_ERROR, "%s: %s: '%s'", key_name,
                  app->Strs[212], operation->Pattern ); /* "Pattern not found in source" */
        return( 1 );
      }
      break;

    case OP_TYPE_AFTER_WORD:
      if( ! operation->Pattern )
        continue;
      key_name = "after_word";
      max_found = Search_after_word( source, operation->Number, operation->Pattern );
      if( max_found ) {
        snprintf( error, SIZE_BUFFER_ERROR, "%s: %u / %u: %s: '%s'", key_name, max_found,
                  operation->Number, app->Strs[212], operation->Pattern ); /* "Pattern not found in source" */
        return( 1 );
      }
      break;

    case OP_TYPE_IN_LINE:
      key_name = "in_line";
      max_found = Search_in_line( source, operation->Number, operation->Activated ? operation->Pattern : 0 );
      if( max_found ) {
        if( operation->Pattern ) {
          --max_found;
          snprintf( error, SIZE_BUFFER_ERROR, "%s: %u / %u: %s: '%s'", key_name, max_found,
                    operation->Number, app->Strs[230], /* "Not enough lines in source matching the pattern" */
                    operation->Pattern );
        }
        else {
          snprintf( error, SIZE_BUFFER_ERROR, "%s: %s: %u / %u", key_name,
                    app->Strs[229], max_found, operation->Number ); /* "Not enough lines in source" */
        }
        return( 1 );
      }
      break;

    case OP_TYPE_IN_SECTION:
      if( ! operation->Delimiter[0] || ! operation->Delimiter[1] )
        continue;
      key_name = "in_section";
      max_found = Search_in_section( source, operation->Number, operation->Delimiter[0], operation->Delimiter[1] );
      if( max_found ) {
        snprintf( error, SIZE_BUFFER_ERROR, "%s: %s: %u / %u", key_name,
                  app->Strs[231], max_found, operation->Number ); /* "Target field not found" */
        return( 1 );
      }
      break;

    case OP_TYPE_IN_FIELD:
      if( ! operation->Delimiter[0] )
        continue;
      key_name = "in_field";
      max_found = Search_in_field( source, operation->Number, operation->Delimiter[0] );
      if( max_found ) {
        snprintf( error, SIZE_BUFFER_ERROR, "%s: %s: %u / %u", key_name,
                  app->Strs[231], max_found, operation->Number ); /* "Target field not found" */
        return( 1 );
      }
      break;

    case OP_TYPE_IN_WORD:
      key_name = "in_word";
      max_found = Search_in_word( source, operation->Number, operation->Activated );
      if( max_found ) {
        snprintf( error, SIZE_BUFFER_ERROR, "%s: %s: %u / %u", key_name,
                  app->Strs[61], max_found - 1, operation->Number ); /* "Not enough words" */
        return( 1 );
      }
      break;

    default:
      ASSERT_NON_FATAL( 1, "Unknown operation->Type = %u", operation->Type );
    }
    if( Check_source_is_empty( source )) {
      snprintf( error, SIZE_BUFFER_ERROR, "%s: %s", key_name,
                app->Strs[233] ); /* "Version found is empty" */
      return( 1 );
    }
  }

  return( 0 );
}

/*
  Purpose: Deleting a search operation to the temporary watcher
*/
static void              Remove_operation_from_watcher_settings_window(GtkWidget *hbox)
{
  s_gui                  *gui = &app->Gui;
  s_wtr                  *watcher = &gui->TmpWatcher;
  s_op                   *operation = 0;
  unsigned int           a;

  for( a = 0; a < gui->File.TemplateCount; ++a )
    if( gui->File.TemplateTable[a]->MainBox == hbox ) {
      Empty_search_block_list( gui );
      gtk_widget_show( app->Gui.WatcherSettings[WS_ADD_BLOCK_OFFICIAL] );
      return;
    }

  for( a = 0; a < watcher->OfficialOperationCount; ++a ) {
    operation = watcher->OfficialOperationTable[a];
    if( operation->GtkTable[0] == hbox ) {
      if( operation->Temporary )
        Free_struct_operation( operation, 0 );
      else
        gtk_container_remove( GTK_CONTAINER(gui->WatcherSettings[WS_VBOX_SEARCH_OFFI]), hbox );

      for( ; a + 1 < watcher->OfficialOperationCount; ++a )
	watcher->OfficialOperationTable[a] = watcher->OfficialOperationTable[a + 1];

      if( watcher->OfficialOperationCount == MAX_COUNT_OPERATION )
	gtk_widget_show( gui->WatcherSettings[WS_ADD_BLOCK_OFFICIAL] );

      --watcher->OfficialOperationCount;
      if( ! watcher->OfficialOperationCount ) {
	free(  watcher->OfficialOperationTable );
	watcher->OfficialOperationTable = 0;
      }
      else {
	REALLOC( watcher->OfficialOperationTable, watcher->OfficialOperationCount * sizeof(*watcher->OfficialOperationTable) );
      }
      Update_official_result();
      return;
    }
  }

  for( a = 0; a < watcher->LocalOperationCount; ++a ) {
    operation = watcher->LocalOperationTable[a];
    if( operation->GtkTable[0] == hbox ) {
      if( operation->Temporary )
        Free_struct_operation( operation, 0 );
      else
        gtk_container_remove( GTK_CONTAINER(gui->WatcherSettings[WS_VBOX_SEARCH_LOCAL]), hbox );

      for( ; a + 1 < watcher->LocalOperationCount; ++a )
	watcher->LocalOperationTable[a] = watcher->LocalOperationTable[a + 1];

      if( watcher->LocalOperationCount == MAX_COUNT_OPERATION )
	gtk_widget_show( gui->WatcherSettings[WS_ADD_BLOCK_LOCAL] );

      --watcher->LocalOperationCount;
      if( ! watcher->LocalOperationCount ) {
	free(  watcher->LocalOperationTable );
	watcher->LocalOperationTable = 0;
      }
      else {
	REALLOC( watcher->LocalOperationTable, watcher->LocalOperationCount * sizeof(*watcher->LocalOperationTable) );
      }
      return;
    }
  }
  ASSERT( 1, "hbox non trouvé" );
}

/*
  Purpose: Deleting a search operation from the template settings window
*/
static void              Remove_operation_from_template_settings_window(GtkWidget *hbox)
{
  s_gui                  *gui = &app->Gui;
  s_op                   *operation;
  s_tpl                  *template = 0;
  unsigned int           a;

  for( a = 0; a < gui->File.TemplateCount; ++a )
    if( gui->File.TemplateTable[a]->MainBox == hbox ) {
      template = gui->File.TemplateTable[a];
      break;
    }

  if( ! template )
    template = &gui->TmpTemplate;

  for( a = 0; a < template->OperationCount; ++a ) {
    operation = template->OperationTable[a];
    if( operation->GtkTable[0] == hbox ) {
      if( operation->Temporary )
        Free_struct_operation( operation, 0 );
      else
        gtk_container_remove( GTK_CONTAINER(gui->TemplateSettings[TS_VBOX_OPERATION]), hbox );

      if( template->OperationCount == MAX_COUNT_OPERATION )
	gtk_widget_show( gui->TemplateSettings[TS_ADD_BLOCK] );

      for( ; a + 1 < template->OperationCount; ++a )
        template->OperationTable[a] = template->OperationTable[a + 1];
      --template->OperationCount;
      if( ! template->OperationCount ) {
        free( template->OperationTable );
        template->OperationTable = 0;
      }
      else {
        REALLOC( template->OperationTable, template->OperationCount * sizeof(*template->OperationTable) );
      }
      return;
    }
  }
  ASSERT( 1, "hbox not found" );
}

/*
  Purpose: Callback for '-' button of the operation blocks
*/
static void              Callback_del_block(GtkWidget *button, GtkWidget *hbox_block)
{
  s_gui                  *gui = &app->Gui;
  GList                  *list;

  /* Suppress from template settings window */
  list = gtk_container_get_children( GTK_CONTAINER(gui->TemplateSettings[TS_VBOX_OPERATION]) );
  for(; list != NULL; list = list->next )
    if( list->data == hbox_block ) {
      g_list_free( list );
      Remove_operation_from_template_settings_window( hbox_block );
      Update_template_result();
      return;
    }
  if( list )
      g_list_free( list );

  Remove_operation_from_watcher_settings_window( hbox_block );
}

/*
  Purpose: Creating a delete buttion ('-') and link it to the container
*/
void                     Create_del_button(GtkWidget *parent)
{
  GtkWidget              *item, *img;

  if( app->Gui.ImgTable[IMG_DEL] ) {
    item = gtk_button_new();
    img = gtk_image_new_from_pixbuf( app->Gui.ImgTable[IMG_DEL] );
    gtk_button_set_image( GTK_BUTTON(item), img );
  }
  else
    item = gtk_button_new_with_mnemonic( " - " );

  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_del_block), parent );
  gtk_box_pack_start( GTK_BOX(parent), item, FALSE, FALSE, 0 );
}

/*
  Purpose: Callback for '+' button in the 'Add template' window
*/
static void         Callback_add_template(GtkWidget *widget, gpointer param)
{
  s_tpl             *template = param;
  s_gui             *gui = &app->Gui;

  Empty_search_block_list( gui );

  ASSERT( !template, "template = 0" );
  ASSERT( !template->Name, "template->Name = 0" );
  free( gui->TmpWatcher.TemplateName );
  STRDUP( gui->TmpWatcher.TemplateName, template->Name );

  gtk_container_add( GTK_CONTAINER(gui->WatcherSettings[WS_VBOX_SEARCH_OFFI]), template->MainBox );
  gtk_widget_show_all( gui->WatcherSettings[WS_VBOX_SEARCH_OFFI] );
  gtk_widget_hide( app->Gui.WatcherSettings[WS_ADD_BLOCK_OFFICIAL] );

  Update_official_result();
}

/*
  Purpose: Opening and mapping a file
*/
unsigned int        Open_file(const char *file, s_chf *dest)
{
  char              *map = 0;
  unsigned int      err = 0;
  struct stat       st;
  int               fd;

  bzero( dest, sizeof(*dest) );
  if(( fd = open( file, O_RDONLY )) < 0 ) {
    err = 143; /* app->Strs[143] "The cache file is not available" */
    ERR_SYS( "open( %s ): %s", file, app->Strs[err] );
    return( err );
  }
  if( fstat( fd, &st ) < 0 ) {
    err = 127; /* app->Strs[127] "Unable to evaluate the cache file" */
    ERR_SYS( "fstat( %s ); %s", file, app->Strs[err] ) ;
    goto error;
  }

  dest->LastUpdate = st.st_ctim.tv_sec;

  if( st.st_size > UINT_MAX ) {
    err = 70; /* app->Strs[70] "Sorry file is too big for me" */
    ERR_APP( "%s: %s", app->Strs[err], file );
    goto error;
  }
  if( ! st.st_size ) {
    err = 142; /* app->Strs[142] "The downloaded file is empty" */
    ERR_APP( "%s: %s", app->Strs[err], file );
    goto error;
  }
  if(( map = mmap( NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0 )) == MAP_FAILED ) {
    err = 2; /* app->Strs[2] "Unable to open file" */
    ERR_SYS( "mmap( %s ): %s", file, app->Strs[err] );
    goto error;
  }

  dest->Source.Mem = map;
  dest->Source.Length = (unsigned int) st.st_size;

 error:
  close( fd );
  return( err );
}

/*
  Purpose: Normalizing the version found in source
*/
_Bool                Normalize_version(s_src *source, char *buffer)
{
  unsigned int       a;
  _Bool              squote = 0, dquote = 0;

  if( ! source->Length )
    goto error;

  ASSERT( source->Pos > source->Length, "source->Pos(%u) > source->Length(%u)", source->Pos, source->Length );
  for( ; source->Pos < source->Length; ++source->Pos )
    if( source->Mem[source->Pos] != ' ' && source->Mem[source->Pos] != '\t' )
      break;

  if( source->Pos >= source->Length )
    goto error;

  source->Mem = &source->Mem[source->Pos];
  source->Length -= source->Pos;
  source->Pos = 0;

  for( a = 0; a < source->Length; ++a ) {
    if( source->Mem[a] == '\n' || source->Mem[a] == '\r' ) {
      source->Length = a;
      break;
    }
    if( source->Mem[a] == '\'' ) {
      if( dquote ) {
	source->Length = a;
	break;
      }
      squote = 1;
    }
    else if( source->Mem[a] == '"' ) {
      if( squote ) {
	source->Length = a;
	break;
      }
      dquote = 1;
    }
  }

  if( source->Length > MAX_LENGTH_VERSION ) {
    memcpy( buffer, source->Mem, MAX_LENGTH_VERSION );
    buffer[MAX_LENGTH_VERSION] = '.';
    buffer[MAX_LENGTH_VERSION + 1] = '.';
    buffer[MAX_LENGTH_VERSION + 2] = '.';
    buffer[MAX_LENGTH_VERSION + 3] = '\0';
  }
  else {
    memcpy( buffer, source->Mem, source->Length );
    buffer[source->Length] = '\0';
  }

  return( 0 );

 error:
  snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[233] ); /* "Version found is empty" */
  return( 1 );
}

/*
  Purpose: Updating the 'Official result' entry in the watcher preferences window
*/
void                     *th_Update_from_search(void *param)
{
  s_so                   *search = param;
  s_chf                  cache;
  unsigned int           err;
  char                   buffer[SIZE_BUFFER_ERROR];

  ASSERT( !search, "search = 0" );
  ASSERT( !search->CacheFile, "search->CacheFile = 0" );

  if( search->Url )
    if(( err = Download_resource_to_cache( search->Url, search->Header, search->CacheFile ))) {
      snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[err] );
      goto error;
    }

  if(( err = Open_file( search->CacheFile, &cache ))) {
    snprintf( buffer, SIZE_BUFFER_ERROR, "%s", app->Strs[err] );
    goto error;
  }

  if( Apply_search_operation( search, &cache.Source, buffer )) {
    if( search->Watcher && search->Watcher->Name ) {
      WARN_APP( "%s: %s", search->Watcher->Name, buffer );
    }
    else {
      WARN_APP( "%s", buffer );
    }
    goto error;
  }

  if( Normalize_version( &cache.Source, buffer ))
    goto error;

  if( search->Watcher ) {
    free( search->Watcher->OfficialVersion );
    STRDUP( search->Watcher->OfficialVersion, buffer );
  }

 error:
  if( search->Result ) {
    LOCK_GTK();
    gtk_entry_set_text( search->Result, buffer );
    UNLOCK_GTK();
  }
  Free_struct_search( search, 1 );
  return( 0 );
}

/*
  Purpose: Copying the operation table to the search structure
*/
s_so                    *Operations_to_search(s_op **table, unsigned char count)
{
  s_so                  *search;
  s_op                   *operation;
  unsigned int           a;

  CALLOC( search, 1, sizeof(*search) );

  if( ! count )
    return( search );

  MALLOC( search->OperationTable, count * sizeof(*search->OperationTable) );
  search->OperationCount = count;

  for( a = 0; a < count; ++a ) {
    CALLOC( operation, 1, sizeof(*operation) );
    operation->Type = table[a]->Type;
    operation->GtkTable = table[a]->GtkTable;
    switch( operation->Type ) {
    case OP_TYPE_AFTER_LINE: Copy_gui_after_line( operation ); break;
    case OP_TYPE_AFTER_WORD: Copy_gui_after_word( operation ); break;
    case OP_TYPE_IN_FIELD: Copy_gui_in_field( operation ); break;
    case OP_TYPE_IN_LINE: Copy_gui_in_line( operation ); break;
    case OP_TYPE_IN_SECTION: Copy_gui_in_section( operation ); break;
    case OP_TYPE_IN_WORD: Copy_gui_in_word( operation ); break;
    default:
      ASSERT( 1, "Unknown Type = %u", operation->Type );
      break;
    }
    operation->GtkTable = 0;
    search->OperationTable[a] = operation;
  }
  return( search );
}

/*
  Purpose: Updating the result in the template window
*/
void                     Update_template_result()
{
  s_tpl                  *template = &app->Gui.TmpTemplate;
  s_so                   *search;
  pthread_t              tmp_id;

  search = Operations_to_search( template->OperationTable, template->OperationCount );
  if( ! search )
    return;

  search->CacheFile = Build_cachefile_name( "tmp_template_result", 19 );
  search->Result = GTK_ENTRY(app->Gui.TemplateSettings[TS_ETY_RESULT]);

  if( pthread_create( &tmp_id, &app->Attr, th_Update_from_search, search )) {
    ERR_APP( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    Free_struct_search( search, 0 );
  }
}

/*
  Purpose: Callback for '+' button in the 'Add watcher' window
*/
static void         Callback_add_block(GtkWidget *widget, gpointer data)
{
  uint16_t          block;
  s_op              *operation = 0;
  s_gui             *gui = &app->Gui;

  memcpy( &block, &data, sizeof(block) );
  if( block & CB_BLOCK_AFTER_LINE ) operation = New_operation_after_line( 0 );
  else if( block & CB_BLOCK_AFTER_WORD ) operation = New_operation_after_word( 1, 0 );
  else if( block & CB_BLOCK_IN_FIELD ) operation = New_operation_in_field( 1, 0 );
  else if( block & CB_BLOCK_IN_LINE ) operation = New_operation_in_line( 1, 0, 0 );
  else if( block & CB_BLOCK_IN_SECTION ) operation = New_operation_in_section( 1, '<', '>' );
  else if( block & CB_BLOCK_IN_WORD ) operation = New_operation_in_word( 1, 0 );
  else {
    ASSERT( 1, "Unknown block operation = %u", block );
  }
  operation->Temporary = 1;

  if( block & CB_OFFICIAL_VERSION ) {
    Add_official_operation_to_watcher( &gui->TmpWatcher, operation );

    gtk_container_add( GTK_CONTAINER(gui->WatcherSettings[WS_VBOX_SEARCH_OFFI]), operation->GtkTable[0] );
    gtk_widget_show_all( gui->WatcherSettings[WS_VBOX_SEARCH_OFFI] );

    if( gui->TmpWatcher.OfficialOperationCount == MAX_COUNT_OPERATION )
      gtk_widget_hide( gui->WatcherSettings[WS_ADD_BLOCK_OFFICIAL] );

    Update_official_result();
  }
  else if( block & CB_TEMPLATE_CREATE ) {
    Add_operation_to_template( &gui->TmpTemplate, operation );

    gtk_container_add( GTK_CONTAINER(gui->TemplateSettings[TS_VBOX_OPERATION]), operation->GtkTable[0] );
    gtk_widget_show_all( gui->TemplateSettings[TS_VBOX_OPERATION] );

    if( gui->TmpTemplate.OperationCount == MAX_COUNT_OPERATION )
      gtk_widget_hide( gui->TemplateSettings[TS_ADD_BLOCK] );

    Update_template_result();
  }
  else if( block & CB_LOCAL_VERSION ) {
    Add_local_operation_to_watcher( &gui->TmpWatcher, operation );

    gtk_container_add( GTK_CONTAINER(gui->WatcherSettings[WS_VBOX_SEARCH_LOCAL]), operation->GtkTable[0] );
    gtk_widget_show_all( gui->WatcherSettings[WS_VBOX_SEARCH_LOCAL] );

    if( gui->TmpWatcher.LocalOperationCount == MAX_COUNT_OPERATION )
      gtk_widget_hide( gui->WatcherSettings[WS_ADD_BLOCK_LOCAL] );
  }
  else {
    ASSERT( 1, "Unknown block target = %u", block );
  }
}

/*
  Purpose: Adding a template into a menu list
*/
void                     Add_template_to_menu(GtkWidget *menu, s_tpl *template)
{
  GtkWidget              *item;

  ASSERT( template->MenuItem, "template->MenuItem != 0" );

#ifndef LIB_HANDY
  //#if GTK_MAJOR_VERSION == 2 || ( GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION < 10 )
  GtkWidget              *img;
  item = gtk_image_menu_item_new_with_label( template->Name );
  if( app->Gui.ImgTable[IMG_ADD] ) {
    img = gtk_image_new_from_pixbuf( app->Gui.ImgTable[IMG_ADD] );
    gtk_image_menu_item_set_image( GTK_IMAGE_MENU_ITEM(item), img );
  }
  else
#endif
    item = gtk_menu_item_new_with_label( template->Name );

  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_template), template );

  template->MenuItem = item;
}

void                     Create_menu_search_block(GtkWidget *parent_box, uint8_t block)
{
  GtkWidget              *menu, *item, *img;
  gpointer               data;
  unsigned int           a;
  uint16_t               flags;

#  define APP_NEW_MENU_ITEM( label )					\
  if( app->Gui.ImgTable[IMG_ADD] ) {					\
    item = gtk_image_menu_item_new_with_mnemonic( label );		\
    img = gtk_image_new_from_pixbuf( app->Gui.ImgTable[IMG_ADD] );	\
    gtk_image_menu_item_set_image( GTK_IMAGE_MENU_ITEM(item), img );	\
  } else								\
    item = gtk_menu_item_new_with_label( label )

  /* Add search block ------------------------------------------------------- */
  menu = gtk_menu_new();
  APP_NEW_MENU_ITEM( app->Strs[234] ); /* "Add search block" */
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu );
  gtk_menu_shell_append(GTK_MENU_SHELL(parent_box), item );

  APP_NEW_MENU_ITEM( app->Strs[209] ); /* " After line containing " */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item );
  flags = block | CB_BLOCK_AFTER_LINE;
  memcpy( &data, &flags, sizeof(flags) );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_block), data );

  APP_NEW_MENU_ITEM( app->Strs[239] ); /* " After word " */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item );
  flags = block | CB_BLOCK_AFTER_WORD;
  memcpy( &data, &flags, sizeof(flags) );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_block), data );

  APP_NEW_MENU_ITEM( app->Strs[108] ); /* " In line number " */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item );
  flags = block | CB_BLOCK_IN_LINE;
  memcpy( &data, &flags, sizeof(flags) );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_block), data );

  APP_NEW_MENU_ITEM( app->Strs[131] ); /* " In field number " */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item );
  flags = block | CB_BLOCK_IN_FIELD;
  memcpy( &data, &flags, sizeof(flags) );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_block), data );

  APP_NEW_MENU_ITEM( app->Strs[174] ); /* " In section number " */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item );
  flags = block | CB_BLOCK_IN_SECTION;
  memcpy( &data, &flags, sizeof(flags) );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_block), data );

  APP_NEW_MENU_ITEM( app->Strs[109] ); /* " In word number " */
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), item );
  flags = block | CB_BLOCK_IN_WORD;
  memcpy( &data, &flags, sizeof(flags) );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_add_block), data );

  if( !( block & CB_OFFICIAL_VERSION ))
    return;

  /* Use a template -------------------------------------------------------- */
  menu = gtk_menu_new();
  app->Gui.WatcherSettings[WS_MENU_TEMPLATE] = menu;

  APP_NEW_MENU_ITEM( app->Strs[247] ); /* "Use a template" */
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu );
  gtk_menu_shell_append(GTK_MENU_SHELL(parent_box), item );
  app->Gui.WatcherSettings[WS_MENU_ITEM_TEMPLATE] = item;

  if( app->Gui.File.TemplateCount ) {
    for( a = 0; a < app->Gui.File.TemplateCount; ++a )
      Add_template_to_menu( menu, app->Gui.File.TemplateTable[a] );
  }
  else
    gtk_widget_set_sensitive( item, 0 );
}
