/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Callback for 'Cancel' button of the template settings window
*/
static void              Callback_cancel(GtkWidget *widget, s_gui *gui)
{
  gtk_widget_hide( gui->TemplateSettings[TS_WIN_MAIN] );
  gui->Flags &= ~( GUI_FLG_TEMPLATE_MODIFY );
}

/*
  Purpose: Moving list of operations in the template
*/
static _Bool            Gui_to_template_operations(s_gui *gui)
{
  s_op                  *operation;
  unsigned int          a;
  unsigned char         ret;

  if( ! gui->TmpTemplate.OperationCount ) {
    gtk_label_set_label( GTK_LABEL(gui->TemplateSettings[TS_LBL_ERROR]), app->Strs[257] ); /* "Missing search block" */
    return( 1 );
  }

  for( a = 0; a < gui->TmpTemplate.OperationCount; ++a ) {
    operation = gui->TmpTemplate.OperationTable[a];
    ret = 0;
    switch( operation->Type ) {
    case OP_TYPE_AFTER_LINE: ret = Get_gui_after_line( operation, gui->TemplateSettings[TS_LBL_ERROR] ); break;
    case OP_TYPE_AFTER_WORD: ret = Get_gui_after_word( operation, gui->TemplateSettings[TS_LBL_ERROR] ); break;
    case OP_TYPE_IN_FIELD:   ret = Get_gui_in_field( operation, gui->TemplateSettings[TS_LBL_ERROR] ); break;
    case OP_TYPE_IN_LINE:    ret = Get_gui_in_line( operation, gui->TemplateSettings[TS_LBL_ERROR] ); break;
    case OP_TYPE_IN_SECTION: ret = Get_gui_in_section( operation, gui->TemplateSettings[TS_LBL_ERROR] ); break;
    case OP_TYPE_IN_WORD:    ret = Get_gui_in_word( operation, gui->TemplateSettings[TS_LBL_ERROR] ); break;
    }
    if( ret == 1 ) return( 1 );
  }

  return( 0 );
}

/*
  Purpose: Retrieving informations of common properties from GUI and put it into template
*/
static _Bool            Gui_to_template_properties(s_gui *gui, s_tpl *template)
{
  const char            *entry;
  unsigned int          a;

  /* Name --------- */
  free( (char*) gui->TmpTemplate.Name );
  gui->TmpTemplate.Name = 0;

  if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->TemplateSettings[TS_ETY_NAME]) )) {
    gtk_label_set_label( GTK_LABEL(gui->TemplateSettings[TS_LBL_ERROR]), app->Strs[186] ); /* "Missing name" */
    return( 1 );
  }

  if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->TemplateSettings[TS_ETY_NAME]) ))) {
    gtk_label_set_label( GTK_LABEL(gui->TemplateSettings[TS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
    return( 1 );
  }

  for( a = 0; a < gui->File.TemplateCount; ++a)
    if( ! strcmp( entry, gui->File.TemplateTable[a]->Name ))
      break;

  if( ! template || ! template->Name || strcmp( entry, template->Name )) {
    if( a != gui->File.TemplateCount ) { /* modified name already exist */
      gtk_label_set_label( GTK_LABEL(gui->TemplateSettings[TS_LBL_ERROR]), app->Strs[97] ); /* "This name is already used" */
      return( 1 );
    }
    if( !( entry = strdup( entry ))) {
      gtk_label_set_label( GTK_LABEL(gui->TemplateSettings[TS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
      return( 1 );
    }
    gui->TmpTemplate.Name = entry;
  }

  if( gtk_entry_get_text_length( GTK_ENTRY(gui->TemplateSettings[TS_ETY_OFFI_HEADER]) )) {
    if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->TemplateSettings[TS_ETY_OFFI_HEADER]) ))) {
      gtk_label_set_label( GTK_LABEL(gui->TemplateSettings[TS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
      return( 1 );
    }
    free( (char*) gui->TmpTemplate.HeaderOptions );
    STRDUP( gui->TmpTemplate.HeaderOptions, entry );
  }

  return( 0 );
}

/*
  Purpose: Callback for 'add' button of the template settings window
*/
static void             Callback_add_template(GtkWidget *widget, gpointer param)
{
  s_gui                 *gui = param;
  s_tpl                 *template = 0;
  GtkTreeModel          *model;
  GtkTreeIter           iter;
  pthread_t             tmp_id;
  unsigned int          a;

  /* Verifying template properties ------------------------------------------ */
  if( Gui_to_template_properties( gui, 0 ))
    goto error;

  /* Verifying template operations ------------------------------------------ */
  if( Gui_to_template_operations( gui ))
    goto error;

  /* Finalisation */
  template = New_template( gui->TmpTemplate.Name );

  template->OperationCount = gui->TmpTemplate.OperationCount;
  template->OperationTable = gui->TmpTemplate.OperationTable;
  gui->TmpTemplate.OperationCount = 0;
  gui->TmpTemplate.OperationTable = 0;
  for( a = 0; a < template->OperationCount; ++a ) {
    template->OperationTable[a]->Temporary = 0;
    gtk_container_remove( GTK_CONTAINER(gui->TemplateSettings[TS_VBOX_OPERATION]), template->OperationTable[a]->GtkTable[0] );
  }

  gtk_widget_hide( gui->TemplateSettings[TS_WIN_MAIN] );

  /* Adding the template ---------------------------------------------------- */
  /* to the file table */
  Add_template_to_file( &gui->File, template );

  /* to the choice list of the watcher settings window */
  Add_template_to_menu( gui->WatcherSettings[WS_MENU_TEMPLATE], template );

  /* to the list of the preferences window */
  if( gtk_tree_selection_get_selected( gui->SelectionTemplate, &model, &iter )) {
    gtk_list_store_insert_before( gui->ListTemplates, &template->Iter, &iter );
    gtk_widget_set_sensitive( gui->PreferenceSettings[PS_BUT_REMOVE_TPL], 1 );
  }
  else
    gtk_list_store_insert_before( gui->ListTemplates, &template->Iter, 0 );
  gtk_list_store_set( gui->ListTemplates, &template->Iter, 0, template->Name, -1 );

  /* to the config file */
  if( Generate_template_section( template, &gui->File.SectionReplace )) {
    ERR_SYS( "%s", app->Strs[146] ); /* "Unable to save into config file" */
  }
  if( pthread_create( &tmp_id, &app->Attr, th_Save_modifications_to_file, &gui->File )) {
    ERR_APP( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    th_Save_modifications_to_file( &gui->File );
  }
  gui->Flags &= ~( GUI_FLG_TEMPLATE_MODIFY );
  return;

 error:
  if( template )
    Free_struct_template( template, 0 );

  gtk_widget_show( gui->TemplateSettings[TS_LBL_ERROR] );
}

/*
  Purpose: Getting the current selected template
*/
s_tpl                    *Get_current_selected_template(s_gui *gui)
{
  gchar                  *name;
  GtkTreeModel           *model;
  GtkTreeIter            iter;
  unsigned int           pos;

  if( gtk_tree_selection_get_selected( gui->SelectionTemplate, &model, &iter )) {
    gtk_tree_model_get( model, &iter, 0, &name, -1 );
    for( pos = 0; pos < gui->File.TemplateCount; ++pos )
      if( ! strcmp( name, gui->File.TemplateTable[pos]->Name ))
	return( gui->File.TemplateTable[pos] );
  }

  FIXME( "Reaching the 'adding in progress' template" );
  return( 0 );
}

/*
  Purpose: Callback for 'save' button of the template settings window
*/
static void             Callback_save_template(GtkWidget *unused1, gpointer param)
{
  s_gui                 *gui = param;
  s_tpl                 *template;
  s_wtr                 *watcher;
  char                  *name;
  unsigned int          a, length;
  pthread_t             tmp_id;

  /* Getting current template --------------------------------------------------*/
  template = Get_current_selected_template( gui );
  ASSERT( !template, "template = 0");

  /* The template properties ------------------------------------------------ */
  if( Gui_to_template_properties( gui, template ))
    goto error;

  /* The template operations ------------------------------------------------ */
  if( Gui_to_template_operations( gui ))
    goto error;

  /* Finalisation  -----------------------------------------------------------*/
  gtk_widget_hide( gui->TemplateSettings[TS_WIN_MAIN] );

  /* to the config file */
  length = strlen( template->Name );
  MALLOC( name, 9 + length + 1 ) ; /* "template:" + '\0' */
  memcpy( name, "template:", 9 );
  memcpy( &name[9], template->Name, length );
  name[9 + length] = '\0';
  Search_section2( &gui->File, &gui->File.SectionTarget, name );
  free( name );

  if( gui->TmpTemplate.Name ) {
    free( (char*) template->Name );
    template->Name = gui->TmpTemplate.Name;
    gui->TmpTemplate.Name = 0;
    gtk_list_store_set( gui->ListTemplates, &template->Iter, 0, template->Name, -1 );
  }

  if( Generate_template_section( template, &gui->File.SectionReplace )) {
    ERR_SYS( "%s", app->Strs[146] ); /* "Unable to save into config file" */
  }
  if( pthread_create( &tmp_id, &app->Attr, th_Save_modifications_to_file, &gui->File )) {
    ERR_APP( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    th_Save_modifications_to_file( &gui->File );
  }

  /* Updating concerned watchers in error state */
  for( a = 0; a < gui->WatcherCount; ++a ) {
    watcher = gui->WatcherTable[a];
    //#LOCK WATCHER
    if( watcher->Flags & ( WTR_FLG_OFFICIAL_ERROR | WTR_FLG_VERSION_MISMATCH )
	&& watcher->TemplateName && ! strcmp( watcher->TemplateName, template->Name )) {
      name = Build_cachefile_name( watcher->Name, watcher->NameLength );
      watcher->Flags &= ~( WTR_FLG_OFFICIAL_ERROR | WTR_FLG_VERSION_MISMATCH );
      Retrieve_official_version_from_cache( watcher, name, 0 );
      free( name );
      Update_watcher_area( watcher, 0 );
    }
  }
  gui->Flags &= ~( GUI_FLG_TEMPLATE_MODIFY );
  return;

 error:
  gtk_widget_show( gui->TemplateSettings[TS_LBL_ERROR] );
}

/*
  Purpose: Callback for the 'delete' event of the template settings window
*/
static gboolean          Callback_on_delete_cancel(GtkWidget *widget, GdkEvent *event, gpointer param)
{
  s_gui                  *gui = param;

  gtk_widget_hide( gui->TemplateSettings[TS_WIN_MAIN] );
  return( TRUE );
}

/*
  Purpose: Creating the frame 'Official version' for the 'Add watcher' window
*/
static void              Create_frame_search(GtkWidget *parent)
{
  GtkWidget              *group, *row, *item, *vbox;

#ifdef LIB_HANDY
  group = hdy_preferences_group_new();
  hdy_preferences_group_set_title( (HdyPreferencesGroup*) group, app->Strs[106] ); /* " In the resource, version can be retrieved " */
  gtk_container_add( GTK_CONTAINER(parent), group );

  /* - Add search block */
  row = hdy_action_row_new();
  gtk_widget_set_halign( row, GTK_ALIGN_START );
  gtk_container_add( GTK_CONTAINER(group), row );

  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(row), vbox );
  app->Gui.TemplateSettings[TS_VBOX_OPERATION] = vbox;

  item = gtk_menu_bar_new();
  gtk_container_add( GTK_CONTAINER(group), item );
  Create_menu_search_block( item, CB_TEMPLATE_CREATE );
  app->Gui.TemplateSettings[TS_ADD_BLOCK] = item;

  parent = group;
#else
  /* Search frame ----------------------------------------------------------- */
  group = gtk_frame_new( app->Strs[106] ); /* " In the resource, version can be retrieved " */
  gtk_box_pack_start( GTK_BOX(parent), group, FALSE, TRUE, 0);

  /* - Add search block */
  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(group), vbox );
  app->Gui.TemplateSettings[TS_VBOX_OPERATION] = vbox;

  item = gtk_menu_bar_new();
  gtk_container_add( GTK_CONTAINER(parent), item );
  Create_menu_search_block( item, CB_TEMPLATE_CREATE );
  app->Gui.TemplateSettings[TS_ADD_BLOCK] = item;
#endif

  /* - Version result */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_label_new( app->Strs[232] ); /* "Result " */
  gtk_widget_set_tooltip_text( item, app->Strs[132] ); /* "Result of the search operations" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[132] ); /* "Result of the search operations" */
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  gtk_entry_set_max_length( GTK_ENTRY(item), 70 );
  gtk_widget_set_sensitive( item, 0 );
  app->Gui.TemplateSettings[TS_ETY_RESULT] = item;
}

/*
  Purpose: Callback for the test file button of the template settings window
*/
static void              Callback_get_cache(GtkWidget *widget, gpointer data)
{
  char                   *path = 0, *tmp = 0;

  gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER(app->Gui.ChooseFile), app->CacheFolder );

  if( gtk_dialog_run( GTK_DIALOG(app->Gui.ChooseFile) ) == GTK_RESPONSE_ACCEPT ) {
    path = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER(app->Gui.ChooseFile ));
    tmp = Build_cachefile_name( "tmp_template_result", 19 );
    if( ! tmp )
      goto error;
    Copy_file( path, tmp );
  }

 error:
  free( tmp );
  if( path ) g_free( path );
  gtk_widget_hide( app->Gui.ChooseFile );
}

static void              Create_option_name(GtkWidget *parent)
{
  GtkWidget              *row, *item;

#ifdef LIB_HANDY
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[45] ); /* " Name " */
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_entry_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_container_add( GTK_CONTAINER(row), item );
#else
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_label_new( app->Strs[45] ); /* " Name " */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
#endif

  gtk_entry_set_max_length( GTK_ENTRY(item), MAX_LENGTH_NAME );
  app->Gui.TemplateSettings[TS_ETY_NAME] = item;
}


static void              Create_option_header(GtkWidget *parent)
{
  GtkWidget              *row, *item;

#ifdef LIB_HANDY
  /* HTTP header - label */
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[115] ); /* "HTTP Header options" */
  if( !( app->Flags & APP_FLG_PHONE )) {
    hdy_action_row_set_subtitle( (HdyActionRow*) row, app->Strs[116] ); /* "HTTP header options separated with comma" */
    g_object_set( row, "subtitle-lines", 2, NULL );
  }
  gtk_container_add( GTK_CONTAINER(parent), row );

  /* HTTP header - entry */
  item = gtk_entry_new();
  gtk_container_add( GTK_CONTAINER(parent), item );
#else
  /* HTTP header - label */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_label_new( app->Strs[115] ); /* "HTTP Header options" */
  gtk_widget_set_tooltip_text( item, app->Strs[116] ); /* "HTTP header options separated with comma" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  /* HTTP header - entry */
  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[116] ); /* "HTTP header options separated with comma" */
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
#endif

  app->Gui.TemplateSettings[TS_ETY_OFFI_HEADER] = item;
}


static void              Create_option_test_file(GtkWidget *parent)
{
  GtkWidget              *row, *item, *img;

#ifdef LIB_HANDY
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[256] ); /* "Use file to test template" */
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_button_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
#else
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_label_new( app->Strs[256] ); /* " File to test template " */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_button_new();
#endif

  img = gtk_image_new();
  gtk_image_set_from_pixbuf( GTK_IMAGE(img), app->Gui.ImgTable[IMG_FILE] );

  gtk_button_set_image( (GtkButton*) item, img );
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_get_cache), 0 );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.TemplateSettings[TS_BUT_CACHE] = item;
}

/*
  Purpose: Creating the name / icon frame of the window 'Add watcher'
*/
static void              Create_frame_name(GtkWidget *parent)
{
  GtkWidget              *group;

#ifdef LIB_HANDY
  group = hdy_preferences_group_new();
  hdy_preferences_group_set_title( (HdyPreferencesGroup*) group, app->Strs[255] ); /* "Template" */
  gtk_container_add( GTK_CONTAINER(parent), group );
#else
  GtkWidget              *frame;

  frame = gtk_frame_new( app->Strs[255] ); /* "Template" */
  gtk_box_pack_start( GTK_BOX(parent), frame, FALSE, TRUE, 0);

  group = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(frame), group );
#endif

  Create_option_name( group );
  Create_option_test_file( group );
  Create_option_header( group );
}

/*
  Purpose: Creating the window to add or modify a watcher
*/
void                     Create_window_template_settings()
{
  GtkWidget              *win, *vbox_main, *line, *align, *hbox, *vbox, *item;
  GError                 *err;

  /* Window - Adding a watcher --------------------------------------------- */
#ifdef LIB_HANDY
  if( app->Flags & APP_FLG_PHONE )
    win = hdy_window_new();
  else
#endif
    win = gtk_window_new( GTK_WINDOW_TOPLEVEL );
  app->Gui.TemplateSettings[TS_WIN_MAIN] = win;
  g_signal_connect( G_OBJECT(win), "delete_event", G_CALLBACK(Callback_on_delete_cancel), &app->Gui );
  g_signal_connect( G_OBJECT(win), "destroy", G_CALLBACK(Callback_on_destroy), &app->Gui );

  gtk_window_set_icon_from_file( GTK_WINDOW(win), DEFAULT_APP_RESOURCES"/icon_win_add.png", &err );

  vbox_main = APP_GTK_VBOX( FALSE, 0 );

#ifdef LIB_HANDY
  GtkWidget *leaf = hdy_leaflet_new();
  gtk_container_add( GTK_CONTAINER(win), leaf );
  gtk_widget_show_all( leaf );

  gtk_container_add( GTK_CONTAINER(leaf), vbox_main );

  if( app->Flags & APP_FLG_PHONE ) {
    GtkWidget *bar = hdy_header_bar_new();
    g_object_set( bar, "show_close_button", TRUE,
		  "can-focus", FALSE,
		  NULL );
    gtk_container_add( GTK_CONTAINER(vbox_main), bar );
    app->Gui.TemplateSettings[TS_BAR_WIN] = bar;
  }
  else
    gtk_window_set_default_size( GTK_WINDOW(win), 360, 350 );
#else
  Set_window_properties( win, vbox_main );
#endif

  /* Frames ----------------------------------------------------------------- */
#ifdef LIB_HANDY
  vbox = hdy_preferences_page_new();
  gtk_container_add( GTK_CONTAINER(vbox_main), vbox );
#else
  if( app->Flags & APP_FLG_PHONE )
    vbox = vbox_main ;
  else {
    hbox = APP_GTK_HBOX( FALSE, 0 );
    gtk_container_add( GTK_CONTAINER(vbox_main), hbox );

    vbox = APP_GTK_VBOX( FALSE, 0 );
    gtk_container_add( GTK_CONTAINER(hbox), vbox );
  }
#endif

  Create_frame_name( vbox );
  Create_frame_search( vbox );

  /* Bottom buttons --------------------------------------------------------- */
  line = APP_GTK_HSEPAR();
  gtk_container_add( GTK_CONTAINER(vbox_main), line );

#if GTK_MAJOR_VERSION < 3 || ( GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION < 14 )
  align = gtk_alignment_new( 1, 0, 0, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_main), align );
#else
  align = vbox_main;
#endif

  hbox = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(align), hbox );
#if GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION >= 14
  gtk_widget_set_halign( hbox, GTK_ALIGN_END );
#endif

  item = gtk_label_new( app->Strs[53] ); /* " !! Incomplete !! " */
  app->Gui.TemplateSettings[TS_LBL_ERROR] = item;
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );

  line = APP_GTK_VSEPAR();
  gtk_container_add( GTK_CONTAINER(hbox), line );

  /* - Add - */
  item = gtk_button_new_with_mnemonic( app->Strs[33] ); /* "_Add" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_add_template), &app->Gui );
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );
  app->Gui.TemplateSettings[TS_BUT_ADD] = item;

  /* - Save - */
  item = gtk_button_new_with_mnemonic( app->Strs[16] ); /* "_Save" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_save_template), &app->Gui );
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );
  app->Gui.TemplateSettings[TS_BUT_SAVE] = item;

  /* - Cancel - */
  item = gtk_button_new_with_mnemonic( app->Strs[17] ); /* "C_ancel" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_cancel), &app->Gui );
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );

  /* End ---------------------------------------------------------------- */
  gtk_widget_show_all( vbox_main );
  gtk_widget_hide( app->Gui.TemplateSettings[TS_BUT_SAVE] );
  gtk_widget_hide( app->Gui.TemplateSettings[TS_LBL_ERROR] );
  if( app->Flags & APP_FLG_PHONE )
    gtk_window_move( GTK_WINDOW(win), 0, 0 );
}
