/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../app.h"

/*
  Purpose: Callback for 'Cancel' button of the window 'Add watcher'
*/
static void              Callback_cancel(GtkWidget *widget, gpointer param)
{
  s_gui                  *gui = param;

  gtk_widget_hide( gui->WatcherSettings[WS_WIN_MAIN] );
  gui->Flags &= ~( GUI_FLG_WATCHER_MODIFY | GUI_FLG_NEW_WATCHER );
}

/*
  Purpose: Resizing a pixbuf, creating a new one
*/
static GdkPixbuf         *Resize_pixbuf(GdkPixbuf *pix, unsigned int width, unsigned int height)
{
  GdkPixbuf              *tmp = pix;
  int                    curr_w, curr_h;

  curr_w = gdk_pixbuf_get_width( pix );
  curr_h = gdk_pixbuf_get_height( pix );
  if( curr_w != width || curr_h != height ) {
    tmp = gdk_pixbuf_scale_simple( pix, width, height, GDK_INTERP_BILINEAR );
    if( ! tmp ) {
      ERR_APP( "%s", app->Strs[228] ); /* "Unable to resize image" */
      tmp = pix;
    }
  }
  return( tmp );
}

/*
  Purpose: Putting a pixbuf to a aporox. 26x26 button
*/
void                     Set_icon_to_button(GtkWidget *button, GdkPixbuf *pix)
{
  GdkPixbuf              *tmp;

  ASSERT( !pix, "pix == 0" );
  tmp = Resize_pixbuf( pix, 26, 26 );
  gtk_image_set_from_pixbuf( GTK_IMAGE(button), tmp );
}

/*
  Purpose: Callback for the logo button of the window 'Add watcher'
*/
static void              Callback_add_logo(GtkWidget *widget, gpointer param)
{
  s_gui                  *gui = param;
  char                   *path, *tmp;
  GError                 *err = 0;
  GdkPixbuf              *pix;

  if( app->LastFolder )
    gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER(gui->ChooseFile), app->LastFolder );
  else
    gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER(gui->ChooseFile), DEFAULT_APP_RESOURCES );

  if( gtk_dialog_run( GTK_DIALOG(gui->ChooseFile) ) == GTK_RESPONSE_ACCEPT ) {
    path = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER(gui->ChooseFile ));
    if( path ) {
      pix = gdk_pixbuf_new_from_file( path, &err );
      if( ! pix ) {
	ERR_APP( "'%s': %s\n=> GTK: %s", path, app->Strs[135], err->message ); /* "Line" "Unable to load image file" */
	g_free( path );
	goto add_logo_FAIL;
      }
      Set_icon_to_button( gui->WatcherSettings[WS_IMG_LOGO], pix );

      if( !( tmp = strdup( path ))) {
	ERR_SYS( "%s", app->Strs[2] ); /* "Couldn't allocate enough memory" */
	g_object_unref( pix ); //#will this destroy pix ?
	g_free( path );
	goto add_logo_FAIL;
      }
      g_free( path );

      if( gui->TmpWatcher.OriginPixbuf )
	g_object_unref( gui->TmpWatcher.OriginPixbuf ); //#will this destroy pix ?
      gui->TmpWatcher.OriginPixbuf = pix;
      g_object_ref( gui->TmpWatcher.OriginPixbuf );
      free( gui->TmpWatcher.ImgPath );
      gui->TmpWatcher.ImgPath = tmp;
    }
  }

 add_logo_FAIL:
  gtk_widget_hide( gui->ChooseFile );
}

/*
  Purpose: Callback for the 'Exec' button of the window 'Add watcher'
*/
static void              Callback_get_exec(GtkWidget *widget, gpointer data)
{
  char                   *path;

  gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER(app->Gui.ChooseFile), "/usr/bin" );

  if( gtk_dialog_run( GTK_DIALOG(app->Gui.ChooseFile) ) == GTK_RESPONSE_ACCEPT ) {
    path = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER(app->Gui.ChooseFile ));
    if( path ) {
      gtk_entry_set_text( GTK_ENTRY(app->Gui.WatcherSettings[WS_ETY_LOCAL_EXEC]), path );
      g_free( path );
    }
    else
      gtk_entry_set_text( GTK_ENTRY(app->Gui.WatcherSettings[WS_ETY_LOCAL_EXEC]), "" );
  }
}

/*
  Purpose: Retrieving version for the added watcher
*/
void                    *th_Retrieve_both_versions(void *param)
{
  s_wtr                 *watcher = param;
  unsigned int          a;

  watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE | WTR_FLG_OFFICIAL_ERROR );
  watcher->Flags |= WTR_FLG_OFFICIAL_IN_PROGRESS | WTR_FLG_LOCAL_IN_PROGRESS;
  Update_watcher_area( watcher, 1 );

  Retrieve_local_version( watcher );

  if( watcher->LocalVersion && watcher->OfficialVersion ) {
    bzero( watcher->Trigger_xor, SIZE_BUFFER_XOR );
    for( a = 0; watcher->LocalVersion[a]; ++a )
      watcher->Trigger_xor[a % SIZE_BUFFER_XOR] = watcher->LocalVersion[a];
    for( a = 0; watcher->OfficialVersion[a]; ++a )
      watcher->Trigger_xor[a % SIZE_BUFFER_XOR] ^= watcher->OfficialVersion[a];
  }

  Update_watcher_area( watcher, 1 );

  watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE | WTR_FLG_OFFICIAL_ERROR );
  Refresh_official_version( watcher );

  return( 0 );
}

/*
   Purpose: Renaming file in same folder
*/
static _Bool            Rename_file(const char *folder, const char *old, const char *new)
{
  char                  *tmp_old, *tmp_new;
  unsigned int          len_old, len_new, len_folder;

  len_folder = strlen( folder );
  len_old = strlen( old );
  len_new = strlen( new );

  MALLOC( tmp_old, len_folder + len_old + 2 ); /* '/' + '\0' */
  tmp_old[len_folder + 1 + len_old] = '\0';
  memcpy( tmp_old, app->CacheFolder, len_folder );
  tmp_old[len_folder] = '/';
  memcpy( &tmp_old[len_folder + 1], old, len_old );

  MALLOC( tmp_new, len_folder + len_new + 2 ); /* '/' + '\0' */
  memcpy( tmp_new, app->CacheFolder, len_folder );
  tmp_new[len_folder] = '/';
  memcpy( &tmp_new[len_folder + 1], new, len_new );
  tmp_new[len_folder + 1 + len_new] = '\0';

  rename( tmp_old, tmp_new );
  FIXME( "Verify return of rename()" );

  free( tmp_old );
  free( tmp_new );
  return( 0 );
}

/*
  Purpose: Retrieving informations of common properties from GUI and put it into watcher
*/
static _Bool            Gui_to_watcher_properties(s_gui *gui, s_wtr *watcher)
{
  const char            *entry;
  unsigned int          a;

  /* Name --------- */
  free( gui->TmpWatcher.Name );
  gui->TmpWatcher.Name = 0;
  gui->TmpWatcher.NameLength = gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_NAME]) );

  if( ! gui->TmpWatcher.NameLength ) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[186] ); /* "Missing name" */
    return( 1 );
  }

  if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_NAME]) ))) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
    return( 1 );
  }

  for( a = 0; a < app->Gui.WatcherCount; ++a)
    if( ! strcmp( entry, app->Gui.WatcherTable[a]->Name ))
      break;

  if( ! watcher->Name || strcmp( entry, watcher->Name )) {
    if( a != app->Gui.WatcherCount ) { /* modified name already exist */
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[97] ); /* "This name is already used" */
      return( 1 );
    }
    if( !( gui->TmpWatcher.Name = strdup( entry ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
      return( 1 );
    }
  }

  return( 0 );
}

/*
  Purpose: Moving list of operations from a watcher to another
*/
static void             Move_local_operation_list(s_wtr *source, s_wtr *destination)
{
  s_op                  *operation;
  unsigned int          a, b;

  ASSERT( !source, "source = 0" );
  ASSERT( !destination, "destination = 0" );

  /* Removing dropped operations */
  for( a = 0; a < destination->LocalOperationCount; ++a ) {
    operation = destination->LocalOperationTable[a];
    for( b = 0; b < source->LocalOperationCount; ++b )
      if( operation->GtkTable[0] == source->LocalOperationTable[b]->GtkTable[0] )
	break;
    if( b == source->LocalOperationCount )
      Free_struct_operation( operation, 0 );
  }

  /* Replacing operations table */
  free( destination->LocalOperationTable );
  destination->LocalOperationCount = source->LocalOperationCount;
  destination->LocalOperationTable = source->LocalOperationTable;
  source->LocalOperationCount = 0;
  source->LocalOperationTable = 0;

  for( a = 0; a < destination->LocalOperationCount; ++a ) {
    operation = destination->LocalOperationTable[a];
    operation->Temporary = 0;
  }
}

/*
  Purpose: Moving list of operations from a watcher to another
*/
static void             Move_official_operation_list(s_wtr *source, s_wtr *destination)
{
  s_op                  *operation;
  unsigned int          a, b;

  ASSERT( !source, "source = 0" );
  ASSERT( !destination, "destination = 0" );

  /* Removing dropped operations */
  for( a = 0; a < destination->OfficialOperationCount; ++a ) {
    operation = destination->OfficialOperationTable[a];
    for( b = 0; b < source->OfficialOperationCount; ++b )
      if( operation->GtkTable[0] == source->OfficialOperationTable[b]->GtkTable[0] )
	break;
    if( b == source->OfficialOperationCount )
      Free_struct_operation( operation, 0 );
  }

  /* Replacing operations table */
  free( destination->OfficialOperationTable );
  destination->OfficialOperationCount = source->OfficialOperationCount;
  destination->OfficialOperationTable = source->OfficialOperationTable;
  source->OfficialOperationCount = 0;
  source->OfficialOperationTable = 0;

  for( a = 0; a < destination->OfficialOperationCount; ++a ) {
    operation = destination->OfficialOperationTable[a];
    operation->Temporary = 0;
  }
}

/*
  Purpose: Retrieving list of local operations in the watcher
*/
static _Bool            Gui_to_watcher_local_operation(s_gui *gui, s_wtr *watcher)
{
  s_op                  *operation;
  unsigned int          a;
  unsigned char         ret;

  if( ! watcher->LocalOperationCount ) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[257] ); /* "Missing search block" */
    return( 1 );
  }

  if( watcher->LocalOperationCount != gui->TmpWatcher.LocalOperationCount )
    watcher->Flags |= WTR_FLG_LOCAL_REFRESH;

  for( a = 0; a < watcher->LocalOperationCount; ++a ) {
    operation = watcher->LocalOperationTable[a];
    ret = 0;
    switch( operation->Type ) {
    case OP_TYPE_AFTER_LINE: ret = Get_gui_after_line( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_AFTER_WORD: ret = Get_gui_after_word( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_FIELD:   ret = Get_gui_in_field( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_LINE:    ret = Get_gui_in_line( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_SECTION: ret = Get_gui_in_section( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_WORD:    ret = Get_gui_in_word( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    }
    if( ret == 1 ) return( 1 );
    if( ret == 2 ) watcher->Flags |= WTR_FLG_LOCAL_REFRESH;
  }
  return( 0 );
}

/*
  Purpose: Retrieving list of official operations in the watcher
*/
static _Bool            Gui_to_watcher_official_operation(s_gui *gui, s_wtr *watcher)
{
  s_op                  *operation;
  unsigned int          a;
  unsigned char         ret;

  if( ! watcher->OfficialOperationCount && ! watcher->TemplateName ) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[257] ); /* "Missing search block" */
    return( 1 );
  }

  if( watcher->OfficialOperationCount != gui->TmpWatcher.OfficialOperationCount )
    watcher->Flags |= WTR_FLG_OFFICIAL_REFRESH;

  if( gui->TmpWatcher.TemplateName && watcher->TemplateName
      && strcmp( gui->TmpWatcher.TemplateName, watcher->TemplateName ))
    watcher->Flags |= WTR_FLG_OFFICIAL_REFRESH;

  for( a = 0; a < watcher->OfficialOperationCount; ++a ) {
    operation = watcher->OfficialOperationTable[a];
    ret = 0;
    switch( operation->Type ) {
    case OP_TYPE_AFTER_LINE: ret = Get_gui_after_line( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_AFTER_WORD: ret = Get_gui_after_word( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_FIELD:   ret = Get_gui_in_field( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_LINE:    ret = Get_gui_in_line( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_SECTION: ret = Get_gui_in_section( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    case OP_TYPE_IN_WORD:    ret = Get_gui_in_word( operation, gui->WatcherSettings[WS_LBL_ERROR] ); break;
    }
    if( ret == 1 ) return( 1 );
    if( ret == 2 ) watcher->Flags |= WTR_FLG_OFFICIAL_REFRESH;
  }
  return( 0 );
}

/*
  Purpose: Retrieving informations of official version from GUI and put it into watcher
*/
static _Bool            Gui_to_watcher_official(s_gui *gui, s_wtr *watcher)
{
  const char            *entry;
  char                  *url;
  unsigned int          pos, length;
  _Bool                 http = 0;

  if( !( length = gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_OFFI_URL]) ))) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[187] ); /* "Missing official URL" */
    return( 1 );
  }
  if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_OFFI_URL]) ))) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
    return( 1 );
  }

  if( ! watcher->OfficialURL || strcmp( entry, watcher->OfficialURL )) {
    STRDUP( url, entry );
    if( url[0] == 'h' && length > 6 && ! strncmp( url, "http://", 7 )) {
      http = 1;
      pos = 7;
    }
    else {
      if( url[0] != 'h' || length < 8 || strncmp( url, "https://", 8 )) {
	REALLOC( url, 8 + length + 1 ); /* "https://" + '\0' */
	memmove( &url[8], url, length );
	memcpy( url, "https://", 8 );
	url[8 + length] = '\0';
      }
      pos = 8;
    }

    if( Domain_is_valid( &url[pos] )) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[120] ); /* "The URL has an unrecognized format" */
      free( url );
      return( 1 );
    }

    if( http )
      watcher->Flags |= WTR_FLG_HTTP | WTR_FLG_OFFICIAL_UPDATE;
    else {
      watcher->Flags |= WTR_FLG_OFFICIAL_UPDATE;
      watcher->Flags &= ~( WTR_FLG_HTTP );
    }
    free( watcher->OfficialURL );
    watcher->OfficialURL = url;
  }

  /* HTTP header options  */
#if LIB_HANDY
  if( APP_GTK_TOGGLE_GET_ACTIVE( gui->WatcherSettings[WS_TOGGLE_HEADER] )) {
#else
  if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(gui->WatcherSettings[WS_TOGGLE_HEADER]) )) {
#endif
    if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_OFFI_HEADER]) )) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with absolute path" */
      return( 1 );
    }
    if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_OFFI_HEADER]) ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
      return( 1 );
    }
    if( !( entry = strdup( entry ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
      return( 1 );
    }
    free( watcher->OfficialHeader );
    watcher->OfficialHeader = (char*) entry;
    watcher->Flags |= WTR_FLG_OFFICIAL_HEADER;
  }
  else
    watcher->Flags &= ~( WTR_FLG_OFFICIAL_HEADER );

  watcher->OfficialTimer = (unsigned int) gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON(gui->WatcherSettings[WS_SPIN_TIMER]) );

  if( ! gui->TmpWatcher.TemplateName ) {
    /* Updating operation structures */
    if( Gui_to_watcher_official_operation( gui, &gui->TmpWatcher ))
      return( 1 );
  }
  return( 0 );
}

/*
  Purpose: Retrieving informations of local version from GUI and put it into watcher
*/
static _Bool            Gui_to_watcher_local(s_gui *gui, s_wtr *watcher)
{
  const char            *entry;

  /* Automatique version ---------------------------------------------------- */
  if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(gui->WatcherSettings[WS_RAD_VERSION_AUTO]) )) {
    if( !( watcher->Flags & WTR_FLG_LOCAL_AUTO ))
      watcher->Flags |= WTR_FLG_LOCAL_AUTO | WTR_FLG_LOCAL_REFRESH;

    /* stderr */
#if LIB_HANDY
    if( APP_GTK_TOGGLE_GET_ACTIVE( gui->WatcherSettings[WS_TOGGLE_STDERR] )) {
#else
    if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(gui->WatcherSettings[WS_TOGGLE_STDERR]) )) {
#endif
      if( !( watcher->Flags & WTR_FLG_STDERR ))
	watcher->Flags |= WTR_FLG_STDERR | WTR_FLG_LOCAL_REFRESH;
    }
    else {
      if( watcher->Flags & WTR_FLG_STDERR )
	watcher->Flags |= WTR_FLG_LOCAL_REFRESH;
      watcher->Flags &= ~( WTR_FLG_STDERR );
    }

    /* Executable */
    if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_LOCAL_EXEC]) )) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with full path" */
      return( 1 );
    }
    if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_LOCAL_EXEC]) ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
      return( 1 );
    }
    if( *entry != '/' ) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with full path" */
      return( 1 );
    }
#if LIB_HANDY
    if( APP_GTK_TOGGLE_GET_ACTIVE( gui->WatcherSettings[WS_TOGGLE_OPTIONNAL] )) {
#else
    if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(gui->WatcherSettings[WS_TOGGLE_OPTIONNAL]) )) {
#endif
      watcher->Flags |= WTR_FLG_OPTIONNAL;
      if( Check_file( &gui->File, entry, 1 ))
	watcher->Flags |= WTR_FLG_DISABLE;
    }
    else
      Check_file( &gui->File, entry, 1 );

    if( ! watcher->LocalExec || strcmp( entry, watcher->LocalExec )) {
      if( !( entry = strdup( entry ))) {
	gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
	return( 1 );
      }
      free( watcher->LocalExec );
      watcher->LocalExec = (char*) entry;
      watcher->Flags |= WTR_FLG_LOCAL_REFRESH;
    }

    /* Argument */
    if( gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_ARGUMENT]) )) {
      if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_ARGUMENT]) ))) {
	gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
	return( 1 );
      }
      if( ! watcher->LocalArg || ! strcmp( entry, watcher->LocalArg )) {
	if( !( entry = strdup( entry ))) {
	  gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
	  return( 1 );
	}
	free( watcher->LocalArg );
	watcher->LocalArg = (char*) entry;
	watcher->Flags |= WTR_FLG_LOCAL_REFRESH;
      }
    }

    if( Gui_to_watcher_local_operation( gui, &gui->TmpWatcher ))
      return( 1 );
  }
  else {
    /* Manual version ------------------------------------------------------- */
    if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_LOCAL_VERSION]) )) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[191] ); /* "Missing manual local version" */
      return( 1 );
    }
    if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_LOCAL_VERSION]) ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
      return( 1 );
    }
    if( ! watcher->LocalVersion || ! strcmp( entry, watcher->LocalVersion )) {
      if( !( entry = strdup( entry ))) {
	gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
	return( 1 );
      }
      free( watcher->LocalVersion );
      watcher->LocalVersion = (char*) entry;
      watcher->Flags |= WTR_FLG_LOCAL_REFRESH;
    }
    watcher->Flags &= ~( WTR_FLG_LOCAL_AUTO );
  }

  return( 0 );
}

/*
  Purpose: Retrieving informations from of other options GUI and put it into watcher
*/
static _Bool            Gui_to_watcher_others(s_gui *gui, s_wtr *watcher)
{
  const char            *entry;

  /* Trigger on mismatch  */
#if LIB_HANDY
  if( APP_GTK_TOGGLE_GET_ACTIVE( gui->WatcherSettings[WS_TOGGLE_MISMATCH] )) {
#else
  if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(gui->WatcherSettings[WS_TOGGLE_MISMATCH]) )) {
#endif
    watcher->Flags |= WTR_FLG_TRIGGER_MISMATCH;
    if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_TRIGGER_MISMATCH]) )) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with absolute path" */
      return( 1 );
    }
    if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_TRIGGER_MISMATCH]) ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
      return( 1 );
    }
    if( *entry != '/' ) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with full path" */
      return( 1 );
    }
    if( !( entry = strdup( entry ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
      return( 1 );
    }
    free( watcher->Trigger_on_mismatch );
    watcher->Trigger_on_mismatch = (char*) entry;
  }

  /* Trigger on error  */
#if LIB_HANDY
  if( APP_GTK_TOGGLE_GET_ACTIVE( gui->WatcherSettings[WS_TOGGLE_ERROR] )) {
#else
  if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(gui->WatcherSettings[WS_TOGGLE_ERROR]) )) {
#endif
    watcher->Flags |= WTR_FLG_TRIGGER_ERROR;
    if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_TRIGGER_ERROR]) )) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with absolute path" */
      return( 1 );
    }
    if( !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_TRIGGER_ERROR]) ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[158] ); /* "A problem occured with a GTK widget" */
      return( 1 );
    }
    if( *entry != '/' ) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[190] ); /* "Missing executable with full path" */
      return( 1 );
    }
    if( !( entry = strdup( entry ))) {
      gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
      return( 1 );
    }
    free( watcher->Trigger_on_error );
    watcher->Trigger_on_error = (char*) entry;
  }

  return( 0 );
}

/*
  Purpose: Callback for 'add' button of the window 'Add watcher'
*/
static void             Callback_add_watcher(GtkWidget *widget, gpointer param)
{
  s_gui                 *gui = param;
  GtkTreeModel          *model;
  const gchar           *name;
  s_wtr                 *watcher = 0;
  pthread_t             tmp_id;
  GtkTreeIter           iter;
  unsigned int          a, count;

  if( !( watcher = calloc( 1, sizeof(*watcher) ))) {
    gtk_label_set_label( GTK_LABEL(gui->WatcherSettings[WS_LBL_ERROR]), app->Strs[1] ); /* "Couldn't allocate enough memory" */
    goto error;
  }
  watcher->Flags = gui->TmpWatcher.Flags;

  /* The watcher properties ------------------------------------------------- */
  if( Gui_to_watcher_properties( gui, watcher ))
    goto error;

  /* Official version ------------------------------------------------------- */
  if( Gui_to_watcher_official( gui, watcher ))
    goto error;

  /* Local version ---------------------------------------------------------- */
  if( Gui_to_watcher_local( gui, watcher ))
    goto error;

  /* Other options ---------------------------------------------------------- */
  if( Gui_to_watcher_others( gui, watcher ))
    goto error;

  /* Finalisation ----------------------------------------------------------- */
  /* Name */
  watcher->Name = gui->TmpWatcher.Name;
  watcher->NameLength = gui->TmpWatcher.NameLength;
  gui->TmpWatcher.Name = 0;

  /* Icon  */
  if( gui->TmpWatcher.OriginPixbuf ) {
    watcher->OriginPixbuf = gui->TmpWatcher.OriginPixbuf;
    watcher->ImgPath = gui->TmpWatcher.ImgPath;
    gui->TmpWatcher.ImgPath = 0;
  }
  else
    watcher->OriginPixbuf = app->Gui.ImgTable[IMG_DEFAULT_WATCHER];

  /* Template  */
  if( gui->TmpWatcher.TemplateName ) {
    watcher->TemplateName = gui->TmpWatcher.TemplateName;
    gui->TmpWatcher.TemplateName = 0;
  }
  else
    Move_official_operation_list( &gui->TmpWatcher, watcher );

  if( watcher->Flags & WTR_FLG_LOCAL_AUTO )
    Move_local_operation_list( &gui->TmpWatcher, watcher );

  /* Adding the watcher ----------------------------------------------------- */
  gtk_widget_hide( gui->WatcherSettings[WS_WIN_MAIN] );

  /* to the table */
  if( gtk_tree_selection_get_selected( gui->SelectionWatcher, &model, &iter )) {
    if( gui->WatcherCount + 1 >= gui->WatcherTableSize ) {
      gui->WatcherTableSize += STEP_SIZE_WATCHER_TABLE;
      REALLOC( gui->WatcherTable, gui->WatcherTableSize * sizeof(*gui->WatcherTable) );
      bzero( &gui->WatcherTable[gui->WatcherCount], STEP_SIZE_WATCHER_TABLE * sizeof(*gui->WatcherTable) );
    }
    gtk_tree_model_get( model, &iter, COL_NAME, &name, -1 );
    for( a = gui->WatcherCount; a ; --a) {
      gui->WatcherTable[a] = gui->WatcherTable[a - 1];
      watcher->Pos = a;
      if( ! strcmp( name, gui->WatcherTable[a]->Name ))
	break;
    }
    gui->WatcherTable[a] = watcher;
    ++gui->WatcherCount;
  }
  else
    Add_watcher_to_gui( gui, watcher );

  /* to the main window */
  Create_new_gui_watcher( watcher );
  Update_main_geometry();
  gtk_frame_set_label( GTK_FRAME(watcher->LabelFrame), watcher->Name );

  /* to the list */
  watcher->CurrentPixbuf = Resize_pixbuf( watcher->OriginPixbuf, app->Gui.WatcherSizeX, app->Gui.WatcherSizeY );
  gtk_list_store_set( gui->ListWatchers, &watcher->Iter, COL_LOGO, watcher->CurrentPixbuf, COL_NAME, watcher->Name, -1 );

  /* to the config file */
  if( Generate_watcher_section( watcher, &gui->File.SectionReplace )) {
    ERR_SYS( "%s", app->Strs[146] ); /* "Unable to save into config file" */
    FIXME( "Throw GUI error" );
  }

  if( pthread_create( &tmp_id, &app->Attr, th_Save_modifications_to_file, &gui->File )) {
    ERR_APP( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    th_Save_modifications_to_file( &gui->File );
  }

  /* Updating gui ----------------------------------------------------------- */
  struct timeval now;
  gettimeofday( &now, NULL );
  Update_timer( watcher, now.tv_sec, 0 );

  if( pthread_create( &tmp_id, &app->Attr, th_Retrieve_both_versions, watcher )) {
    gtk_label_set_label( GTK_LABEL(watcher->LabelLocalVersionResult),
			 app->Strs[179] ); /* "Unable to create a thread" */
    gtk_label_set_label( GTK_LABEL(watcher->LabelOfficialVersionResult),
			 app->Strs[179] ); /* "Unable to create a thread" */
    ERR_SYS( "%s", app->Strs[179] ); /* "Unable to create a thread" */
  }

  count = 0;
  for( a = 0; a < gui->WatcherCount; ++a )
    if( !( gui->WatcherTable[a]->Flags & WTR_FLG_DISABLE ))
      gui->WatcherTable[a]->VisiblePos = count++;
  if( count )
    gui->Flags &= ~( GUI_FLG_NO_WATCHER );

  gui->Flags &= ~( GUI_FLG_WATCHER_MODIFY );
  return;

 error:
  if( watcher )
    Free_struct_watcher( watcher, 1, 0 );
  gtk_widget_show( gui->WatcherSettings[WS_LBL_ERROR] );
}

/*
  Purpose: Callback for 'save' button of the watcher settings window
*/
static void             Callback_save_watcher(GtkWidget *unused1, s_gui *gui)
{
  s_wtr                 *watcher;
  pthread_t             tmp_id;

  /* Getting current watcher --------------------------------------------------*/
  watcher = Get_current_selected_watcher( gui );
  ASSERT( !watcher, "watcher = 0");
  FIXME( "Lock watcher " );

  /* The watcher properties ------------------------------------------------- */
  if( Gui_to_watcher_properties( gui, watcher ))
    goto error;

  /* Official version ------------------------------------------------------- */
  if( Gui_to_watcher_official( gui, watcher ))
    goto error;

  /* Local version ---------------------------------------------------------- */
  if( Gui_to_watcher_local( gui, watcher ))
    goto error;

  /* Other options ---------------------------------------------------------- */
  if( Gui_to_watcher_others( gui, watcher ))
    goto error;

  /* Finalisation  -----------------------------------------------------------*/
  gtk_widget_hide( gui->WatcherSettings[WS_WIN_MAIN] );
  gtk_widget_hide( gui->PreferenceSettings[PS_BUT_SAVE] );
  gtk_widget_hide( gui->WatcherSettings[WS_BUT_SAVE] );

  /* Icon  */
  if( gui->TmpWatcher.ImgPath ) {
    free( watcher->ImgPath );
    watcher->ImgPath = gui->TmpWatcher.ImgPath;
    gui->TmpWatcher.ImgPath = 0;
  }

  watcher->OriginPixbuf = gui->TmpWatcher.OriginPixbuf;
  gui->TmpWatcher.OriginPixbuf = 0;

  if( gui->TmpWatcher.OriginPixbuf && watcher->OriginPixbuf != gui->TmpWatcher.OriginPixbuf ) {
    watcher->OriginPixbuf = gui->TmpWatcher.OriginPixbuf;
    gui->TmpWatcher.OriginPixbuf = 0;
    gtk_list_store_set( gui->ListWatchers, &watcher->Iter, COL_LOGO, watcher->OriginPixbuf, -1 );
    if( watcher->CurrentPixbuf )
      g_object_unref( watcher->CurrentPixbuf );
    watcher->CurrentPixbuf = Resize_pixbuf( watcher->OriginPixbuf, app->Gui.WatcherSizeX, app->Gui.WatcherSizeY );
  }

  /* Template */
  if( gui->TmpWatcher.TemplateName ) {
    if( ! watcher->TemplateName || strcmp( gui->TmpWatcher.TemplateName, watcher->TemplateName )) {
      watcher->TemplateName = gui->TmpWatcher.TemplateName;
      gui->TmpWatcher.TemplateName = 0;
      watcher->Flags |= WTR_FLG_OFFICIAL_REFRESH;
    }
  }
  else {
    if( watcher->TemplateName ) {
      free( watcher->TemplateName );
      watcher->TemplateName = 0;
    }
    Move_official_operation_list( &gui->TmpWatcher, watcher );
    watcher->Flags |= WTR_FLG_OFFICIAL_REFRESH;
  }

  if( watcher->Flags & WTR_FLG_LOCAL_AUTO ) {
    Move_local_operation_list( &gui->TmpWatcher, watcher );
    watcher->Flags |= WTR_FLG_LOCAL_REFRESH;
  }

  /* Saving to the config file */
  Search_section2( &gui->File, &gui->File.SectionTarget, watcher->Name );

  if( gui->TmpWatcher.Name ) {
    Rename_file( app->CacheFolder, watcher->Name, gui->TmpWatcher.Name );

    free( watcher->Name );
    watcher->Name = gui->TmpWatcher.Name;
    watcher->NameLength = gui->TmpWatcher.NameLength;
    gui->TmpWatcher.Name = 0;
    gtk_list_store_set( gui->ListWatchers, &watcher->Iter, COL_NAME, watcher->Name, -1 );
    gtk_frame_set_label( GTK_FRAME(watcher->LabelFrame), watcher->Name );
    watcher->Flags |= WTR_FLG_AREA_REFRESH;
  }

  if( Generate_watcher_section( watcher, &gui->File.SectionReplace )) {
    ERR_SYS( "%s", app->Strs[146] ); /* "Unable to save into config file" */
    FIXME( "Throw GUI error" );
  }
  if( pthread_create( &tmp_id, &app->Attr, th_Save_modifications_to_file, &gui->File )) {
    ERR_APP( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    th_Save_modifications_to_file( &gui->File );
  }

  struct timeval now;
  gettimeofday( &now, NULL );
  Update_timer( watcher, now.tv_sec, 0 );

  /* Updating informations */
  if(( watcher->Flags & WTR_FLG_LOCAL_REFRESH )
     && ( watcher->Flags & ( WTR_FLG_OFFICIAL_UPDATE | WTR_FLG_OFFICIAL_REFRESH ))) {
    watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE );
    if( pthread_create( &tmp_id, &app->Attr, th_Retrieve_both_versions, watcher )) {
      gtk_label_set_label( GTK_LABEL(watcher->LabelLocalVersionResult),
			   app->Strs[179] ); /* "Unable to create a thread" */
      gtk_label_set_label( GTK_LABEL(watcher->LabelOfficialVersionResult),
			   app->Strs[179] ); /* "Unable to create a thread" */
      ERR_SYS( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    }
    gui->Flags &= ~( GUI_FLG_WATCHER_MODIFY );
    return;
  }
  if( watcher->Flags & WTR_FLG_LOCAL_REFRESH ) {
    watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE );
    if( pthread_create( &tmp_id, &app->Attr, th_Retrieve_local_version, watcher )) {
      gtk_label_set_label( GTK_LABEL(watcher->LabelLocalVersionResult),
			   app->Strs[179] ); /* "Unable to create a thread" */
      ERR_SYS( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    }
  }
  else if(( watcher->Flags & ( WTR_FLG_OFFICIAL_UPDATE | WTR_FLG_OFFICIAL_REFRESH ))) {
    watcher->Flags &= ~( WTR_FLG_NO_AUTO_UPDATE );
    if( pthread_create( &tmp_id, &app->Attr, th_Refresh_official_version, watcher )) {
      gtk_label_set_label( GTK_LABEL(watcher->LabelOfficialVersionResult),
			   app->Strs[179] ); /* "Unable to create a thread" */
      ERR_SYS( "%s", app->Strs[179] ); /* "Unable to create a thread" */
    }
  }

  gui->Flags &= ~( GUI_FLG_WATCHER_MODIFY | GUI_FLG_NEW_WATCHER );

  return;

  error:
  bzero( &gui->TmpWatcher, sizeof(gui->TmpWatcher) );
  gtk_widget_show( gui->WatcherSettings[WS_LBL_ERROR] );
}

/*
  Purpose: Callback for radio 'Manual' of local frame
*/
static void              Callback_local_manual(GtkWidget *widget, s_gui *gui)
{
  gtk_widget_set_sensitive( gui->WatcherSettings[WS_ETY_LOCAL_VERSION], 1 );
#ifdef LIB_HANDY
  hdy_expander_row_set_enable_expansion( (HdyExpanderRow*) gui->WatcherSettings[WS_BOX_VERSION_AUTO], 0 );
#else
  gtk_widget_set_sensitive( gui->WatcherSettings[WS_BOX_VERSION_AUTO], 0 );
#endif
}

/*
  Purpose: Callback for radio 'Automatic' of local frame
*/
static void              Callback_local_automatic(GtkWidget *widget, s_gui *gui)
{
  gtk_widget_set_sensitive( gui->WatcherSettings[WS_ETY_LOCAL_VERSION], 0 );
#ifdef LIB_HANDY
  hdy_expander_row_set_enable_expansion( (HdyExpanderRow*) gui->WatcherSettings[WS_BOX_VERSION_AUTO], 1 );
#else
  gtk_widget_set_sensitive( gui->WatcherSettings[WS_BOX_VERSION_AUTO], 1 );
#endif
}

/*
  Purpose: Callback for check box which sets sensitive or not the given widget
*/
void                     Callback_check_to_sensitive(GtkWidget *widget, GtkWidget *checkbox)
{
  gtk_widget_set_sensitive( checkbox, gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(widget) ));
}

/*
  Purpose: Callback for switch which sets sensitive or not the given widget
*/
void                     Callback_switch_to_sensitive(GtkWidget *widget, gboolean state, gpointer data)
{
  gtk_widget_set_sensitive( data, state );
}

/*
  Purpose: Callback for the delete event of the watcher settings window
*/
static gboolean          Callback_on_delete_cancel(GtkWidget *widget, GdkEvent *event, gpointer param)
{
  Callback_cancel( widget, param );
  return( TRUE );
}

/*
  Purpose: Creating the frame for the locale version of the window 'Add watcher'
*/
static void              Create_frame_local_version(GtkWidget *parent)
{
  GtkWidget              *group, *vbox, *row, *item, *img;

#ifdef LIB_HANDY
  group = hdy_preferences_group_new();
  gtk_container_add( GTK_CONTAINER(parent), group );

  GtkWidget *exprow = hdy_expander_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) exprow, app->Strs[43] ); /* "Retrieving local version" */
  if( !( app->Flags & APP_FLG_PHONE ))
    hdy_expander_row_set_subtitle( (HdyExpanderRow*) exprow, app->Strs[111] ); /* "Parameters to determine version" */
  gtk_container_add( GTK_CONTAINER(group), exprow );
  app->Gui.WatcherSettings[WS_EXPROW_LOCAL] = exprow;

  /* - Manual ----------------- */
  row = hdy_action_row_new();
  gtk_container_add( GTK_CONTAINER(exprow), row );

  item = gtk_radio_button_new_with_label_from_widget( NULL, app->Strs[51] ); /* "Manual version" */
  hdy_action_row_add_prefix( (HdyActionRow*) row, item );
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_local_manual), &app->Gui );
  gtk_widget_set_tooltip_text( item, app->Strs[101] ); /* "local_version: Official version will be compared to this value" */
  app->Gui.WatcherSettings[WS_RAD_VERSION_MANUAL] = item;

  /* version - Entry */
  item = gtk_entry_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  gtk_widget_set_tooltip_text( item, app->Strs[101] ); /* "local_version: Official version will be compared to this value" */
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_entry_set_max_length( GTK_ENTRY(item), MAX_LENGTH_VERSION + 3 ); /* 50 */
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_ETY_LOCAL_VERSION] = item;

  /* - Automatic -------------- */
  row = hdy_expander_row_new();
  gtk_container_add( GTK_CONTAINER(exprow), row );
  app->Gui.WatcherSettings[WS_BOX_VERSION_AUTO] = row;
  exprow = row;

  item = gtk_radio_button_new_with_label_from_widget( GTK_RADIO_BUTTON(app->Gui.WatcherSettings[WS_RAD_VERSION_MANUAL]), app->Strs[52] ); /* "Version automatic" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_local_automatic), &app->Gui );
  hdy_expander_row_add_prefix( (HdyExpanderRow*) exprow, item );
  app->Gui.WatcherSettings[WS_RAD_VERSION_AUTO] = item;

  /* optionnal - check box */
  row = hdy_action_row_new();
  hdy_action_row_set_subtitle( (HdyActionRow*) row, app->Strs[177] ); /* "optionnal: The watcher will be disable at start, if the executable is not available" */
  g_object_set( row, "subtitle-lines", 2, NULL );
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[176] ); /* "Optionnal" */
  gtk_container_add( GTK_CONTAINER(exprow), row );

  item = gtk_switch_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_TOGGLE_OPTIONNAL] = item;

  /* Executable - Entry */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(exprow), row );

  item = gtk_label_new( app->Strs[64] ); /* " Executable " */
  gtk_widget_set_tooltip_text( item, app->Strs[102] ); /* "exec: Executable file launched to retrieve the local version" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[102] ); /* "exec: Executable file launched to retrieve the local version" */
  gtk_entry_set_max_length( GTK_ENTRY(item), 250 );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_LOCAL_EXEC] = item;

  /* Executable - Button */
  img = gtk_image_new();
  gtk_image_set_from_pixbuf( GTK_IMAGE(img), app->Gui.ImgTable[IMG_FILE] );

  item = gtk_button_new();
  gtk_button_set_image( (GtkButton*) item, img );
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_get_exec), 0 );
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_BUT_EXEC] = item;

  /* Argument - Entry */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(exprow), row );

  item = gtk_label_new( app->Strs[65] ); /* " Argument " */
  gtk_widget_set_tooltip_text( item, app->Strs[103] ); /* "arg: Argument given to the executable to retrieve the local version" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[103] ); /* "arg: Argument given to the executable to retrieve the local version" */
  gtk_entry_set_max_length( GTK_ENTRY(item), 250 );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_ARGUMENT] = item;

  /* Search frame ----------------------------------------------------------- */
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[200] ); /* " In the command output, version can be retrieved " */
  gtk_container_add( GTK_CONTAINER(exprow), row );

  /* stderr - check box */
  row = hdy_action_row_new();
  if( !( app->Flags & APP_FLG_PHONE )) {
    hdy_action_row_set_subtitle( (HdyActionRow*) row, app->Strs[105] ); /* "stderr: Because some developers don't understand what 'stderr' stand for" */
    g_object_set( row, "subtitle-lines", 2, NULL );
  }
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[104] ); /* "From the error standard outpout" */
  gtk_container_add( GTK_CONTAINER(exprow), row );

  item = gtk_switch_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_TOGGLE_STDERR] = item;

  /* - Add search block */
  row = hdy_action_row_new();
  gtk_widget_set_halign( row, GTK_ALIGN_START );
  gtk_container_add( GTK_CONTAINER(exprow), row );

  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(row), vbox );
  app->Gui.WatcherSettings[WS_VBOX_SEARCH_LOCAL] = vbox;

  item = gtk_menu_bar_new();
  gtk_container_add( GTK_CONTAINER(exprow), item );
  Create_menu_search_block( item, CB_LOCAL_VERSION );
  app->Gui.WatcherSettings[WS_ADD_BLOCK_LOCAL] = item;
#else
  GtkWidget *vbox_auto;
  /* Frame ---------------------------------------------------------------- */
  group = gtk_frame_new( app->Strs[43] ); /* "Retrieving local version" */
  gtk_box_pack_start( GTK_BOX(parent), group, FALSE, TRUE, 0);

  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(group), vbox );

  /* - Manual ----------------- */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox), row );

  item = gtk_radio_button_new_with_label_from_widget( NULL, app->Strs[51] ); /* "Manual version" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_local_manual), &app->Gui );
  gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(item), TRUE );
  gtk_widget_set_tooltip_text( item, app->Strs[101] ); /* "local_version: Official version will be compared to this value" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, TRUE, 0);
  app->Gui.WatcherSettings[WS_RAD_VERSION_MANUAL] = item;

  /* - Entry - version - */
  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[101] ); /* "local_version: Official version will be compared to this value" */
  gtk_entry_set_max_length( GTK_ENTRY(item), MAX_LENGTH_VERSION + 3 ); /* 50 */
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_LOCAL_VERSION] = item;

  /* - Automatic -------------- */
  item = gtk_radio_button_new_with_label_from_widget( GTK_RADIO_BUTTON(app->Gui.WatcherSettings[WS_RAD_VERSION_MANUAL]), app->Strs[52] ); /* "Version automatic:" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_local_automatic), &app->Gui );
  gtk_box_pack_start( GTK_BOX(vbox), item, FALSE, TRUE, 0);
  app->Gui.WatcherSettings[WS_RAD_VERSION_AUTO] = item;

  vbox_auto = APP_GTK_VBOX( FALSE, 0 );
  gtk_widget_set_sensitive( vbox_auto, 0 );
  gtk_container_add( GTK_CONTAINER(vbox), vbox_auto );
  app->Gui.WatcherSettings[WS_BOX_VERSION_AUTO] = vbox_auto;

  /* - check box - optionnal  */
  item = gtk_check_button_new_with_label( app->Strs[176] ); /* "Optionnal" */
  gtk_widget_set_tooltip_text( item, app->Strs[177] ); /* "optionnal: The watcher will be disable at start, if the executable is not available" */
  gtk_box_pack_start( GTK_BOX(vbox_auto), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_TOGGLE_OPTIONNAL] = item;

  /* Executable - Entry */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_auto), row );

  item = gtk_label_new( app->Strs[64] ); /* " Executable: " */
  gtk_widget_set_tooltip_text( item, app->Strs[102] ); /* "exec: Executable file launched to retrieve the local version" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[102] ); /* "exec: Executable file launched to retrieve the local version" */
  gtk_entry_set_max_length( GTK_ENTRY(item), 250 );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_LOCAL_EXEC] = item;

  /* Executable - Button */
  img = gtk_image_new();
  gtk_image_set_from_pixbuf( GTK_IMAGE(img), app->Gui.ImgTable[IMG_FILE] );

  item = gtk_button_new();
  gtk_button_set_image( (GtkButton*) item, img );
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_get_exec), 0 );
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_BUT_EXEC] = item;

  /* Argument - Entry */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_auto), row );

  item = gtk_label_new( app->Strs[65] ); /* " Argument: " */
  gtk_widget_set_tooltip_text( item, app->Strs[103] ); /* "arg: Argument given to the executable to retrieve the local version" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[103] ); /* "arg: Argument given to the executable to retrieve the local version" */
  gtk_entry_set_max_length( GTK_ENTRY(item), 250 );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_ARGUMENT] = item;

  group = gtk_frame_new( app->Strs[200] ); /* " In the command output, version can be retrieved: " */
  gtk_box_pack_start( GTK_BOX(vbox_auto), group, FALSE, TRUE, 0);

  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(group), vbox );

  /* stderr - check box */
  item = gtk_check_button_new_with_label( app->Strs[104] ); /* "From the error standard outpout" */
  gtk_widget_set_tooltip_text( item, app->Strs[105] ); /* "stderr: Because some developers don't understand what 'stderr' stand for" */
  gtk_box_pack_start( GTK_BOX(vbox), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_TOGGLE_STDERR] = item;

  /* Search frame ----------------------------------------------------------- */
  group = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox), group );
  app->Gui.WatcherSettings[WS_VBOX_SEARCH_LOCAL] = group;

  /* - Add search block */
  item = gtk_menu_bar_new();
  gtk_container_add( GTK_CONTAINER(vbox), item );
  Create_menu_search_block( item, CB_LOCAL_VERSION );
  app->Gui.WatcherSettings[WS_ADD_BLOCK_LOCAL] = item;
#endif
}

/*
  Purpose: Creating the an entry option activated by a switch/check_button
*/
static void              Create_entry_options(GtkWidget *parent, unsigned int id_switch, const char *label, const char *tooltip1, unsigned int id_entry, const char *tooltip2)
{
  GtkWidget              *row, *item;

#ifdef LIB_HANDY
  row = hdy_action_row_new();
  if( !( app->Flags & APP_FLG_PHONE )) {
    hdy_action_row_set_subtitle( (HdyActionRow*) row, tooltip1 );
    g_object_set( row, "subtitle-lines", 2, NULL );
  }
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, label );
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_switch_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[id_switch] = item;

  item = gtk_entry_new();
  g_signal_connect( G_OBJECT(app->Gui.WatcherSettings[id_switch]), "state-set", G_CALLBACK(Callback_switch_to_sensitive), item );
  gtk_container_add( GTK_CONTAINER(parent), item );
#else
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(parent), row );

  item = gtk_check_button_new_with_label( label );
  gtk_widget_set_tooltip_text( item, tooltip1 );
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[id_switch] = item;

  item = gtk_entry_new();
  g_signal_connect( G_OBJECT(app->Gui.WatcherSettings[id_switch]), "clicked", G_CALLBACK(Callback_check_to_sensitive), item );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
#endif

  if( tooltip2 )
    gtk_widget_set_tooltip_text( item, tooltip2 );
  gtk_entry_set_max_length( GTK_ENTRY(item), 500 );
  app->Gui.WatcherSettings[id_entry] = item;
}

/*
  Purpose: Creating the options frame of the window 'Add watcher'
*/
static void              Create_frame_options(GtkWidget *parent)
{
  GtkWidget              *group;

#ifdef LIB_HANDY
  group = hdy_preferences_group_new();
  hdy_preferences_group_set_title( (HdyPreferencesGroup*) group, app->Strs[170] ); /* "Trigger when..." */
  gtk_container_add( GTK_CONTAINER(parent), group );

  /* trigger different */
  Create_entry_options( group, WS_TOGGLE_MISMATCH, app->Strs[197], /* "...versions mismatch" */
			app->Strs[203], /* "trigger_on_mismatch: To trigger an action when a mismatch is detected between the official and local version */
			WS_ETY_TRIGGER_MISMATCH, app->Strs[204] ); /* "Entry is waiting an executable with absolute path. @name provides watcher's name, @local provides local version, @official provides official version to the executable" */

  /* trigger not found */
  group = hdy_preferences_group_new();
  gtk_container_add( GTK_CONTAINER(parent), group );

  Create_entry_options( group, WS_TOGGLE_ERROR, app->Strs[213], /* "... official version not found :" */
			app->Strs[214], /* "trigger_on_error: To trigger an action official version is not found inside the web resource" */
			WS_ETY_TRIGGER_ERROR, app->Strs[215] ); /* "Entry is waiting an executable with absolute path. @name provides watcher's name" */
#else
  GtkWidget              *vbox;

  /* Frame ---------------------------------------------------------------- */
  group = gtk_frame_new( app->Strs[170] ); /* "Trigger when..." */
  gtk_box_pack_start( GTK_BOX(parent), group, FALSE, TRUE, 0);

  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(group), vbox );

  /* trigger different */
  Create_entry_options( vbox, WS_TOGGLE_MISMATCH, app->Strs[197], /* "...versions mismatch" */
			app->Strs[203], /* "trigger_on_mismatch: To trigger an action when a mismatch is detected between the official and local version */
			WS_ETY_TRIGGER_MISMATCH, app->Strs[204] ); /* "Entry is waiting an executable with absolute path. @name provides watcher's name, @local provides local version, @official provides official version to the executable" */

  /* trigger not found */
  Create_entry_options( vbox, WS_TOGGLE_ERROR, app->Strs[213], /* "... official version not found :" */
			app->Strs[214], /* "trigger_on_error: To trigger an action official version is not found inside the web resource" */
			WS_ETY_TRIGGER_ERROR, app->Strs[215] ); /* "Entry is waiting an executable with absolute path. @name provides watcher's name" */
#endif
}

/*
  Purpose: Updating the 'Official result' entry after URL modification by the user
*/
static gboolean          Callback_url_focus_out(GtkWidget *widget, GdkEvent *event, gpointer param)
{
  s_gui                  *gui = param;
  const char             *entry;
  char                   *url;
  s_wtr                  *watcher;
  unsigned int           pos, length;

  watcher = Get_current_selected_watcher( gui );

  if( ! gtk_entry_get_text_length( GTK_ENTRY(gui->WatcherSettings[WS_ETY_OFFI_URL]) )
      || !( entry = gtk_entry_get_text( GTK_ENTRY(gui->WatcherSettings[WS_ETY_OFFI_URL]) )))
    return( FALSE );

  if( ! watcher || ! watcher->OfficialURL || strcmp( entry, watcher->OfficialURL ) || ! gui->TmpWatcher.CacheFile.Source.Length ) {
    if( !( url = strdup( entry )))
      return( FALSE );

    if( url[0] == 'h' && ! strncmp( url, "http://", 7 ))
      pos = 7;
    else {
      length = strlen( url );
      if( url[0] != 'h' || length < 8 || strncmp( url, "https://", 8 )) {
	length = strlen( url );
	REALLOC( url, 8 + length + 1 ); /* "https://" + '\0' */
	memmove( &url[8], url, length );
	memcpy( url, "https://", 8 );
	url[8 + length] = '\0';
      }
      pos = 8;
    }

    if( Domain_is_valid( &url[pos] )) {
      free( url );
      return( FALSE );
    }

    gui->TmpWatcher.OfficialURL = url;
    Update_official_result();
  }
  return( FALSE );
}

static void              Callback_url_activate(GtkSpinButton *spinbutton, gpointer param)
{
  s_gui                  *gui = param;

  if( !( app->Flags & APP_FLG_INIT ))
    Callback_url_focus_out( 0, 0, gui );
}

/*
  Purpose: Creating the frame 'Official version' for the 'Add watcher' window
*/
static void              Create_frame_official_version(GtkWidget *parent)
{
  GtkWidget              *group, *row, *item, *vbox_main, *vbox;

#ifdef LIB_HANDY
  group = hdy_preferences_group_new();
  hdy_preferences_group_set_title( (HdyPreferencesGroup*) group, app->Strs[192] ); /* "Retrieving official version" */
  gtk_container_add( GTK_CONTAINER(parent), group );

  /* URL - row */
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[119] ); /* " URL " */
  if( !( app->Flags & APP_FLG_PHONE )) {
    hdy_action_row_set_subtitle( (HdyActionRow*) row, app->Strs[126] ); /* "url: URL used to retrieve the web resource containing the official version" */
    g_object_set( row, "subtitle-lines", 2, NULL );
  }
  gtk_container_add( GTK_CONTAINER(group), row );

  /* URL - Entry */
  item = gtk_entry_new();
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  gtk_entry_set_max_length( GTK_ENTRY(item), 300 );
  g_signal_connect( G_OBJECT(item), "focus-out-event", G_CALLBACK(Callback_url_focus_out), &app->Gui );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_url_activate), &app->Gui );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_ETY_OFFI_URL] = item;

  /* - Spin - Timer */
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[122] ); /* " Check version every " */
  if( !( app->Flags & APP_FLG_PHONE )) {
    hdy_action_row_set_subtitle( (HdyActionRow*) row, app->Strs[124] ); /* "official_timer: Delay between 2 checks of the official version" */
    g_object_set( row, "subtitle-lines", 2, NULL );
  }
  gtk_container_add( GTK_CONTAINER(group), row );

  item = gtk_spin_button_new_with_range( 60, 525600, 1 );
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(item), DEFAULT_TIMER );
  gtk_widget_set_tooltip_text( item, "60 -> 525600" );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_SPIN_TIMER] = item;

  /* HTTP header */
  Create_entry_options( group, WS_TOGGLE_HEADER, app->Strs[115], /* "HTTP Header options" */
                        app->Strs[116], /* "HTTP header options separated with comma" */
                        WS_ETY_OFFI_HEADER, 0 );

  /* Search frame ----------------------------------------------------------- */
  group = hdy_preferences_group_new();
  gtk_container_add( GTK_CONTAINER(parent), group );

  row = hdy_expander_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[106] ); /* " In the web resource, version can be retrieved: " */
  if( !( app->Flags & APP_FLG_PHONE ))
    hdy_expander_row_set_subtitle( (HdyExpanderRow*) row, app->Strs[111] ); /* "Parameters to determine version" */
  gtk_container_add( GTK_CONTAINER(group), row );
  app->Gui.WatcherSettings[WS_EXPROW_OFFICIAL] = row;
  group = row;

  /* - Add search block */
  row = hdy_action_row_new();
  gtk_widget_set_halign( row, GTK_ALIGN_START );
  gtk_container_add( GTK_CONTAINER(group), row );

  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(row), vbox );
  app->Gui.WatcherSettings[WS_VBOX_SEARCH_OFFI] = vbox;

  item = gtk_menu_bar_new();
  gtk_container_add( GTK_CONTAINER(vbox), item );
  Create_menu_search_block( item, CB_OFFICIAL_VERSION );
  app->Gui.WatcherSettings[WS_ADD_BLOCK_OFFICIAL] = item;

  vbox_main = group;
#else
  GtkWidget              *frame_search;
  /* Main frame ------------------------------------------------------------- */
  group = gtk_frame_new( app->Strs[44] ); /* "Retrieving official version" */
  gtk_box_pack_start( GTK_BOX(parent), group, FALSE, TRUE, 0);

  vbox_main = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(group), vbox_main );

  /* - Entry - URL - */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_main), row );

  item = gtk_label_new( app->Strs[119] ); /* " URL " */
  gtk_widget_set_tooltip_text( item, app->Strs[126] ); /* "url: URL used to retrieve the web resource containing the official version" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_entry_set_max_length( GTK_ENTRY(item), 300 );
  g_signal_connect( G_OBJECT(item), "focus-out-event", G_CALLBACK(Callback_url_focus_out), &app->Gui );
  g_signal_connect( G_OBJECT(item), "activate", G_CALLBACK(Callback_url_activate), &app->Gui );
  gtk_widget_set_tooltip_text( item, app->Strs[126] ); /* "url: URL used to retrieve the web resource containing the official version" */
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_OFFI_URL] = item;

  /* HTTP header - label */
  Create_entry_options( vbox_main, WS_TOGGLE_HEADER, app->Strs[115], /* "HTTP Header options" */
                        app->Strs[116], /* "HTTP header options separated with comma" */
                        WS_ETY_OFFI_HEADER, 0 );

  /* - Spin - Timer */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_main), row );

  item = gtk_label_new( app->Strs[122] ); /* " Check version every " */
  gtk_widget_set_tooltip_text( item, app->Strs[124] ); /* "official_timer: Delay between 2 checks of the official version" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_spin_button_new_with_range( 60, 525600, 1 );
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(item), DEFAULT_TIMER );
  gtk_widget_set_tooltip_text( item, "60 -> 525600" );
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_SPIN_TIMER] = item;

  /* Search frame ----------------------------------------------------------- */
  frame_search = gtk_frame_new( app->Strs[106] ); /* " In the web resource, version can be retrieved: " */
  gtk_box_pack_start( GTK_BOX(vbox_main), frame_search, FALSE, TRUE, 0);

  /* - Add search block */
  vbox = APP_GTK_VBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(frame_search), vbox );
  app->Gui.WatcherSettings[WS_VBOX_SEARCH_OFFI] = vbox;

  item = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox_main), item, TRUE, TRUE, 0);
  app->Gui.WatcherSettings[WS_ADD_BLOCK_OFFICIAL] = item;
  Create_menu_search_block( item, CB_OFFICIAL_VERSION );
#endif

  /* Official version result */
  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_main), row );

  item = gtk_label_new( app->Strs[232] ); /* "Result : " */
  gtk_widget_set_tooltip_text( item, app->Strs[132] ); /* "Result of the search operations" */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_widget_set_tooltip_text( item, app->Strs[132] ); /* "Result of the search operations" */
  gtk_entry_set_max_length( GTK_ENTRY(item), 70 );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  gtk_widget_set_sensitive( item, 0 );
  app->Gui.WatcherSettings[WS_ETY_OFFI_RESULT] = item;
}

/*
  Purpose: Creating the name / icon frame
*/
static void              Create_frame_name(GtkWidget *parent)
{
  GtkWidget              *group, *row, *item, *img;

#ifdef LIB_HANDY
  group = hdy_preferences_group_new();
  gtk_container_add( GTK_CONTAINER(parent), group );

  /* - Entry - Name - */
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[45] ); /* " Name " */
  gtk_container_add( GTK_CONTAINER(group), row );

  item = gtk_entry_new();
  hdy_action_row_set_activatable_widget( (HdyActionRow*) row, item );
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  gtk_entry_set_max_length( GTK_ENTRY(item), MAX_LENGTH_NAME );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_ETY_NAME] = item;

  /* - Button - Icon - */
  row = hdy_action_row_new();
  hdy_preferences_row_set_title( (HdyPreferencesRow*) row, app->Strs[46] ); /* " Icon " */
  if( !( app->Flags & APP_FLG_PHONE )) {
    hdy_action_row_set_subtitle( (HdyActionRow*) row, app->Strs[112] ); /* "icon: To choose the image file which will represent the watcher" */
    g_object_set( row, "subtitle-lines", 2, NULL );
  }
  gtk_container_add( GTK_CONTAINER(group), row );

  img = gtk_image_new();
  gtk_image_set_from_pixbuf( GTK_IMAGE(img), app->Gui.ImgTable[IMG_DEFAULT_WATCHER] );
  app->Gui.WatcherSettings[WS_IMG_LOGO] = img;

  item = gtk_button_new();
  gtk_widget_set_valign( item, GTK_ALIGN_CENTER );
  gtk_widget_set_size_request( item, 38, 38 );
  gtk_button_set_image( (GtkButton*) item, img );
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_add_logo), &app->Gui );
  gtk_container_add( GTK_CONTAINER(row), item );
  app->Gui.WatcherSettings[WS_BUT_LOGO] = item;
#else
  group = gtk_frame_new( app->Strs[30] ); /* "Watcher" */
  gtk_box_pack_start( GTK_BOX(parent), group, FALSE, TRUE, 0);

  row = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(group), row );

  /* - Entry - Name - */
  item = gtk_label_new( app->Strs[45] ); /* " Name " */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );

  item = gtk_entry_new();
  gtk_entry_set_max_length( GTK_ENTRY(item), MAX_LENGTH_NAME );
  gtk_box_pack_start( GTK_BOX(row), item, TRUE, TRUE, 0 );
  app->Gui.WatcherSettings[WS_ETY_NAME] = item;

  /* - Button - Icon - */
  item = gtk_label_new( app->Strs[46] ); /* " Icon " */
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );
  gtk_widget_set_tooltip_text( item, app->Strs[112] ); /* "icon: To choose the image file which will represent the watcher" */

  img = gtk_image_new();
  gtk_image_set_from_pixbuf( GTK_IMAGE(img), app->Gui.ImgTable[IMG_DEFAULT_WATCHER] );
  app->Gui.WatcherSettings[WS_IMG_LOGO] = img;

  item = gtk_button_new();
  gtk_widget_set_size_request( item, 38, 38 );
  gtk_button_set_image( (GtkButton*) item, img );
  gtk_widget_set_tooltip_text( item, app->Strs[112] ); /* "icon: To choose the image file which will represent the watcher" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_add_logo), &app->Gui );
  gtk_box_pack_start( GTK_BOX(row), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_BUT_LOGO] = item;
#endif

  app->Gui.ChooseFile = gtk_file_chooser_dialog_new( app->Strs[50], GTK_WINDOW(app->Gui.WatcherSettings[WS_WIN_MAIN]),  /* "Logo choice" */
						     GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
						     GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL );
  gtk_window_set_modal( GTK_WINDOW(app->Gui.ChooseFile), TRUE );
}

/*
  Purpose: Creating the window to add or modify a watcher
*/
void                     Create_window_watcher_settings()
{
  GtkWidget              *win, *vbox_main, *line, *align, *hbox, *vbox, *item;
  GError                 *err;

  /* Window - Adding a watcher --------------------------------------------- */
#ifdef LIB_HANDY
  if( app->Flags & APP_FLG_PHONE )
    win = hdy_window_new();
  else
#endif
    win = gtk_window_new( GTK_WINDOW_TOPLEVEL );
  app->Gui.WatcherSettings[WS_WIN_MAIN] = win;
  g_signal_connect( G_OBJECT(win), "delete_event", G_CALLBACK(Callback_on_delete_cancel), &app->Gui );
  g_signal_connect( G_OBJECT(win), "destroy", G_CALLBACK(Callback_on_destroy), 0 );

  gtk_window_set_icon_from_file( GTK_WINDOW(win), DEFAULT_APP_RESOURCES"/icon_win_add.png", &err );

  vbox_main = APP_GTK_VBOX( FALSE, 0 );

#ifdef LIB_HANDY
  GtkWidget *leaf = hdy_leaflet_new();
  gtk_container_add( GTK_CONTAINER(win), leaf );
  gtk_widget_show_all( leaf );

  gtk_container_add( GTK_CONTAINER(leaf), vbox_main );

  if( app->Flags & APP_FLG_PHONE ) {
    GtkWidget *bar = hdy_header_bar_new();
    g_object_set( bar,
		  "visible", TRUE,
		  "show_close_button", TRUE,
		  "can-focus", FALSE,
		  NULL );
    gtk_container_add( GTK_CONTAINER(vbox_main), bar );
    app->Gui.WatcherSettings[WS_BAR_WIN] = bar;
  }
#else
  Set_window_properties( win, vbox_main );
#endif

  /* Frames ----------------------------------------------------------------- */
#ifdef LIB_HANDY
  vbox = hdy_preferences_page_new();
  gtk_container_add( GTK_CONTAINER(vbox_main), vbox );

  Create_frame_name( vbox );
  Create_frame_official_version( vbox );
  Create_frame_local_version( vbox );
  Create_frame_options( vbox );
#else
  if( app->Flags & APP_FLG_PHONE )
    hbox = vbox = vbox_main ;
  else {
    hbox = APP_GTK_HBOX( FALSE, 0 );
    gtk_container_add( GTK_CONTAINER(vbox_main), hbox );

    vbox = APP_GTK_VBOX( FALSE, 0 );
    gtk_container_add( GTK_CONTAINER(hbox), vbox );
  }

  Create_frame_name( vbox );
  Create_frame_official_version( vbox );
  Create_frame_local_version( hbox );
  Create_frame_options( vbox_main );
#endif

  /* Bottom buttons --------------------------------------------------------- */
  line = APP_GTK_HSEPAR();
  gtk_container_add( GTK_CONTAINER(vbox_main), line );

#if GTK_MAJOR_VERSION < 3 || ( GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION < 14 )
  align = gtk_alignment_new( 1, 0, 0, 0 );
  gtk_container_add( GTK_CONTAINER(vbox_main), align );
#else
  align = vbox_main;
#endif

  hbox = APP_GTK_HBOX( FALSE, 0 );
  gtk_container_add( GTK_CONTAINER(align), hbox );
#if GTK_MAJOR_VERSION == 3 && GTK_MINOR_VERSION > 14
  gtk_widget_set_halign( hbox, GTK_ALIGN_END );
#endif

  item = gtk_label_new( app->Strs[53] ); /* " !! Incomplete !! " */
  app->Gui.WatcherSettings[WS_LBL_ERROR] = item;
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );

  line = APP_GTK_VSEPAR();
  gtk_container_add( GTK_CONTAINER(hbox), line );

  /* - Add - */
  item = gtk_button_new_with_mnemonic( app->Strs[33] ); /* "_Add" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_add_watcher), &app->Gui );
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_BUT_ADD] = item;

  /* - Save - */
  item = gtk_button_new_with_mnemonic( app->Strs[16] ); /* "_Save" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_save_watcher), &app->Gui );
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );
  app->Gui.WatcherSettings[WS_BUT_SAVE] = item;

  /* - Cancel - */
  item = gtk_button_new_with_mnemonic( app->Strs[17] ); /* "C_ancel" */
  g_signal_connect( G_OBJECT(item), "clicked", G_CALLBACK(Callback_cancel), &app->Gui );
  gtk_box_pack_start( GTK_BOX(hbox), item, FALSE, FALSE, 0 );

  /* End ---------------------------------------------------------------- */
  gtk_widget_show_all( vbox_main );
  gtk_widget_hide( app->Gui.WatcherSettings[WS_BUT_SAVE] );
  gtk_widget_hide( app->Gui.WatcherSettings[WS_LBL_ERROR] );
  if( app->Flags & APP_FLG_PHONE )
    gtk_window_move( GTK_WINDOW(win), 0, 0 );
}
