/*
 *  'UpdatesWarner' permits you to survey new versions
 *  Copyright (C) 2017-2021 Vincent Deca. known as Bigcake@ubuntu-fr.org
 *  (user@forum not email)
 *
 *  vincent.deca@neutralite.org
 *
 *  This file is part of 'UpdatesWarner'.
 *
 * 'UpdatesWarner' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  'UpdatesWarner' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'UpdatesWarner'.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "app.h"

/*
  Gobal pointer to the main structure of the application
  - The content of the structure should be modified with locking corresponding mutexes if required
*/
s_app             * restrict app;

/*
  Purpose: Existing with a message and a system error (or not)
*/
void              Exit_with_msg(const char *msg, char flg)
{
  if( app->Flags & APP_FLG_NO_COLORS )
    fprintf( stderr, "\n%s : %s", app->Strs[0], msg ); /* "Error" */
  else
    fprintf( stderr, "\n\033[31m%s\033[0m : %s", app->Strs[0], msg ); /* "Error" */
  if( flg )
    fprintf( stderr, ": %s\n\n", strerror( errno ));
  else
    fprintf( stderr, "\n\n" );
  exit( EXIT_FAILURE );
}

/*
  Purpose: Quitting with needed cleans / saves / ...
*/
void              Quit_app()
{
  /*
    - no sensible data to clean for the moment
    - no user data to save for the moment
    - nothing else to do, system will free resources much faster than me
  */
  if( app->Flags & APP_FLG_SYSTRAY )
    gtk_status_icon_set_visible( app->Systray, 0 ); /* Too ease some WM */
  exit( EXIT_SUCCESS );
}

/*
  Purpose: Getting the current user home path
*/
static void              Get_user_home()
{
  struct passwd          *info;
  const char             *home = 0;

  home = getenv( "HOME" );
  if( home && home[0] == '/' )
    goto found;

  if(( info = getpwuid( getuid() )) != NULL )
    if( info->pw_dir && info->pw_dir[0] == '/' ) {
      home = info->pw_dir;
      goto found;
    }

  ERR_APP( "%s", app->Strs[244] ); /* "Unable to determine user home path, fallback to /tmp/uw_cache" */
  STRDUP( home, "/tmp/uw_cache" );

 found:
  app->UserHome = home;
  app->UserHomeLength = strlen( home );
}

/*
  Purpose: Initializing the config folder
*/
static void       Prepare_config_folder( )
{
  struct stat     st;
  unsigned int    length;

#ifdef MODE_DEV
  STRDUP( app->Gui.File.Name, "./"DEFAULT_CONFIG_FILE );
  return;
#endif /* MODE_DEV */

  length = app->UserHomeLength;

  MALLOC( app->Gui.File.Name, length + 1 + sizeof(DEFAULT_FOLDER_CONFIG) + sizeof(DEFAULT_CONFIG_FILE) ); /* '/', second '/' and '\0' are counted with sizeof() */

  memcpy( app->Gui.File.Name, app->UserHome, length );
  if( app->UserHome[length - 1] != '/' )
    app->Gui.File.Name[length++] = '/';

  memcpy( &app->Gui.File.Name[length], DEFAULT_FOLDER_CONFIG, sizeof(DEFAULT_FOLDER_CONFIG) );

  if( stat( app->Gui.File.Name, &st )) {
    if( mkdir( app->Gui.File.Name, S_IRWXU )) {
      ERR_SYS( "mkdir( %s )", app->Gui.File.Name );
      exit( EXIT_FAILURE );
    }
    if( stat( app->Gui.File.Name, &st )) {
      ERR_SYS( "stat( %s )", app->Gui.File.Name );
      exit( EXIT_FAILURE );
    }
  }

  if( ! S_ISDIR( st.st_mode )) {
    ERR_APP( "%s: %s", app->Gui.File.Name, app->Strs[156] ); /* "Already exists and is not a folder" */
    exit( EXIT_FAILURE );
  }

  length += sizeof(DEFAULT_FOLDER_CONFIG) - 1;
  app->Gui.File.Name[length++] = '/';
  memcpy( &app->Gui.File.Name[length], DEFAULT_CONFIG_FILE, sizeof(DEFAULT_CONFIG_FILE) );
}

/*
  Purpose: Giving the default cache folder name
*/
char              *Get_default_cache_folder_name()
{
  char            *name;

  MALLOC( name, app->UserHomeLength + 1 + sizeof(DEFAULT_FOLDER_CACHE) ); /* '/', '\0' is counted with sizeof() */

  memcpy( name, app->UserHome, app->UserHomeLength );
  name[app->UserHomeLength] = '/';

  memcpy( &name[app->UserHomeLength + 1], DEFAULT_FOLDER_CACHE, sizeof(DEFAULT_FOLDER_CACHE) );
  return( name );
}

/*
  Purpose: Initializing the cache folder
*/
static void       Prepare_cache_folder()
{
  struct stat     st;

  if( ! app->CacheFolder )
    if( !( app->CacheFolder = Get_default_cache_folder_name()))
      goto Get_cache_FAIL;

  if( stat( app->CacheFolder, &st )) {
    if( mkdir( app->CacheFolder, S_IRWXU )) {
      ERR_SYS( "%s: mkdir( %s )", app->Strs[194], app->CacheFolder ); /* "Switching cache folder to a fallback folder" */
      goto Get_cache_FAIL;
    }
    if( stat( app->CacheFolder, &st )) {
      ERR_SYS( "%s: stat( %s )", app->Strs[194], app->CacheFolder ); /* "Switching cache folder to a fallback folder" */
      goto Get_cache_FAIL;
    }
  }

  if( ! S_ISDIR( st.st_mode )) {
    ERR_APP( "%s: %s: %s", app->CacheFolder, app->Strs[156], /* "Already exists and is not a folder" */
	     app->Strs[194] ); /* "Switching cache folder to a fallback folder" */
    goto Get_cache_FAIL;
  }

  return;

 Get_cache_FAIL:
  free( app->CacheFolder );
  STRDUP( app->CacheFolder, "/tmp/uw_cache" );

  if( stat( app->CacheFolder, &st )) {
    if( mkdir( app->CacheFolder, S_IRWXU )) {
      ERR_SYS( "%s: mkdir( %s )", app->Strs[196], app->CacheFolder ); /* "You will have to create the cache folder yourself, because me... I give up..." */
      exit( EXIT_FAILURE );
    }
    if( stat( app->CacheFolder, &st )) {
      ERR_SYS( "%s: stat( %s )", app->Strs[196], app->CacheFolder ); /* "You will have to create the cache folder yourself, because me... I give up..." */
      exit( EXIT_FAILURE );
    }
  }

  if( ! S_ISDIR( st.st_mode )) {
    ERR_APP( "%s: %s: That's it ! I give up !", app->CacheFolder, app->Strs[156] ); /* "Already exists and is not a folder" */
    sleep( 2 );
    ERR_APP( "Geez... %s", app->Strs[196] ); /* "You will have to create the cache folder yourself, because me... I give up..." */
    exit( EXIT_FAILURE );
  }
}

/*
  Purpose: Loading language adapted to env
*/
static void            Init_language()
{
  const char           *envir;
  struct dirent        *dp;
  DIR                  *dirp;
  char                 *curr_dir = 0;
  int                  fd;
  struct stat          st;

  app->Strs = English;
  envir = getenv( "LANG" );
  if( ! envir )
    return;

  if( !( curr_dir = getcwd( NULL, 0 ))) {
    ERR_SYS( "getcwd( %s )", APP_LANG_DIR );
    return;
  }
  if( !( dirp = opendir( APP_LANG_DIR ))) {
    ERR_SYS( "opendir( %s )", APP_LANG_DIR );
    free( curr_dir );
    return;
  }
  if(( fd = dirfd( dirp )) < 0 ) {
    ERR_SYS( "dirfd( %s )", APP_LANG_DIR );
  }
  else if( fchdir( fd )) {
    ERR_SYS( "fchdir( %s )", APP_LANG_DIR );
  }
  else {
    while(( dp = readdir( dirp )) > 0 ) {
      if( ! strncasecmp( envir, dp->d_name, 2 )
	  && stat( dp->d_name, &st ) != -1 && S_ISREG( st.st_mode )) {
	Load_file_lang( dp->d_name );
	break;
      }
    }
  }
  if( chdir( curr_dir )) {
    ERR_SYS( "chdir( %s )", curr_dir );
  }
  free( curr_dir );
  closedir( dirp );
}


/*
  Purpose: Tries to explain why gtk_init_check() failed
*/
static void       Error_gtk_init()
{
  const char           *envir;

  if( getenv( "SSH_CLIENT" ) || getenv( "SSH_TTY" )) {
    envir = getenv( "DISPLAY" );
    if( envir )
      printf("export DISPLAY=%s\n", envir );
    else {
      WARN_APP( "%s", app->Strs[243] ); /* "You probably forgot to export DISPLAY=:0" */
    }
  }
}

/*
  Purpose: Initializing application
*/
static void       Init_app()
{
  GError          *err = 0;
  int             a;

  /* ------------------------- APP initialization ----------------------- */
  if(( app = calloc( 1, sizeof(*app) )) == NULL )
    Exit_with_msg( English[1], 1 ); /* "Couldn't allocate enough memory" */

  Init_language();

  app->Flags = APP_FLG_INIT | APP_FLG_SYSTRAY;

  /* ------------------------- Thread initialization ----------------------- */
  if(( app->CoreCount = sysconf( _SC_NPROCESSORS_ONLN )) < 1 ) {
    ERR_SYS( "%s", app->Strs[55] ); /* "Unable to retrieve number of CPU cores" */
    app->CoreCount = 1;
  }

  CALLOC( app->IdWorkers, app->CoreCount + 1, sizeof(*app->IdWorkers) );
  CALLOC( app->ThreadFlags, app->CoreCount, sizeof(*app->ThreadFlags) );

  if( pthread_attr_init( &app->Attr ) || pthread_attr_setdetachstate( &app->Attr, PTHREAD_CREATE_DETACHED ))
    Exit_with_msg( app->Strs[56], 1 ); /* "Unable to initialize thread attributs" */

  if( pthread_mutex_init( &app->LockCache, NULL )
      || pthread_mutex_init( &app->LockPrint, NULL )
      || pthread_mutex_init( &app->LockCond, NULL )
      || pthread_cond_init( &app->TimerCond, NULL )
      || pthread_mutex_init( &app->Gui.LockWatchers, NULL )
      || pthread_mutex_init( &app->Gui.File.LockFile, NULL ))
    Exit_with_msg( app->Strs[57], 1 ); /* "Unable to initialize a mutex" */

  /* ------------------------- GTK initialization ----------------------- */
#if GLIB_MAJOR_VERSION < 3 || ( GLIB_MAJOR_VERSION == 3 && GLIB_MINOR_VERSION < 36 )
  g_type_init();
#endif

#ifndef GLIB_VERSION_2_26
  g_thread_init(NULL);

  //#elif GDK_MAJOR_VERSION == 3
  //#  if GDK_MINOR_VERSION < 6
  //gdk_threads_init();
  //#  else
  //#todo gdk_threads_add_idle () ?
  //#  include <X11/Xlib.h>
  //XInitThreads();
  //#  endif
#else
  gdk_threads_init();
#endif

  if( gtk_init_check( NULL, NULL ) == FALSE ) {
    ERR_APP( "%s: gtk_init_check() failed", app->Strs[10] ); /* "Failed to initialize GTK library" */
    Error_gtk_init();
    exit( EXIT_FAILURE );
  }
#if GTK_MAJOR_VERSION >= 3
  //  getpid();
  app->Gui.App = gtk_application_new( "org.gtkmm." APP_SH_NAME, G_APPLICATION_NON_UNIQUE );
#endif

  /* ------------------------- GUI initialization ----------------------- */
  app->Gui.MainX = 200;
  app->Gui.MainY = 200; //#todo: middle of current monitor would be nice
  app->Gui.MaxXWatcher = DEFAULT_COUNT_WIDTH;
  app->Gui.MaxYWatcher = DEFAULT_COUNT_HEIGHT;
  app->Gui.WatcherSizeX = DEFAULT_ICON_WIDTH;
  app->Gui.WatcherSizeY = DEFAULT_ICON_HEIGHT;
  app->Gui.Flags |= GUI_FLG_ON_TOP | GUI_FLG_NO_WATCHER | GUI_FLG_FIRST_INIT;

  app->Gui.MainColor.red = DEFAULT_COLOR_RED;
  app->Gui.MainColor.green = DEFAULT_COLOR_RED;
  app->Gui.MainColor.blue = DEFAULT_COLOR_RED;
#if GTK_MAJOR_VERSION >= 3
  app->Gui.MainColor.alpha = DEFAULT_COLOR_ALPHA;
#else
  app->Gui.MainOpacity = DEFAULT_OPACITY;
#endif

  CALLOC( app->Gui.ImgFile, IMG_COUNT, sizeof(*app->Gui.ImgFile) );
  app->Gui.ImgFile[IMG_DEFAULT_WATCHER] = DEFAULT_APP_RESOURCES"/default_watcher26.png";
  app->Gui.ImgFile[IMG_ADD] = DEFAULT_APP_RESOURCES"/icon_add.png";
  app->Gui.ImgFile[IMG_ABOUT] = DEFAULT_APP_RESOURCES"/icon_about.png";
  app->Gui.ImgFile[IMG_DEL] = DEFAULT_APP_RESOURCES"/icon_del.png";
  app->Gui.ImgFile[IMG_APP] = DEFAULT_APP_RESOURCES"/updates-warner.png";
  app->Gui.ImgFile[IMG_VERSION_ERROR] = DEFAULT_APP_RESOURCES"/version_error26.png";
  app->Gui.ImgFile[IMG_VERSION_MISMATCH] = DEFAULT_APP_RESOURCES"/version_mismatch26.png";
  app->Gui.ImgFile[IMG_VERSION_IN_PROGRESS] = DEFAULT_APP_RESOURCES"/in_progress26.png";
  app->Gui.ImgFile[IMG_FILE] = DEFAULT_APP_RESOURCES"/icon_file.png";
  app->Gui.ImgFile[IMG_MODIFY] = DEFAULT_APP_RESOURCES"/icon_config_watcher.png";
  app->Gui.ImgFile[IMG_REFRESH] = DEFAULT_APP_RESOURCES"/icon_refresh.png";
  app->Gui.ImgFile[IMG_SETTINGS] = DEFAULT_APP_RESOURCES"/icon_preferences.png";
  app->Gui.ImgFile[IMG_INFOS_BOTTOM] = DEFAULT_APP_RESOURCES"/infos_bottom.png";
  app->Gui.ImgFile[IMG_INFOS_TOP] = DEFAULT_APP_RESOURCES"/infos_top.png";
  app->Gui.ImgFile[IMG_INFOS_LEFT] = DEFAULT_APP_RESOURCES"/infos_left.png";
  app->Gui.ImgFile[IMG_INFOS_RIGHT] = DEFAULT_APP_RESOURCES"/infos_right.png";

  CALLOC( app->Gui.ImgTable, IMG_COUNT, sizeof(*app->Gui.ImgTable) );
  app->Gui.ImgCount = IMG_COUNT;
  for( a = 0; a < IMG_COUNT; ++a) {
    if( !( app->Gui.ImgTable[a] = gdk_pixbuf_new_from_file( app->Gui.ImgFile[a], &err ))) {
      ERR_APP( "'%s': %s\n=> GTK: %s", app->Gui.ImgFile[a], app->Strs[135], err->message ); /* "Line" "Unable to load image file" */
    }
    g_object_ref( G_OBJECT( app->Gui.ImgTable[a] )); /* 'cause I don't want that 'self-declared smarter than everyone' GTK to mess up my pix */
  }

  if( ! app->Gui.ImgTable[IMG_DEFAULT_WATCHER] ) {
#include "fallback_icons.c" /* mmmm I like it sooo much, don't you ? */
    app->Gui.ImgTable[IMG_DEFAULT_WATCHER] = gdk_pixbuf_new_from_xpm_data( default_watcher26_xpm );
    if( ! app->Gui.ImgTable[IMG_VERSION_ERROR] )
      app->Gui.ImgTable[IMG_VERSION_ERROR] = gdk_pixbuf_new_from_xpm_data( version_error26_xpm );
    if( ! app->Gui.ImgTable[IMG_VERSION_MISMATCH] )
      app->Gui.ImgTable[IMG_VERSION_MISMATCH] = gdk_pixbuf_new_from_xpm_data( version_mismatch26_xpm );
    if( ! app->Gui.ImgTable[IMG_VERSION_IN_PROGRESS] )
      app->Gui.ImgTable[IMG_VERSION_IN_PROGRESS] = gdk_pixbuf_new_from_xpm_data( in_progress26_xpm );
    g_object_ref( G_OBJECT( app->Gui.ImgTable[IMG_DEFAULT_WATCHER] )); /* Don't you love when GTK destroys a resource because it knows you won't use it later ? */
    g_object_ref( G_OBJECT( app->Gui.ImgTable[IMG_VERSION_ERROR] )); /* YES THAT WAS SARCASTIC ! */
    g_object_ref( G_OBJECT( app->Gui.ImgTable[IMG_VERSION_MISMATCH] )); /* WHY THE F$#ù! DON'T YOU LET ME MANAGE MY MEMORY !???!?!! */
    g_object_ref( G_OBJECT( app->Gui.ImgTable[IMG_VERSION_IN_PROGRESS] )); /* I hate your resources management sooo much */
  }

#ifdef LIB_HANDY
  hdy_init();
#endif

  /* ------------------------------ Others ------------------------------ */
  signal( SIGCHLD, SIG_IGN );
  Get_user_home();
}

/*
  Purpose: Waking up the worker threads every minutes
*/
static void       *th_Timer(void *param)
{
  while( 1 ) {
    sleep( 60 );
    pthread_cond_broadcast( &app->TimerCond );
  }
  return( 0 );
}

/*
  Purpose: Launching all needed threads
*/
static void       Launch_threads()
{
  int             a;

  app->Flags &= ~( APP_FLG_INIT );
#pragma GCC diagnostic ignored "-Wrestrict" /* Useless use of restrict from pthread_create() in 4th parameter */
  for( a = 0; a < app->CoreCount; ++a )
    if( pthread_create( &app->IdWorkers[a], &app->Attr, th_Updater, &app->IdWorkers[a] ))
      Exit_with_msg( app->Strs[179], 1 ); /* "Unable to create a thread" */

  if( pthread_create( &app->IdWorkers[a++], &app->Attr, th_Timer, 0 ))
    Exit_with_msg( app->Strs[179], 1 ); /* "Unable to create a thread" */
}

/*
  Purpose: Initializing the GUI
*/
static void       Build_GUI()
{
  int             width, height;

  Create_window_main();
  Create_window_about();
  Create_window_watcher_settings();
  Create_window_template_settings();
  Create_systray();
  Create_window_preferences();
  Create_window_error();
  Create_menu_watcher();
  Create_watchers();

  bzero( &app->Gui.CancelPref, sizeof(app->Gui.CancelPref) );
  gtk_widget_show( app->Gui.Main[MW_WIN_MAIN] );

  SCREEN_WIDTH_HEIGHT( width, height );
  if( app->Gui.MainX < width && app->Gui.MainY < height )
    gtk_window_move( GTK_WINDOW(app->Gui.Main[MW_WIN_MAIN]), app->Gui.MainX, app->Gui.MainY );
  if( app->Flags & APP_FLG_PHONE ) {
    gtk_widget_hide( app->Gui.PreferenceSettings[PS_COMBOROW_EXPANSION] );
    gtk_widget_hide( app->Gui.PreferenceSettings[PS_HBOX_ALWAYS_ON] );
    gtk_widget_hide( app->Gui.PreferenceSettings[PS_ROW_LOCK_POS] );
  }
  Launch_threads();
}

int                main(int ac, char **av)
{
  Init_app();
  Get_options( ac, av );

  Prepare_config_folder();
  Load_file_config( &app->Gui.File, app->Gui.File.Name );

  if( app->Flags & APP_FLG_PHONE ) {
    app->Flags &= ~( APP_FLG_SYSTRAY );
    app->Gui.MainX = 0;
    app->Gui.MainY = 0;
    if( app->Gui.WatcherSizeX == DEFAULT_ICON_WIDTH ) {
      app->Gui.WatcherSizeX = 2 * DEFAULT_ICON_WIDTH;
      app->Gui.WatcherSizeY = 2 * DEFAULT_ICON_HEIGHT;
    }
  }

  Prepare_cache_folder();

#if GTK_MAJOR_VERSION < 3
  Build_GUI();
  gtk_main();
#else
  g_signal_connect( app->Gui.App, "activate", G_CALLBACK( Build_GUI ), NULL );
  if( g_application_run( G_APPLICATION( app->Gui.App ), 0, 0 ))
    Exit_with_msg( app->Strs[10], 1 ); /* "Failed to initialize GTK library" */
#endif

  Quit_app();
}
